﻿namespace NETing.ExperimentsConsole.Experiments.Abstract
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using Commands;
    using Index.Controller.Configurations;
    using System.Linq;

    internal abstract class AbstractConsoleCommand : IIndexConsoleCommand
    {
        protected AbstractIndexConsole IndexConsole { get; private set; }

        public abstract string Name { get; }
        public abstract string ShortDesc { get; }

        public int KeyPrefix { get; set; }

        public void Setup(AbstractIndexConsole console)
        {
            IndexConsole = console;
        }

        abstract public void Execute(string arg);

        #region Input helpers

        internal static string GetStringInput(string message, string defaultValue)
        {
            Console.Write(message + " (default: " + defaultValue + "): ");
            string input = Console.ReadLine();
            string result = !string.IsNullOrWhiteSpace(input) && !string.IsNullOrEmpty(input) ? input.Trim() : defaultValue;
            Console.WriteLine("Will use: " + result);
            return result;
        }

        internal static int GetIntegerInput(string message, int defaultValue)
        {
            Console.Write(message + " (default: " + defaultValue + "): ");
            int result = int.TryParse(Console.ReadLine() ?? "", out result) ? result : defaultValue;
            Console.WriteLine("Will use: " + result);
            return result;
        }

        internal static bool GetBooleanInput(string message, bool defaultValue)
        {
            Console.Write(message + "(Y/N) (default: " + (defaultValue ? "Y" : "N") + "): ");
            string input = Console.ReadLine();
            input = (string.IsNullOrEmpty(input) ? (defaultValue ? "Y" : "N") : input).Trim().ToLowerInvariant();

            bool result = input == "y" ? true : false;
            Console.WriteLine("Will use: " + (result ? "Y" : "N"));
            return result;
        }

        internal static IEnumerable<IPostingFileConfig> GetPostingFileConfigs(
            string message = "Select posting file configuration to use.")
        {
            Console.WriteLine(message);

            int index = 0;
            foreach (var cfg in Configure.configs)
            {
                Console.WriteLine(index++ + ": " + cfg.Name);
            }

            Console.Write("Select configs, if multiple separate by space: (default: all) ");
            string[] inp = (Console.ReadLine() ?? "").Trim().Split(' ');
            var selected = new List<IPostingFileConfig>();

            if (inp.Length == 0 || string.IsNullOrEmpty(inp[0]))
            {
                selected.AddRange(Configure.configs);
            }
            else
            {
                selected.AddRange(inp
                    .Where(s => int.TryParse(s, out index))
                    .Select(s => Configure.configs[index]));
            }

            var builder = new StringBuilder();
            builder.Append("Will use: ");
            foreach (IPostingFileConfig config in selected)
            {
                builder.Append(config.Name + ", ");
            }

            builder.Length -= 2;
            Console.WriteLine(builder.ToString());

            return selected;
        }

        internal static FileInfo GetFileInput(string message, bool existing = true)
        {
            Console.WriteLine(message);
            var file = new FileInfo(Console.ReadLine() ?? "");
            while (existing && !file.Exists)
            {
                Console.WriteLine("File must exist, enter new path.");
                file = new FileInfo(Console.ReadLine() ?? "");
            }

            return file;
        }

        internal static IEnumerable<FileInfo> GetInputFiles(string message, bool existing = true)
        {
            Console.WriteLine(message);
            Console.WriteLine("You can enter several files, stop entry by writing #.");
            var files = new List<FileInfo>();
            Console.Write(" >");
            string inp = Console.ReadLine() ?? "";
            while (inp != "#")
            {
                var file = new FileInfo(inp);
                while (existing && !file.Exists && inp == "#")
                {
                    inp = Console.ReadLine() ?? "";
                    file = new FileInfo(inp);
                }

                if (inp == "#")
                {
                    return files;
                }

                files.Add(file);

                Console.Write(" >");
                inp = Console.ReadLine() ?? "";

                if (inp == "#")
                {
                    return files;
                }
            }

            return files;
        }

        internal static DirectoryInfo GetDirectoryInput(string message, bool existing = false, bool create = true)
        {
            Console.WriteLine(message);
            var dir = new DirectoryInfo(Console.ReadLine() ?? "");

            while (existing && !dir.Exists)
            {
                Console.WriteLine("Directory must exist, enter new path.");
                dir = new DirectoryInfo(Console.ReadLine() ?? "");
            }

            if (!dir.Exists && create)
            {
                dir.Create();
                Console.WriteLine("Created directory: " + dir);
            }

            return dir;
        }

        public static T GetEnumSelection<T>(string message)
        {
            Console.WriteLine(message);
            string[] names = typeof(T).GetEnumNames();
            for (int i = 0; i < names.Length; i++)
            {
                Console.WriteLine(i + ": " + names[i]);
            }

            object selected = GetIntegerInput("Select one: ", 0);
            Console.WriteLine("Will use: " + names[(int)selected]);
            return (T)selected;
        }

        #endregion
    }
}
