﻿namespace NETing.ExperimentsConsole.Experiments.Abstract
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    
    using Commands;
    
    using Index.Controller;
    using Index.Controller.Configurations;
    using Index.Controller.Configurations.ConcreteConfigurations;
    using Index.Documents.DocumentModel.Scope;

    internal abstract class AbstractIndexConsole
    {
        private readonly IIndexConsoleCommand[] basicCommands = new IIndexConsoleCommand[]
        {
            new PrintState(),
            new ListCommands(),
            new Configure(),
            new Shutdown(),
        };

        private sealed class DefaultBuilderServiceConfig : IBuilderServiceConfig
        {
            public int StoreSize
            {
                get { return (1024 + 512) * 1024 * 1024; }
            }

            public int InitialListSize
            {
                get { return 16; }
            }
        }

        internal sealed class DefaultIndexSystemConfig : IIndexSystemConfig
        {
            public DefaultIndexSystemConfig()
            {
                BlockSize = 16 * 1024;
                BufferCount = 1024 << 8;
                BufferSlices = 8;
                BufferFlushInterval = 100;
                BuffersReadOnly = false;
            }

            public int BlockSize { get; set; }

            public int BufferCount { get; set; }

            public int BufferSlices { get; set; }

            public bool BuffersReadOnly { get; set; }

            public int BufferFlushInterval { get; set; }
        }

        internal bool running = true;

        internal readonly DefaultIndexSystemConfig IndexSystemConfig = new DefaultIndexSystemConfig();
        internal readonly IBuilderServiceConfig BuilderSreviceConfig = new DefaultBuilderServiceConfig();

        internal IPostingFileConfig PostingFileConfig;
        internal StorageProfile StorageProfile;
        internal IndexController Controller;

        private readonly IList<IIndexConsoleCommand> commands = new List<IIndexConsoleCommand>();

        private readonly IDictionary<string, IIndexConsoleCommand> commandMapping
            = new Dictionary<string, IIndexConsoleCommand>();

        private readonly IDictionary<string, IIndexConsoleCommand> keyMapping
            = new Dictionary<string, IIndexConsoleCommand>();

        internal void ConfigureIndexSystem(IPostingFileConfig postingFileConfig)
        {
            PostingFileConfig = postingFileConfig;
            StorageProfile = new StorageProfile(
                new FileInfo("../index/" + postingFileConfig.Name + "/dictionary.neting"),
                new FileInfo("../index/" + postingFileConfig.Name + "/scopeDirectory.neting"),
                new FileInfo("../index/" + postingFileConfig.Name + "/postingFile.neting"));

            if (Controller != null)
            {
                Controller.Shutdown();
            }

            Controller = new IndexController(StorageProfile, IndexSystemConfig, PostingFileConfig);
        }

        internal void ConfigureIndexSystem(
            ScopeDirectory scopeDirectory, IPostingFileConfig postingFileConfig)
        {
            PostingFileConfig = postingFileConfig;
            StorageProfile = new StorageProfile(
                new FileInfo("../index/" + postingFileConfig.Name + "/dictionary.neting"),
                new FileInfo("../index/" + postingFileConfig.Name + "/scopeDirectory.neting"),
                new FileInfo("../index/" + postingFileConfig.Name + "/postingFile.neting"));

            if (Controller != null)
            {
                Controller.Shutdown();
            }

            Controller = new IndexController(
                scopeDirectory, StorageProfile, IndexSystemConfig, PostingFileConfig);
        }

        public void Execute()
        {
            PrepareConsole();
            Console.WriteLine("Syntax: <command> [<argument(s)>]");
            ListCommands();
            while (running)
            {
                Step();
            }
        }

        protected abstract IEnumerable<IIndexConsoleCommand> GetCommands();

        private void PrepareConsole()
        {
            commands.Clear();
            foreach (var command in basicCommands.Concat(GetCommands()))
            {
                command.Setup(this);
                commands.Add(command);
                commandMapping.Add(command.Name.ToLowerInvariant(), command);
                int prefixLength = 1;
                while (keyMapping.ContainsKey(
                    command.Name.ToLowerInvariant().Substring(0, prefixLength)))
                {
                    prefixLength++;
                }

                keyMapping.Add(command.Name.ToLowerInvariant().Substring(0, prefixLength), command);
                command.KeyPrefix = prefixLength;
            }

            ConfigureIndexSystem(new RawPostingFileConfig());
        }

        private void Step()
        {
            Console.Write(">> ");
            string[] command = (Console.ReadLine() ?? "").Split(' ');
            IIndexConsoleCommand cmd;
            if (commandMapping.TryGetValue(command[0].ToLowerInvariant(), out cmd)
                || keyMapping.TryGetValue(command[0].ToLowerInvariant(), out cmd))
            {
                cmd.Execute(command.Length > 1 ? command[1] : null);
            }
            else
            {
                Console.WriteLine("Unknown command: " + command[0].ToLowerInvariant());
                Console.WriteLine("Existing commands are:");
                ListCommands();
            }
        }

        public void ListCommands()
        {
            Console.WriteLine(line);

            int maxKey = Enumerable.Select(commands, c => c.Name.Length).Max() + 4;
            int lineLength = line.Length - maxKey;
            string linePrefix = new string(Enumerable.Range(0, maxKey).Select(i => ' ').ToArray());

            foreach (var command in commands)
            {
                WriteCommand(command, linePrefix, lineLength, maxKey);
            }

            Console.WriteLine(line);
        }

        private const string line = "-------------------------------------------------------------------";

        private static void WriteCommand(IIndexConsoleCommand command, string linePrefix, int lineLength, int maxKey)
        {
            string spaces = new string(Enumerable.Range(0, maxKey - command.Name.Length - 3).Select(a => ' ').ToArray());
            Console.Write("(" + command.Name.Substring(0, command.KeyPrefix) + ")"
                          + command.Name.Substring(command.KeyPrefix)+ ":" + spaces);

            Console.WriteLine(command.ShortDesc.Substring(0, Math.Min(command.ShortDesc.Length, lineLength)));
            for (int i = lineLength; i < command.ShortDesc.Length; i += lineLength)
            {
                Console.WriteLine(linePrefix + command.ShortDesc.Substring(i, Math.Min(command.ShortDesc.Length - i, lineLength)));
            }
        }
    }
}
