﻿namespace NETing.ExperimentsConsole.Experiments.QueryExecutionExperiment.Commands
{
    using System;
    using System.Collections.Generic;
    
    using System.IO;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;

    using Abstract;
    using Abstract.QueryTrace;

    using Index.Controller.Services;
    using Index.QueryLanguage;
    using Index.Utils.Benchmark;
    using Index.Utils.Compression;

    internal abstract class AbstractQueryTraceCommand : AbstractConsoleCommand
    {
        internal sealed class TraceExecutionStatistic
        {
            public double TotalLatency { get; set; }
            public double TotalMillisecond { get; set; }
            public int ExecutedQueries { get; set; }

            public double Throughput
            {
                get { return ExecutedQueries / (TotalMillisecond / 1000); }
            }

            public double AvgLatency
            {
                get { return ((double)TotalLatency) / ExecutedQueries; }
            }
        }

        private QueryService currentQueryService;
        private int executedQueries;
        private long latency;

        protected Query[] LoadQueries(FileInfo traceFile)
        {
            IndexConsole.Controller.ScopeDirectoryService.Read();

            Query[] queries;
            using (var queryReader = new QueryTraceReader(
                IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory,
                new BinaryReader(traceFile.OpenRead())))
            {
                queries = queryReader.ToArray();
            }

            return queries;
        }

        protected TraceExecutionStatistic DoExecuteQueryTrace(
            IEnumerable<Query> queries,
            bool parallel,
            int concurrency,
            bool safety = false)
        {
            concurrency = Utils.Max(1, concurrency);

            latency = 0;
            executedQueries = 0;

            currentQueryService = IndexConsole.Controller.QueryService;
            
            for (int i = 0; i < 5; i++)
            {
                GC.Collect();
                GC.WaitForFullGCComplete();
            }

            long before = Timing.PerformanceCounter;

            if (parallel)
            {
                var action = safety ? (Action<Query>) SafeTimeExecuteQuery : TimeExecuteQuery;
                Parallel.ForEach(
                    queries,
                    new ParallelOptions {MaxDegreeOfParallelism = concurrency},
                    action);
            }
            else
            {
                if (safety)
                {
                    foreach (Query t in queries)
                    {
                        SafeTimeExecuteQuery(t);
                    }
                }
                else
                {
                    foreach (Query t in queries)
                    {
                        TimeExecuteQuery(t);
                    }
                }
            }

            long total = Timing.PerformanceCounter - before;
            double millis = (Timing.PerformanceCountToNanoSeconds(total) / 1000000);

            return new TraceExecutionStatistic
                        {
                            ExecutedQueries = executedQueries,
                            TotalLatency = (Timing.PerformanceCountToNanoSeconds(latency) / 1000000),
                            TotalMillisecond = millis
                        };
        }

        private void SafeTimeExecuteQuery(Query query)
        {
            try
            {
                TimeExecuteQuery(query);
            }
            catch (Exception e)
            {
                Console.WriteLine("Query failed: " + query.Id + " with message: " + e.Message);
                throw e;
            }
        }

        private void TimeExecuteQuery(Query query)
        {
            ExecuteQuery(query);
        }

        private void ExecuteQuery(Query query)
        {
            long before = Timing.PerformanceCounter;
            using (IEnumerator<object> result = currentQueryService.Query(query))
            {
                while (result.MoveNext())
                {
                }
            }

            long total = Timing.PerformanceCounter - before;
            Interlocked.Increment(ref executedQueries);
            Interlocked.Add(ref latency, total);
        }
    }
}