﻿namespace NETing.ExperimentsConsole.Experiments.QueryExecutionExperiment.Commands
{
    using System;
    using System.IO;
    using System.Text;
    using System.Linq;
    using Abstract.Commands;
    using Experiments;
    using Index.Controller.Configurations;
    using Index.QueryLanguage;
    using Utils;

    internal sealed class RunExperiment : AbstractQueryTraceCommand
    {
        public override string Name
        {
            get { return "RunExperiment.";  }
        }

        public override string ShortDesc
        {
            get { return "Will run a preconfigured experiment.";  }
        }

        private void ExecuteExperiments(ExperimentGroup experiments, DirectoryInfo outDir)
        {
            var subDir = new DirectoryInfo(outDir.FullName + "\\" + experiments.Name);
            if (!subDir.Exists)
            {
                subDir.Create();
            }

            foreach (IPostingFileConfig config in Configure.configs)
            {
                try
                {
                    Console.WriteLine("Configuring index for: " + config.Name);
                    IndexConsole.ConfigureIndexSystem(config);
                    var cfg = config;
                    foreach (IQueryExperiment experiment in experiments.Experiments.Where(e => e.Configs.Contains(cfg)))
                    {
                        ExecuteExperiment(experiment, config, subDir);
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine("Experiments wit config: " + config.Name + " failed: " + e.Message);
                }
            }
        }

        private void ExecuteExperiment(IQueryExperiment experiment, IPostingFileConfig config, DirectoryInfo outDir)
        {
            Console.WriteLine("Starting experiments: " + experiment.Name);
            var subDir = new DirectoryInfo(outDir.FullName + "\\" + experiment.Name);
            if (!subDir.Exists)
            {
                subDir.Create();
            }

            DoExecuteExperiment(experiment, config, subDir);
        }
        
        private void DoExecuteExperiment(
            IQueryExperiment experiment,
            IPostingFileConfig config,
            DirectoryInfo experimentDir)
        {
            var experimentLog = new FileInfo(experimentDir.FullName + "\\experimentLog." + experiment.Name + ".log");
            var queries = experiment.Traces.Aggregate(Enumerable.Empty<Query>(), (current, trace) => current.Concat(LoadQueries(trace))).ToArray();
            queries = experiment.Filter(queries).ToArray();

            using (TextWriter logWriter = new StreamWriter(experimentLog.OpenWrite()))
            {
                var logLineBuilder = new StringBuilder(1024);
                var teededLogWriter = new Tee(Console.Out, logWriter);
                teededLogWriter.WriteLine("Starting experiment: " + experiment.Name);

                var configResultFile = new FileInfo(experimentDir.FullName + "\\" + experiment.Name + "." + config.Name + ".timing");
                teededLogWriter.WriteLine("Results will be in: " + configResultFile);
                    
                if (experiment.WarmUp > 0)
                {
                    teededLogWriter.WriteLine("Running warm up rounds.");
                    for (int i = 0; i < experiment.WarmUp; i++)
                    {
                        DoExecuteQueryTrace(queries, experiment.Concurrent, experiment.ConcurrencyLevel);
                    }

                    teededLogWriter.WriteLine("Warm up rounds done.");
                }

                teededLogWriter.WriteLine("Starting experiments for config: " + config.Name + ", experiment: " + experiment.Name);
                using (TextWriter resultWriter = new StreamWriter(configResultFile.OpenWrite()))
                {
                    var teededResultWriter = new Tee(teededLogWriter, resultWriter);

                    for (int i = 0; i < experiment.Runs; i++)
                    {
                        logLineBuilder.Clear();
                        var timing = DoExecuteQueryTrace(queries, experiment.Concurrent, experiment.ConcurrencyLevel);
                        logLineBuilder
                            .Append(timing.ExecutedQueries).Append('\t')
                            .Append(timing.TotalMillisecond).Append('\t')
                            .Append(timing.TotalLatency).Append('\t')
                            .Append(timing.Throughput);

                        teededResultWriter.WriteLine(logLineBuilder.ToString());
                    }
                }

                teededLogWriter.WriteLine("Finished experiments for config: " + config.Name + "experiment: " + experiment.Name);
            }
        }

        private static readonly ExperimentGroup[] experiments = new ExperimentGroup[]
        {
            new ExperimentGroup
                {
                    Name = "ThesisExperiments",
                    Experiments = new IQueryExperiment[]
                                      {
                                          /*
                                          new HistogramIntersectionExperiment("max11res") { Runs = 5 }, // 1
                                          new HistogramIntersectionExperiment("max21res") { Runs = 5 },
                                          new HistogramIntersectionExperiment("max31res") { Runs = 5 },
                                          new HistogramIntersectionExperiment("max41res") { Runs = 10 },
                                          new HistogramIntersectionExperiment("max51res") { Runs = 20 },
                                          
                                          new HistogramPathExperiment("sentence", "max11res") { Runs = 5 }, // 2
                                          new HistogramPathExperiment("sentence", "max21res") { Runs = 5 },
                                          new HistogramPathExperiment("sentence", "max31res") { Runs = 5 },
                                          new HistogramPathExperiment("sentence", "max41res") { Runs = 10 },
                                          new HistogramPathExperiment("sentence", "max51res") { Runs = 20 },
                                          */

                                          /*
                                          new ThreadedExperiment(1), // 3
                                          new ThreadedExperiment(2), // 4
                                          new ThreadedExperiment(4), // 5
                                          new ThreadedExperiment(8), // 6
                                          */
                                          
                                          new PathExperiment("person", false), // 11
                                          new PathExperiment("person", true), // 12
                                          new PathExperiment("company", false), // 13
                                          new PathExperiment("company", true), // 14
                                          new PathExperiment("sentence", false), // 7
                                          new PathExperiment("sentence", true), // 8
                                          new PathExperiment("title", false), // 9
                                          new PathExperiment("title", true), // 10
                                          
                                      }
                },  
 
            new ExperimentGroup
                {
                    Name = "VBtests",
                    Experiments = new IQueryExperiment[]
                        {
                            new HistogramIntersectionExperiment("max11res") { Runs = 5 }, // 1
                            new HistogramIntersectionExperiment("max21res") { Runs = 5 },
                            new HistogramIntersectionExperiment("max31res") { Runs = 5 },
                            new HistogramIntersectionExperiment("max41res") { Runs = 10 },
                            new HistogramIntersectionExperiment("max51res") { Runs = 20 },    
                        }
                },

            new ExperimentGroup
                {
                    Name = "ProfileIntersection",
                    Experiments = new IQueryExperiment[] { new IntersectionExperiment10() }
                }, 
        };

        public override void Execute(string arg)
        {
            int index = 0;
            foreach (ExperimentGroup experiment in experiments)
            {
                Console.WriteLine(index++ + ": " + experiment.Name);
            }

            int selected = GetIntegerInput("Select experiment group", 0);
            DirectoryInfo output = GetDirectoryInput("Input path to result driectory.");
            ExecuteExperiments(experiments[selected], output);
        }
    }
}
