﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.IO;

    using Abstract;
    using Abstract.QueryTrace;
    using Abstract.Trec;

    using Index.Controller.Services;
    using Index.Documents.Parser;
    using Index.PathMatching;
    using Index.QueryLanguage;
    using Index.Utils.Compression;

    internal abstract class AbstractQueryTraceCommand : AbstractConsoleCommand
    {
        private sealed class ResultWriter : IDisposable
        {
            private readonly FileInfo outputFile;
            private readonly QueryTraceWriter[] outputWriters;

            private readonly int minResults;
            private readonly int bucketSize;
            private readonly int buckets;

            public ResultWriter(int bucketSize, int buckets, int minResults, FileInfo outputFile)
            {
                this.outputFile = outputFile;
                this.bucketSize = bucketSize;
                this.buckets = buckets;
                this.minResults = minResults;
                outputWriters = new QueryTraceWriter[buckets + 1];

                SetupWriters();
            }

            private void SetupWriters()
            {
                int min = minResults;
                for (int i = 0; i < buckets; i++)
                {
                    var file = new FileInfo(outputFile.FullName + ".min" + minResults + "res.max" + (min + bucketSize) + "res");
                    outputWriters[i] = new QueryTraceWriter(new BinaryWriter(file.OpenWrite()));
                    min += bucketSize;
                }

                var owerflow = new FileInfo(outputFile.FullName + ".min" + minResults + "res.max" + (int.MaxValue) + "res");
                outputWriters[buckets] = new QueryTraceWriter(new BinaryWriter(owerflow.OpenWrite()));
            }

            public void Add(Query query, int results)
            {
                int bucket = Utils.Min(results / bucketSize, buckets);
                outputWriters[bucket].AddQuery(query);
            }
            
            public void Dispose()
            {
                foreach (QueryTraceWriter queryTraceWriter in outputWriters)
                {
                    queryTraceWriter.Dispose();
                }
            }
        }

        private static bool IsWithinThreshold(int lower, int value, int upper)
        {
            return lower <= value && value <= upper;
        }

        protected void DoCompileQueryTrace(
            FileInfo trecLog,
            FileInfo outputTraceFile,
            string pathExpression,
            int termCountLowerThreshold,
            int termCountUpperThreshold,
            int resultsLowerThreshold,
            int histogramBuckets,
            int histogramBucketSize,
            TextWriter logOutput = null)
        {
            if (logOutput == null)
            {
                logOutput = Console.Out;
            }

            using (var trec = new TrecQueryLogReader(trecLog.OpenText()))
            {
                using (var output = new ResultWriter(histogramBucketSize, histogramBuckets, resultsLowerThreshold, outputTraceFile))
                {
                    QueryService service = IndexConsole.Controller.PrepareQuery();
                    var pathMatcher = new PathMatcher(IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory);
                    foreach (Tuple<int, Term[]> trecQuery in trec)
                    {
                        if (!IsWithinThreshold(termCountLowerThreshold, trecQuery.Item2.Length, termCountUpperThreshold))
                        {
                            continue;
                        }

                        Query query;
                        if (pathExpression == null)
                        {
                            query = new Query(trecQuery.Item2);
                        }
                        else
                        {
                            query = new Query(trecQuery.Item2, pathMatcher.MatchingPaths(pathExpression));
                        }

                        int results = CountResults(service, query);
                        if (IsWithinThreshold(resultsLowerThreshold, results, int.MaxValue))
                        {
                            output.Add(query, results);
                            logOutput.WriteLine("Query: " + trecQuery.Item1 + " accepted with " + results + " hits.");
                        }
                    }
                }
            }
        }

        private static int CountResults(QueryService service, Query query)
        {
            int cnt = 0;
            using (IEnumerator<object> enumerator = service.Query(query))
            {
                while (enumerator.MoveNext())
                {
                    cnt++;
                }
            }

            return cnt;
        }
    }
}