﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    
    using Abstract;
 
    using Index.Controller.Services;
    using Index.Documents.DocumentModel;
    using Index.Documents.DocumentModel.Scope;
    using Index.Documents.Parser;
    using Index.PathMatching;
    using Index.PostingFile;
    using Index.Utils.Collections;

    internal sealed class IndexStatistics : AbstractConsoleCommand
    {
        internal delegate void NewList();
        internal delegate void NewDocument(uint docId);
        internal delegate void NewPosting(Posting posting);

        internal event NewList newList;
        internal event NewDocument newDoc;
        internal event NewPosting newPosting;

        private abstract class AbstractStatisticsCollector
        {
            protected AbstractStatisticsCollector(IndexStatistics command)
            {
                command.newList += NewList;
                command.newDoc += NewDocument;
                command.newPosting += NewPosting;
            }

            protected virtual void NewList()
            {
            }

            protected virtual void NewDocument(uint docId)
            {
            }

            protected virtual void NewPosting(Posting posting)
            {
            }

            internal virtual void WriteResults(TextWriter output)
            {
            }
        }

        private class PathTypeStatistics : AbstractStatisticsCollector
        {
            private readonly MutableBitArray mask = new MutableBitArray();
            private readonly string pathExpression;
            
            private long documents;
            private long lists;
            private bool hasMatches;

            private int matchingPaths;
            private int matchingOccs;

            public PathTypeStatistics(string pathExpression, IndexStatistics command) : base(command)
            {
                this.pathExpression = pathExpression;
                command.IndexConsole.Controller.ScopeDirectoryService.Read();
                var matcher = new PathMatcher(command.IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory);
                foreach (var pt in matcher.MatchingPaths(pathExpression))
                {
                    mask[pt.PathId] = true;
                    matchingPaths++;
                }
            }

            protected override void NewDocument(uint docId)
            {
                if (hasMatches)
                {
                    hasMatches = false;
                    documents++;
                }
            }

            protected override void NewPosting(Posting posting)
            {
                lists++;
                if (mask[posting.Scope.PathType.PathId])
                {
                    hasMatches = true;
                    matchingOccs++;
                }
            }

            internal override void WriteResults(TextWriter output)
            {
                output.WriteLine(pathExpression + " matcing paths: " + matchingPaths);
                output.WriteLine(pathExpression + " matcing occs: " + matchingOccs);
                output.WriteLine(pathExpression + " matcing docs: " + documents);
                output.WriteLine(pathExpression + " doc lists: " + lists);
            }
        }

        private class ListDocumentLengths : AbstractStatisticsCollector
        {
            private long totalLength;
            private long currentLength;
            private long lists;
            private long maxDf = long.MinValue;
            private long minDf = long.MaxValue;


            public ListDocumentLengths(IndexStatistics command) : base(command)
            {
            }

            protected override void NewList()
            {
                if (lists++ == 0)
                {
                    totalLength = 0;
                }
                else
                {
                    totalLength += currentLength;
                    maxDf = Math.Max(maxDf, currentLength);
                    minDf = Math.Min(minDf, currentLength);
                }

                currentLength = 0;
            }

            protected override void NewDocument(uint docId)
            {
                currentLength++;
            }

            internal override void WriteResults(TextWriter output)
            {
                totalLength += currentLength;
                output.WriteLine("Average DF:\t\t" + ((double)totalLength / lists));
                output.WriteLine("Max DF:\t\t" + (maxDf));
                output.WriteLine("Max DF:\t\t" + (minDf));
            }
        }
        
        private class Tokens : AbstractStatisticsCollector
        {
            private long tokens;

            public Tokens(IndexStatistics command) : base(command)
            {
            }

            protected override void NewPosting(Posting posting)
            {
                tokens++;
            }

            internal override void WriteResults(TextWriter output)
            {
                output.WriteLine("Total tokens:\t\t" + tokens);
            }
        }

        private class DistinctPathTypes : AbstractStatisticsCollector
        {
            private readonly int ptc;

            internal DistinctPathTypes(IndexStatistics indexStatistics) : base(indexStatistics)
            {
                indexStatistics.IndexConsole.Controller.ScopeDirectoryService.Read();
                ptc = indexStatistics.IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory.NumberOfPaths;
            }

            internal override void WriteResults(TextWriter output)
            {
                output.WriteLine("Number of path types: \t" + ptc);
            }
        }

        private class DeweyStats : AbstractStatisticsCollector
        {
            private readonly long[] histogram = new long[ushort.MaxValue];

            private int maxDepth = int.MinValue;
            private int minDepth = int.MaxValue;

            private int maxElement = int.MinValue;

            private long totalDepth;
            private long occs;

            internal DeweyStats(IndexStatistics command) : base(command)
            {
            }

            protected override void NewPosting(Posting posting)
            {
                occs++;
                Scope scope = posting.Scope;
                totalDepth += scope.PathType.DeweyDepth;
                maxDepth = Math.Max(maxDepth, scope.PathType.DeweyDepth);
                minDepth = Math.Min(minDepth, scope.PathType.DeweyDepth);
                
                for (int i = 0; i < scope.PathType.DeweyDepth; i++)
                {
                    ushort elem = scope.Dewey[i];
                    maxElement = Math.Max(maxElement, elem);
                    histogram[elem]++;
                }
            }

            internal override void WriteResults(TextWriter output)
            {
                output.WriteLine("Dewey max depth: \t" + maxDepth);
                output.WriteLine("Dewey min depth: \t" + minDepth);
                output.WriteLine("Dewey max element: \t" + maxElement);

                for (int i = 0; i < histogram.Length - 1; i++)
                {
                    histogram[i + 1] += histogram[i];
                }

                output.WriteLine("Fraction of dewey elements belove");
                double divideby = totalDepth;
                for (int i = 0; i < 256; i++)
                {
                    output.WriteLine(i + "\t" + (histogram[i] / divideby));
                }
            }
        }
        
        public override string Name
        {
            get { return "IndexStatistics"; }
        }

        public override string ShortDesc
        {
            get { return "Will dump statistics regarding the data in the index."; }
        }

        public override void Execute(string arg)
        {
            var collectors = new AbstractStatisticsCollector[]
                             {
                                 new Tokens(this),
                                 new DistinctPathTypes(this),
                                 new DeweyStats(this),
                                 new ListDocumentLengths(this),
                                 new PathTypeStatistics("paragraph", this),
                                 new PathTypeStatistics("sentence", this),
                                 new PathTypeStatistics("title", this),
                                 new PathTypeStatistics("person", this),
                                 new PathTypeStatistics("company", this),
                             };

            using (QueryService service = IndexConsole.Controller.PrepareQuery())
            {
                foreach (Term term in IndexConsole.Controller.DictionaryService.GetAllTerms())
                {
                    newList();
                    uint prevDoc = uint.MaxValue;
                    using (PostingFileEnumerator enumerator = service.Lookup(term))
                    {
                        while (enumerator.MoveNext())
                        {
                            Posting posting = enumerator.Current;
                            if (posting.DocumentId != prevDoc)
                            {
                                newDoc(posting.DocumentId);
                            }

                            newPosting(posting);
                            prevDoc = posting.DocumentId;
                        }
                    }
                }
            }

            WriteResults(collectors);
            newList = null;
            newDoc = null;
            newPosting = null;
        }

        private void WriteResults(IEnumerable<AbstractStatisticsCollector> collectors)
        {
            foreach (AbstractStatisticsCollector abstractStatisticsCollector in collectors)
            {
                abstractStatisticsCollector.WriteResults(Console.Out);
            }
        }
    }
}
