﻿namespace NETing.ExperimentsConsole.Experiments.UtilityConsole.Commands
{
    using System;
    using System.Collections.Generic;
    using System.IO;

    using Abstract;
    using Abstract.QueryTrace;
    
    using Index.Documents.DocumentModel.Scope;
    using Index.QueryLanguage;
    
    using System.Linq;

    internal sealed class MergeQueryTraces : AbstractConsoleCommand
    {
        private enum MergeModes
        {
            Randomize,
            WeightedRandom,
            Sequential
        }

        public override string Name
        {
            get { return "MergeTraces"; }
        }

        public override string ShortDesc
        {
            get { return "Merges multiple query traces into one."; }
        }

        public override void Execute(string arg)
        {
            IndexConsole.Controller.ScopeDirectoryService.Read();

            IEnumerable<FileInfo> traceFiles = GetInputFiles("Enter path to query traces: ");
            FileInfo resultFile = GetFileInput("Enter path to resulting query trace: ", false);
            var mode = GetEnumSelection<MergeModes>("Select merge method: ");
            switch (mode)
            {
                case MergeModes.Sequential:
                    SequentialMerge(traceFiles, resultFile);
                    return;
                case MergeModes.Randomize:
                    RandomizedMerge(traceFiles, resultFile);
                    return;
                case MergeModes.WeightedRandom:
                    WeightedRandom(traceFiles, resultFile);
                    return;
            }
        }

        private void WeightedRandom(IEnumerable<FileInfo> traceFiles, FileInfo resultFile)
        {
            var weights = new int[traceFiles.Count()];
            int pos = 0, sum = 0;
            foreach (var traceFile in traceFiles)
            {
                weights[pos++] = GetIntegerInput("Enter integer weight for file: " + traceFile.Name, 1);
                sum = weights[pos - 1];
            }

            int[] mapping = new int[sum];
            pos = 0;
            for (int i = 0; i < weights.Length; i++)
            {
                for (int j = 0; j < weights[i]; j++)
                {
                    mapping[pos++] = i;
                }
            }

            ScopeDirectory scopeDirectory = IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory;
            var rnd = new Random();
            using (var writer = new QueryTraceWriter(new BinaryWriter(resultFile.OpenWrite())))
            {
                int remaining = traceFiles.Count();
                int count = remaining;
                var enumerators = new IEnumerator<Query>[remaining];
                var readers = new QueryTraceReader[remaining];
                pos = 0;
                foreach (FileInfo traceFile in traceFiles)
                {
                    readers[pos] = new QueryTraceReader(scopeDirectory, new BinaryReader(traceFile.OpenRead()));
                    enumerators[pos] = readers[pos].GetEnumerator();
                    pos++;
                }

                while (remaining > 0)
                {
                    int selected = mapping[rnd.Next(sum)];

                    if (enumerators[selected].MoveNext())
                    {
                        writer.AddQuery(enumerators[selected].Current);
                    }
                    else
                    {
                        remaining--;
                        enumerators[selected].Dispose();
                        enumerators[selected] = enumerators[remaining];

                        sum -= weights[selected];
                        int start;
                        for (start = 0; start < sum && mapping[start] != selected; start++)
                        {
                        }

                        Array.Copy(mapping, start + weights[selected], mapping, start, weights[selected]);
                    }
                }

                foreach (var reader in readers)
                {
                    reader.Dispose();
                }
            }
        }

        private void RandomizedMerge(IEnumerable<FileInfo> traceFiles, FileInfo resultFile)
        {
            ScopeDirectory scopeDirectory = IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory;
            var rnd = new Random();
            using (var writer = new QueryTraceWriter(new BinaryWriter(resultFile.OpenWrite())))
            {
                int remaining  = traceFiles.Count();
                int count = remaining;
                var enumerators = new IEnumerator<Query>[remaining];
                var readers = new QueryTraceReader[remaining];
                int pos = 0;
                foreach (FileInfo traceFile in traceFiles)
                {
                    readers[pos] = new QueryTraceReader(scopeDirectory, new BinaryReader(traceFile.OpenRead()));
                    enumerators[pos] = readers[pos].GetEnumerator();
                    pos++;
                }

                while (remaining > 0)
                {
                    int selected = rnd.Next(remaining);
                    if (enumerators[selected].MoveNext())
                    {
                        writer.AddQuery(enumerators[selected].Current);
                    }
                    else
                    {
                        remaining--;
                        enumerators[selected].Dispose();
                        enumerators[selected] = enumerators[remaining];
                    }
                }

                foreach (var reader in readers)
                {
                    reader.Dispose();
                }
            }
        }

        private void SequentialMerge(IEnumerable<FileInfo> traceFiles, FileInfo resultFile)
        {
            using (var writer = new QueryTraceWriter(new BinaryWriter(resultFile.OpenWrite())))
            {
                foreach (FileInfo traceFile in traceFiles)
                {
                    using (var reader = new QueryTraceReader(
                        IndexConsole.Controller.ScopeDirectoryService.ScopeDirectory,
                        new BinaryReader(traceFile.OpenRead())))
                    {
                        foreach (Query query in reader)
                        {
                            writer.AddQuery(query);
                        }
                    }
                }
            }
        }
    }
}
