﻿namespace NETing.Index.Inversion
{
    using System.Collections.Generic;
    using System.IO;
    
    using Documents.DocumentModel.Scope;
    using Documents.Parser;
    
    using NUnit.Framework;
    
    [TestFixture]
    public class DocumentInverterTest
    {
        private ScopeDirectory ScopeDirectory { get; set; }

        private TermStorage TermStorage { get; set; }

        private DocumentParser Parser { get; set; }

        private DocumentInverter documentInverter;

        [TestFixtureSetUp]
        public void SetupFixture()
        {
            ScopeDirectory = new ScopeDirectory();
            TermStorage = new TermStorage();
            documentInverter = new DocumentInverter(TermStorage);
            Parser = new DocumentParser(ScopeDirectory);
        }

        private void AssertPosting(
            InvertedDocumentOccurrence occ, int[] positions, PathType[] paths, Dewey[] deweys)
        {
            for (int i = 0; i < occ.Frequency; i++)
            {
                Assert.AreEqual(positions[i], occ.Positions[i]);
                Assert.AreEqual(new Scope(paths[i], deweys[i]), occ.ScopeAt(i, ScopeDirectory));
            }
        }

        [Test]
        public void TestDocumentInversion()
        {
            /*
             * hello : <0, a[0]/b[0]/0>
             * hi    : <1, a[0]/1>, <8, a[0]/3>
             * you   : <2, a[0]/c[2]/0>, <6, a[0]/3>
             * all   : <3, a[0]/c[2]/d[1]/0>, <4, a[0]/c[2]/1>
             * are   : <5, a[0]/3>
             * and   : <7, a[0]/3>
             * */

            Stream document = new MemoryStream();
            TextWriter writer = new StreamWriter(document);
            writer.Write("<?xml version=\"1.0\" encoding=\"utf-8\"?><a><b>hello</b>hi<c>you<d>all</d>all</c>are you and hi</a>");
            writer.Flush();
            // writer.Close();
            document.Seek(0, SeekOrigin.Begin);
            var inverted = documentInverter.Invert(Parser.ParseDocument(document));
            var inversion = new Dictionary<Term, InvertedDocumentOccurrence>();
            while (inverted.MoveNext())
            {
                inversion.Add(inverted.Current.First, inverted.Current.Second.Duplicate());
            }

            AssertPosting(
                inversion[new Term("hello")], 
                new[] { 0 }, 
                new[] { ScopeDirectory.GetPath("a", "b") },
                new[] { new Dewey(new ushort[] { 0, 0 }) });

            AssertPosting(
                inversion[new Term("hi")],
                new[] { 1, 8 },
                new[] { ScopeDirectory.GetPath("a"), ScopeDirectory.GetPath("a") },
                new[] { new Dewey(new ushort[] { 1 }), new Dewey(new ushort[] { 3 }) });

            AssertPosting(
                inversion[new Term("you")],
                new[] { 2, 6 },
                new[] { ScopeDirectory.GetPath("a", "c"), ScopeDirectory.GetPath("a") },
                new[] { new Dewey(new ushort[] { 2, 0 }), new Dewey(new ushort[] { 3 }) });

            AssertPosting(
                inversion[new Term("all")],
                new[] { 3, 4 },
                new[] { ScopeDirectory.GetPath("a", "c", "d"), ScopeDirectory.GetPath("a", "c") },
                new[] { new Dewey(new ushort[] { 2, 1, 0 }), new Dewey(new ushort[] { 2, 2 }) });

            AssertPosting(
                inversion[new Term("are")],
                new[] { 5 },
                new[] { ScopeDirectory.GetPath("a") },
                new[] { new Dewey(new ushort[] { 3 }) });

            AssertPosting(
                inversion[new Term("and")],
                new[] { 7 },
                new[] { ScopeDirectory.GetPath("a") },
                new[] { new Dewey(new ushort[] { 3 }) });
        }
    }
}