﻿namespace NETing.Index.Inversion
{
    using System.Collections.Generic;
    using System.Linq;
    
    using Documents.DocumentModel.Scope;
    using Documents.Parser;
    
    using NUnit.Framework;
    
    using Utils.Collections;
    using Utils.Testing;

    [TestFixture]
    public class ScalingInMemoryInverterTest
    {
        private ScopeDirectory ScopeDirectory { get; set; }

        private DocumentParser Parser { get; set; }

        private ScalingInMemoryInverter Inverter { get; set; }

        private readonly DocumentInverter documentInverter = new DocumentInverter(new TermStorage());

        private readonly string[] documents = new[]
        {
            "<doc>hello you <a>people</a> next is this</doc>",
            "<doc>this is a text<a>with documents and stuff</a>next one<a><b>is</b></a></doc>",
            "<doc>no way this is it</doc>",
            "<doc><b>hello</b> this is <a>a<b>serious</b></a> atempth to create a document</doc>"
        };

        private readonly string[] terms = new[]
        {
            "a", "and", "atempth", "create", "document", "documents",
            "hello", "is", "it", "next", "no", "one", "people", "serious",
            "stuff", "text", "this", "to", "way", "with", "you"
        };

        private readonly uint[][][] positions = new[]
        {
            new[] { new uint[0], new uint[] { 2 }, new uint[0], new uint[] { 3, 8 } },
            new[] { new uint[0], new uint[] { 6 }, new uint[0], new uint[0] },
            new[] { new uint[0], new uint[0], new uint[0], new uint[] { 5 } },
            new[] { new uint[0], new uint[0], new uint[0], new uint[] { 7 } },
            new[] { new uint[0], new uint[0], new uint[0], new uint[] { 9 } },
            new[] { new uint[0], new uint[] { 5 }, new uint[0], new uint[0] },
            new[] { new uint[] { 0 }, new uint[0], new uint[0], new uint[] { 0 } },
            new[] { new uint[] { 4 }, new uint[] { 1, 10 }, new uint[] { 3 }, new uint[] { 2 } },
            new[] { new uint[0], new uint[0], new uint[] { 4 }, new uint[0] },
            new[] { new uint[] { 3 }, new uint[] { 8 }, new uint[0], new uint[0] },
            new[] { new uint[0], new uint[0], new uint[] { 0 }, new uint[0] },
            new[] { new uint[0], new uint[] { 9 }, new uint[0], new uint[0] },
            new[] { new uint[] { 2 }, new uint[0], new uint[0], new uint[0] },
            new[] { new uint[0], new uint[0], new uint[0], new uint[] { 4 } },
            new[] { new uint[0], new uint[] { 7 }, new uint[0], new uint[0] },
            new[] { new uint[0], new uint[] { 3 }, new uint[0], new uint[0] },
            new[] { new uint[] { 5 }, new uint[] { 0 }, new uint[] { 2 }, new uint[] { 1 } },
            new[] { new uint[0], new uint[0], new uint[0], new uint[] { 6 } },
            new[] { new uint[0], new uint[0], new uint[] { 1 }, new uint[0] },
            new[] { new uint[0], new uint[] { 4 }, new uint[0], new uint[0] },
            new[] { new uint[] { 1 }, new uint[0], new uint[0], new uint[0] },
        };

        private static readonly string[][] nilPath = new string[0][];
        private static readonly string[] doc = new[] { "doc" };
        private static readonly string[] docA = new[] { "doc", "a" };
        private static readonly string[] docAB = new[] { "doc", "a", "b" };
        private static readonly string[] docB = new[] { "doc", "b" };

        private readonly string[][][][] paths = new[]
        {
            new[] { nilPath, new[] { doc }, nilPath, new[] { docA, doc } },
            new[] { nilPath, new[] { docA }, nilPath, nilPath },
            new[] { nilPath, nilPath, nilPath, new[] { doc } },
            new[] { nilPath, nilPath, nilPath, new[] { doc } },
            new[] { nilPath, nilPath, nilPath, new[] { doc } },
            new[] { nilPath, new[] { docA }, nilPath, nilPath },
            new[] { new[] { doc }, nilPath, nilPath, new[] { docB } },
            new[] { new[] { doc }, new[] { doc, docAB }, new[] { doc }, new[] { doc } },
            new[] { nilPath, nilPath, new[] { doc }, nilPath },
            new[] { new[] { doc }, new[] { doc }, nilPath, nilPath },
            new[] { nilPath, nilPath, new[] { doc }, nilPath },
            new[] { nilPath, new[] { doc }, nilPath, nilPath },
            new[] { new[] { docA }, nilPath, nilPath, nilPath },
            new[] { nilPath, nilPath, nilPath, new[] { docAB } },
            new[] { nilPath, new[] { docA }, nilPath, nilPath },
            new[] { nilPath, new[] { doc }, nilPath, nilPath },
            new[] { new[] { doc }, new[] { doc }, new[] { doc }, new[] { doc } },
            new[] { nilPath, nilPath, nilPath, new[] { doc } },
            new[] { nilPath, nilPath, new[] { doc }, nilPath },
            new[] { nilPath, new[] { docA }, nilPath, nilPath },
            new[] { new[] { doc }, nilPath, nilPath, nilPath },
        };

        private static ushort[][] nilDewey = new ushort[0][];
        private static ushort[] docDewey = new ushort[1];

        private readonly ushort[][][][] deweys = new[]
        {
            new[] { nilDewey, new[] { docDewey }, nilDewey, new[] { new ushort[] { 0, 1 }, docDewey } },
            new[] { nilDewey, new[] { new ushort[] { 0, 0 } }, nilDewey, nilDewey },
            new[] { nilDewey, nilDewey, nilDewey, new[] { docDewey } },
            new[] { nilDewey, nilDewey, nilDewey, new[] { docDewey } },
            new[] { nilDewey, nilDewey, nilDewey, new[] { docDewey } },
            new[] { nilDewey, new[] { new ushort[] { 0, 0 } }, nilDewey, nilDewey },
            new[] { new[] { docDewey }, nilDewey, nilDewey, new[] { new ushort[] { 0, 0 } } },
            new[] { new[] { docDewey }, new[] { docDewey, new ushort[] { 0, 1, 0 } }, new[] { docDewey }, new[] { docDewey } },
            new[] { nilDewey, nilDewey, new[] { docDewey }, nilDewey },
            new[] { new[] { docDewey }, new[] { docDewey }, nilDewey, nilDewey },
            new[] { nilDewey, nilDewey, new[] { docDewey }, nilDewey },
            new[] { nilDewey, new[] { docDewey }, nilDewey, nilDewey },
            new[] { new[] { new ushort[] { 0, 0 } }, nilDewey, nilDewey, nilDewey },
            new[] { nilDewey, nilDewey, nilDewey, new[] { new ushort[] { 0, 1, 0 } } },
            new[] { nilDewey, new[] { new ushort[] { 0, 0 } }, nilDewey, nilDewey },
            new[] { nilDewey, new[] { docDewey }, nilDewey, nilDewey },
            new[] { new[] { docDewey }, new[] { docDewey }, new[] { docDewey }, new[] { docDewey } },
            new[] { nilDewey, nilDewey, nilDewey, new[] { docDewey } },
            new[] { nilDewey, nilDewey, new[] { docDewey }, nilDewey },
            new[] { nilDewey, new[] { new ushort[] { 0, 0 } }, nilDewey, nilDewey },
            new[] { new[] { docDewey }, nilDewey, nilDewey, nilDewey }
        };

        [TestFixtureSetUp]
        public void SetupFixture()
        {
            ScopeDirectory = new ScopeDirectory();

            Parser = new DocumentParser(ScopeDirectory);
            Inverter = new ScalingInMemoryInverter(1024, 24, ScopeDirectory);
        }

        private void AssertTermPostings(int termIndex, Term term, IEnumerator<InvertedDocumentOccurrence> posting)
        {
            Assert.AreEqual(terms[termIndex], term.ToString());
            int occurences = 0;
            while (posting.MoveNext())
            {
                occurences++;
                AssertPositions(
                    positions[termIndex][(long)posting.Current.DocumentId - 1],
                    posting.Current.Positions.Take((int)posting.Current.Frequency));
                AssertScope(
                    paths[termIndex][(long)posting.Current.DocumentId - 1],
                    deweys[termIndex][(long)posting.Current.DocumentId - 1],
                    Enumerable.Range(0, posting.Current.Frequency).Select(i => posting.Current.ScopeAt(i, ScopeDirectory)));
            }

            int count = 0;
            foreach (var post in positions[termIndex])
            {
                if (post.Length > 0)
                {
                    count++;   
                }
            }

            Assert.AreEqual(count, occurences);
        }

        private static void AssertPositions(IEnumerable<uint> expected, IEnumerable<ushort> actual)
        {
            Assert.That(expected.SequenceEqual(actual.Cast<uint>()));
        }

        private void AssertScope(
            IEnumerable<string[]> pathTypes, IEnumerable<ushort[]> deweys, IEnumerable<Scope> actual)
        {
            IEnumerator<string[]> pathTypeEnum = pathTypes.GetEnumerator();
            IEnumerator<ushort[]> deweyEnum = deweys.GetEnumerator();
            IEnumerator<Scope> actualEnum = actual.GetEnumerator();

            while (pathTypeEnum.MoveNext())
            {
                Assert.That(deweyEnum.MoveNext());
                Assert.That(actualEnum.MoveNext());
                Assert.AreEqual(
                    ScopeDirectory.GetPath(pathTypeEnum.Current), 
                    actualEnum.Current.PathType);
                Assert.AreEqual(
                    new Dewey(deweyEnum.Current), 
                    actualEnum.Current.Dewey);
            }

            Assert.IsFalse(actualEnum.MoveNext());
        }

        [Test]
        [Ignore("Broken by text node tagging.")]
        public void TestInMemoryInverter()
        {
            uint documentId = 0;
            foreach (string document in documents)
            {
                var parsed = Parser.ParseDocument(document.AsStream());
                var inverted = documentInverter.Invert(parsed);
                while (inverted.MoveNext())
                {
                    Inverter.AddDocumentOccurences(inverted.Current.First, ++documentId, inverted.Current.Second);
                }
            }

            int termIndex = 0;
            foreach (Pair<Term, IEnumerator<InvertedDocumentOccurrence>> pair in 
                Inverter.ReadInvertedIndex())
            {
                AssertTermPostings(termIndex++, pair.First, pair.Second);
            }
        }
    }
}
