﻿namespace NETing.Index.Utils.ByteArrays
{
    using System;
    using Benchmark;
    using NUnit.Framework;

    [TestFixture]
    public class TestByteArrayExtensions
    {
        private static int[] integers = new[]
                                            {
                                                int.MinValue, int.MaxValue, 0, 1, 128, 256, 1024, 111,
                                                16, 2, -16, -128, -3900, int.MaxValue - 1
                                            };

        [Test]
        public void TestSetGetValueInteger()
        {
            var buffer = new byte[8];
            foreach (int integer in integers)
            {
                buffer.SetValueAt(0, integer);
                buffer.SetValueAt(4, integer);
                
                int outHolder = buffer.GetInt32At(0);
                Assert.AreEqual(integer, outHolder);
                outHolder = buffer.GetInt32At(4);
                Assert.AreEqual(integer, outHolder);
            }
        }

        [Test]
        public void TestSetValueIntegerBitConverterCompatibility()
        {
            var buffer = new byte[8];
            foreach (int integer in integers)
            {
                buffer.SetValueAt(0, integer);
                buffer.SetValueAt(4, integer);
                Assert.AreEqual(integer, BitConverter.ToInt32(buffer, 0));
                Assert.AreEqual(integer, BitConverter.ToInt32(buffer, 4));
            }
        }

        [Test]
        [Category("Performance-tests")]
        public void TestByteArrayExtensionsPerformance()
        {
            const int Iterations = 10000000;

            long before = Timing.PerformanceCounter;
            var buffer = new byte[4]; 
            for (int i = 0; i < Iterations; i++)
            {
                buffer.SetValueAt(0, i);
            }

            long after = Timing.PerformanceCounter;
            long extensionsTime = after - before;

            before = Timing.PerformanceCounter;
            for (int i = 0; i < Iterations; i++)
            {
                byte[] bytes = BitConverter.GetBytes(i);
                Buffer.BlockCopy(bytes, 0, buffer, 0, 4);
            }

            after = Timing.PerformanceCounter;
            long bitConverterTime = after - before;

            before = Timing.PerformanceCounter;
            var holder = new int[1];
            for (int i = 0; i < Iterations; i++)
            {
                holder[0] = i;
                Buffer.BlockCopy(holder, 0, buffer, 0, 4);
            }

            after = Timing.PerformanceCounter;
            long blockCopyTime = after - before;

            Console.WriteLine("----------------------------------------");
            Console.WriteLine("Storing values into byte arrays:");
            Console.WriteLine("----------------------------------------");
            Console.WriteLine(
                "ByteArrayUtils nanoseconds per Int32: {0:0.00}", 
                Timing.PerformanceCountToNanoSeconds(extensionsTime) / Iterations);
            Console.WriteLine(
                "BitConverter nanoseconds per Int32: {0:0.00}",
                Timing.PerformanceCountToNanoSeconds(bitConverterTime) / Iterations);
            Console.WriteLine(
                "Buffer.BlockCopy nanoseconds per Int32: {0:0.00}",
                Timing.PerformanceCountToNanoSeconds(blockCopyTime) / Iterations);

            before = Timing.PerformanceCounter;
            buffer = new byte[4];
            int output;
            for (int i = 0; i < Iterations; i++)
            {
                output = buffer.GetInt32At(0);
            }

            after = Timing.PerformanceCounter;
            extensionsTime = after - before;

            before = Timing.PerformanceCounter;
            for (int i = 0; i < Iterations; i++)
            {
                output = BitConverter.ToInt32(buffer, 0);
            }

            after = Timing.PerformanceCounter;
            bitConverterTime = after - before;

            before = Timing.PerformanceCounter;
            holder = new int[1];
            for (int i = 0; i < Iterations; i++)
            {
                Buffer.BlockCopy(buffer, 0, holder, 0, 4);
            }

            after = Timing.PerformanceCounter;
            blockCopyTime = after - before;

            Console.WriteLine("----------------------------------------");
            Console.WriteLine("Retrieving values from byte arrays:");
            Console.WriteLine("----------------------------------------");
            Console.WriteLine(
                "ByteArrayUtils nanoseconds per Int32: {0:0.00}",
                Timing.PerformanceCountToNanoSeconds(extensionsTime) / Iterations);
            Console.WriteLine(
                "BitConverter nanoseconds per Int32: {0:0.00}",
                Timing.PerformanceCountToNanoSeconds(bitConverterTime) / Iterations);
            Console.WriteLine(
                "Buffer.BlockCopy nanoseconds per Int32: {0:0.00}",
                Timing.PerformanceCountToNanoSeconds(blockCopyTime) / Iterations);
        }
    }
}


