﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;
    using Documents.DocumentModel.Scope;
 
    using PostingFile;
    
    using Readers;

    using Utils.Collections;
    using Utils.Compression;

    public sealed class CNPforScopeColumnReader : AbstractScopeColumnReader
    {
        private readonly int padLimit;

        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private readonly NPforDecodingWrapper wrapper;
            private byte[] dataBuffer = new byte[1024];

            private int pathTypeOffset;
            private int currentDeweyOffset;

            private int positionInPathTypeBuffer;

            private readonly uint[] pathTypeBuffer = new uint[128];
            private readonly uint[] deweyBuffer = new uint[16 * 128];
            private readonly uint[] columns = new uint[16];

            private readonly ushort[] deweyArray = new ushort[32];
            private readonly Dewey deweyHolder;

            public ChunkEnumerator(ScopeDirectory scopeDirectory, NPforDecodingWrapper wrapper) 
                : base(scopeDirectory)
            {
                this.wrapper = wrapper;

                deweyHolder = new Dewey(deweyArray);
                holder.Dewey = deweyHolder;
            }

            public override bool MoveNext()
            {
                CurrentOccurrenceId++;
                if (CurrentOccurrenceId >= TotalOccurrences)
                {
                    return false;
                }

                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                }

                PathType pathType = pathTypes[pathTypeBuffer[++positionInPathTypeBuffer]];
                holder.PathType = pathType;

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                LoadDewey(pathType.DeweyDepth);
                return true;
            }

            public override void MoveTo(int targetPosition)
            {
                int remainingInChunk = 128 - positionInPathTypeBuffer - 1;
                int skipDistance = targetPosition - CurrentOccurrenceId - 1;

                if (skipDistance >= remainingInChunk)
                {
                    if (positionInPathTypeBuffer > -1)
                    {
                        CurrentOccurrenceId += remainingInChunk;
                        skipDistance -= remainingInChunk;
                        positionInPathTypeBuffer = -1;
                    }

                    int skipChunks = skipDistance >> 7;
                    if (skipChunks > 0)
                    {
                        SkipChunks(skipChunks);
                        CurrentOccurrenceId += (skipChunks << 7);
                    }
                }

                CurrentOccurrenceId++;
                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                    positionInPathTypeBuffer = -1;
                }

                CurrentOccurrenceId = targetPosition;
                int stopPositionInPathTypeBuffer = targetPosition & 127;
                int[] endedAt = new int[16];
                int maxDepth = 0, passed = 0;
                while(++positionInPathTypeBuffer < stopPositionInPathTypeBuffer)
                {
                    int depth = pathTypes[pathTypeBuffer[positionInPathTypeBuffer]].DeweyDepth;
                    maxDepth = Utils.Max(maxDepth, depth);
                    passed++;
                    endedAt[depth]++;
                }

                endedAt[0] = passed;
                for (int i = 1; i < maxDepth; i++)
                {
                    endedAt[i] = endedAt[i - 1] - endedAt[i];
                    columns[i] += (uint)endedAt[i];
                }

                holder.PathType = pathTypes[pathTypeBuffer[positionInPathTypeBuffer]];
                LoadDewey(holder.PathType.DeweyDepth);

                if (positionInPathTypeBuffer > 126)
                {
                    positionInPathTypeBuffer = -1;
                }
            }

            public override int SkipToPathType(MutableBitArray candidatePathTypes)
            {
                PathType pathType = holder.PathType;
                int switchedPathTypes = 0, startAt = positionInPathTypeBuffer;
                if (pathType != null && candidatePathTypes[pathType.PathId])
                {
                    return CurrentOccurrenceId;
                }

                if (pathType == null)
                {
                    if (!MoveNext())
                    {
                        return CurrentOccurrenceId;
                    }

                    return SkipToPathType(candidatePathTypes);
                }

                uint pathId = (uint)pathType.PathId;
                while (!candidatePathTypes[pathId])
                {
                    CurrentOccurrenceId++;
                    if (CurrentOccurrenceId >= TotalOccurrences)
                    {
                        return CurrentOccurrenceId;
                    }

                    if (positionInPathTypeBuffer < 0)
                    {
                        DecodePathTypeChunk();
                        switchedPathTypes++;
                        startAt = -1;
                    }

                    pathId = pathTypeBuffer[++positionInPathTypeBuffer];
                    
                    if (positionInPathTypeBuffer >= 127)
                    {
                        positionInPathTypeBuffer = -1;
                    }
                }

                if (switchedPathTypes > 0)
                {
                    if (switchedPathTypes > 1)
                    {
                        SkipDeweyChunks(switchedPathTypes - 1);
                    }

                    DecodeDeweyChunk();
                }

                holder.PathType = pathTypes[pathId];
                var deweySkip = new int[16];
                int limit = positionInPathTypeBuffer != -1 ? positionInPathTypeBuffer : 127;
                for (int i = startAt; i < limit; )
                {
                    pathType = pathTypes[pathTypeBuffer[++i]];
                    deweySkip[pathType.DeweyDepth]++;
                }

                deweySkip[pathType.DeweyDepth]--;
                deweySkip[0] = limit - startAt - 1;
                for (int i = 1; i < 16; i++)
                {
                    deweySkip[i] = deweySkip[i - 1] - deweySkip[i];
                    columns[i] += (uint)deweySkip[i];
                }

                
                LoadDewey(holder.PathType.DeweyDepth);

                if (positionInPathTypeBuffer > 126)
                {
                    positionInPathTypeBuffer = -1;
                }

                return CurrentOccurrenceId;
            }

            private void LoadDewey(int depth)
            {
                holder.Dewey.Count = depth;
                for (int i = 0; i < depth; i++)
                {
                    deweyArray[i] = (ushort)deweyBuffer[columns[i]++];
                }
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                uint deweyOffset;
                offset = PrefixedVByte.Decode(buffer, offset, out deweyOffset);
                currentDeweyOffset = (int)deweyOffset;
                return offset;
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                SetupChunk(occurrences);
                Reset();

                return this;
            }

            public override void Reset()
            {
                base.Reset();
                pathTypeOffset = 0;
                positionInPathTypeBuffer = -1;
                Array.Clear(deweyArray, 0, 16);
            }

            private void DecodeChunk()
            {
                DecodePathTypeChunk();
                DecodeDeweyChunk();

                for (int i = 0; i < 16; i++)
                {
                    deweyArray[i] = 0;
                }
            }

            private void DecodeDeweyChunk()
            {
                uint groupMax, fullPrefix;
                currentDeweyOffset = PrefixedVByte.Decode(
                    dataBuffer, currentDeweyOffset, out groupMax);
                currentDeweyOffset = PrefixedVByte.Decode(
                    dataBuffer, currentDeweyOffset, out fullPrefix);

                for (int i = (int)fullPrefix; i < groupMax; i++)
                {
                    currentDeweyOffset = PrefixedVByte.Decode(
                        dataBuffer, currentDeweyOffset, out columns[i]);
                }

                for (int i = 0; i < fullPrefix; i++)
                {
                    columns[i] = (uint)i << 7;
                }

                currentDeweyOffset = wrapper.DecodeMulti(
                    dataBuffer, currentDeweyOffset, deweyBuffer, 0, (int)fullPrefix);

                for (int i = (int)fullPrefix; i < groupMax; i++)
                {
                    currentDeweyOffset = wrapper.DecodeSingle(
                        dataBuffer, currentDeweyOffset, (int)columns[i], deweyBuffer, i << 7);
                    columns[i] = (uint)i << 7;
                }
            }

            private void DecodePathTypeChunk()
            {
                pathTypeOffset = wrapper.DecodeSingle(
                    dataBuffer, pathTypeOffset, TotalOccurrences - CurrentOccurrenceId, pathTypeBuffer, 0);
                positionInPathTypeBuffer = -1;
            }

            private void SkipChunks(int chunks)
            {
                pathTypeOffset = wrapper.SkipMultiple(
                    dataBuffer, pathTypeOffset, chunks, TotalOccurrences - CurrentOccurrenceId - 1);
                SkipDeweyChunks(chunks);
            }

            private void SkipDeweyChunks(int chunks)
            {
                for (int i = 0; i < chunks; i++)
                {
                    uint groupMax, fullPrefix;
                    currentDeweyOffset = PrefixedVByte.Decode(
                        dataBuffer, currentDeweyOffset, out groupMax);
                    currentDeweyOffset = PrefixedVByte.Decode(
                        dataBuffer, currentDeweyOffset, out fullPrefix);

                    for (int j = 0; j < fullPrefix; j++)
                    {
                        columns[j] = 128;
                    }

                    uint skipPrefix = fullPrefix;
                    for (uint j = fullPrefix; j < groupMax; j++)
                    {
                        currentDeweyOffset = PrefixedVByte.Decode(
                            dataBuffer, currentDeweyOffset, out columns[j]);

                        if (columns[j] >= wrapper.PadLimit)
                        {
                            skipPrefix++;
                        }
                    }

                    currentDeweyOffset = wrapper.SkipMultiple(
                        dataBuffer, currentDeweyOffset, (int)skipPrefix, 128 * (int)skipPrefix);
                    for (uint j = skipPrefix; j < groupMax; j++)
                    {
                        currentDeweyOffset = wrapper.SkipSingle(dataBuffer, currentDeweyOffset, (int)columns[j]);
                    }
                }
            }
        }
        
        public CNPforScopeColumnReader(int padLimit = NPforSettings.PadLimit)
        {
            this.padLimit = padLimit;
        }
        
        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory, new NPforDecodingWrapper(padLimit));
        }
    }
}
