﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;
    using Documents.DocumentModel.Scope;

    using PostingFile;

    using Readers;

    using Utils.Compression;

    public sealed class PfxCNPforScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private readonly NPforDecodingWrapper wrapper;
            private byte[] dataBuffer = new byte[1024];

            private int pathTypeOffset;
            
            private uint maxDepth;
            private readonly uint[] offsets = new uint[16];
            private readonly uint[] columnCounts = new uint[16];
            private int currentPrefixOffset;

            private int positionInPathTypeBuffer;

            private readonly uint[] pathTypeBuffer = new uint[128];
            private readonly uint[] prefixBuffer = new uint[128];

            private readonly int[] positions = new int[16];
            private readonly uint[] chunkRemaining = new uint[16];
            private readonly uint[][] deweyColumnBuffer = new uint[16][];
            private readonly int[] prefixWaterFront = new int[17];
            private readonly int[] depthWaterFront = new int[17];
            private readonly int[] columnMoves = new int[17];

            private readonly ushort[] deweyArray = new ushort[32];
            private readonly Dewey deweyHolder;

            public ChunkEnumerator(ScopeDirectory scopeDirectory, NPforDecodingWrapper wrapper) 
                : base(scopeDirectory)
            {
                this.wrapper = wrapper;
                for (int i = 0; i < deweyColumnBuffer.Length; i++)
                {
                    deweyColumnBuffer[i] = new uint[128];
                }

                deweyHolder = new Dewey(deweyArray);
                holder.Dewey = deweyHolder;
            }

            public override bool MoveNext()
            {
                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                }

                CurrentOccurrenceId++;

                PathType pathType = pathTypes[pathTypeBuffer[++positionInPathTypeBuffer]];
                uint prefixLength = prefixBuffer[positionInPathTypeBuffer];

                holder.PathType = pathType;

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                if (CurrentOccurrenceId < TotalOccurrences)
                {
                    LoadDewey(pathType.DeweyDepth, prefixLength);
                    return true;
                }

                return false;
            }

            public override void MoveTo(int targetPosition)
            {
                int remainingInChunk = 128 - positionInPathTypeBuffer - 1;
                int skipDistance = targetPosition - CurrentOccurrenceId - 1;

                if (skipDistance >= remainingInChunk)
                {
                    Array.Clear(columnMoves, 0, (int)maxDepth);

                    if (positionInPathTypeBuffer > -1)
                    {
                        CurrentOccurrenceId += remainingInChunk;
                        skipDistance -= remainingInChunk;
                        positionInPathTypeBuffer = -1;

                        for (int i = 0; i < maxDepth; i++)
                        {
                            columnMoves[i] = (int)chunkRemaining[i];
                        }                     
                    }
                    
                    int skipChunks = skipDistance >> 7;
                    if (skipChunks > 0)
                    {
                        SkipChunks(skipChunks, columnMoves);
                        CurrentOccurrenceId += (skipChunks << 7);
                    }

                    for (int i = 0; i < maxDepth; i++)
                    {
                        if (columnMoves[i] >= 1)
                        {
                            if (positions[i] > -1)
                            {
                                int remainingInCurrent = (128 - positions[i]);
                                if (columnMoves[i] > remainingInCurrent)
                                {
                                    columnMoves[i] -= remainingInCurrent;
                                    positions[i] = -1;
                                }
                            }

                            if (columnMoves[i] > 128)
                            {
                                skipChunks = columnMoves[i] >> 7;
                                int skipped = skipChunks << 7;
                                offsets[i] = (uint)wrapper.SkipMultiple(
                                    dataBuffer, (int) offsets[i], skipChunks, skipped);
                                columnCounts[i] -= (uint) skipped;
                                columnMoves[i] -= skipped;
                                positions[i] = -1;
                            }

                            if (columnMoves[i] > 0)
                            {
                                if (positions[i] < 0)
                                {
                                    LoadColumn(i);
                                }

                                positions[i] += columnMoves[i];

                                if (positions[i] > 127)
                                {
                                    positions[i] = -1;
                                }
                            }
                        }
                    }

                    Array.Clear(deweyArray, 0, 16);
                }

                Array.Clear(prefixWaterFront, 0, (int)maxDepth + 1);
                Array.Clear(depthWaterFront, 0, (int)maxDepth + 1);
                skipDistance = targetPosition - CurrentOccurrenceId;

                if (skipDistance < 1)
                {
                    return;
                }

                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                }

                PathType pathType = null;
                for (int i = 0; i < skipDistance; i++)
                {
                    CurrentOccurrenceId++;
                    pathType = pathTypes[pathTypeBuffer[++positionInPathTypeBuffer]];
                    prefixWaterFront[prefixBuffer[positionInPathTypeBuffer]]++;
                    depthWaterFront[pathType.DeweyDepth]++;
                    holder.PathType = pathType;
                }

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }
                
                for (int i = 0; i < maxDepth; i++)
                {
                    int read = prefixWaterFront[i] - depthWaterFront[i];
                    prefixWaterFront[i + 1] += prefixWaterFront[i];
                    depthWaterFront[i + 1] += depthWaterFront[i];
                    
                    if (positions[i] < 0 && read > 0)
                    {
                        LoadColumn(i);    
                    }

                    positions[i] += read;
                    chunkRemaining[i] -= (uint)read;

                    if (positions[i] > 128)
                    {
                        int newPos = positions[i] & 127;
                        LoadColumn(i);
                        positions[i] = newPos;
                    }

                    if (positions[i] == 128)
                    {
                        positions[i] = -1;
                        deweyArray[i] = (ushort)deweyColumnBuffer[i][127];
                    }

                    if (positions[i] > 0)
                    {
                        deweyArray[i] = (ushort)deweyColumnBuffer[i][positions[i] - 1];
                    }
                }

                deweyHolder.Count = holder.PathType.DeweyDepth;
            }

            private void LoadColumn(int i)
            {
                offsets[i] = (uint)wrapper.DecodeSingle(
                    dataBuffer,
                    (int)offsets[i],
                    (int)columnCounts[i],
                    deweyColumnBuffer[i],
                    0);

                columnCounts[i] -= Utils.Min(128, columnCounts[i]);
                positions[i] = 0;
            }

            private void SkipChunks(int skipChunks, int[] columnMoves)
            {
                pathTypeOffset = wrapper.SkipMultiple(
                    dataBuffer, pathTypeOffset, skipChunks, TotalOccurrences - CurrentOccurrenceId - 1);

                SkipPrefixChunks(skipChunks, columnMoves);
            }

            private void SkipPrefixChunks(int skipChunks, int[] columnMoves)
            {
                for (int i = 0; i < skipChunks; i++)
                {
                    uint localDept;
                    currentPrefixOffset = PrefixedVByte.Decode(dataBuffer, currentPrefixOffset, out localDept);
                    for (int j = 0; j < localDept; j++)
                    {
                        uint columnCount;
                        currentPrefixOffset = PrefixedVByte.Decode(dataBuffer, currentPrefixOffset, out columnCount);
                        columnMoves[j] += (int)columnCount;
                    }

                    currentPrefixOffset = wrapper.SkipSingle(
                        dataBuffer, currentPrefixOffset, TotalOccurrences - CurrentOccurrenceId - 1);
                }
            }

            private void LoadDewey(int depth, uint prefixLength)
            {
                deweyHolder.Count = depth;

                for (int i = (int)prefixLength; i < depth; i++)
                {
                    if (positions[i] < 0)
                    {
                        LoadColumn(i);
                    }

                    deweyArray[i] = (ushort)deweyColumnBuffer[i][positions[i]++];
                    chunkRemaining[i]--;

                    if (positions[i] > 127)
                    {
                        positions[i] = -1;
                    }
                }
            }

            private void DecodeChunk()
            {
                DecodePathChunk();
                DecodePrefixChunk();
            }

            private void DecodePrefixChunk()
            {
                uint localDept;
                currentPrefixOffset = PrefixedVByte.Decode(dataBuffer, currentPrefixOffset, out localDept);
                for (int i = 0; i < localDept; i++)
                {
                    currentPrefixOffset = PrefixedVByte.Decode(dataBuffer, currentPrefixOffset, out chunkRemaining[i]);
                }

                currentPrefixOffset = wrapper.DecodeSingle(
                    dataBuffer, currentPrefixOffset, TotalOccurrences - CurrentOccurrenceId - 1, prefixBuffer, 0);

                for (int i = 0; i < localDept; i++)
                {
                    if (positions[i] < 0)
                    {
                        if (columnCounts[i] > 0)
                        {
                            LoadColumn(i);
                        }
                        else
                        {
                            deweyColumnBuffer[i][0] = 0;
                        }

                        positions[i] = 0;
                    }
                }

                Array.Clear(deweyArray, 0, (int)maxDepth);
            }

            private void DecodePathChunk()
            {
                pathTypeOffset = wrapper.DecodeSingle(
                    dataBuffer, pathTypeOffset, TotalOccurrences - CurrentOccurrenceId - 1, pathTypeBuffer, 0);
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                uint prefixOffset;
                offset = PrefixedVByte.Decode(buffer, offset, out maxDepth);
                offset = PrefixedVByte.Decode(buffer, offset, out prefixOffset);
                offset = PrefixedVByte.Decode(buffer, offset, out offsets[0]);

                currentPrefixOffset = (int)prefixOffset;

                offsets[0] += prefixOffset;
                for (int i = 1; i < maxDepth; i++)
                {
                    offset = PrefixedVByte.Decode(buffer, offset, out offsets[i]);
                    offsets[i] += offsets[i - 1];
                }

                for (int i = 0; i < maxDepth; i++)
                {
                    offset = PrefixedVByte.Decode(buffer, offset, out columnCounts[i]);
                }

                for (int i = 0; i < maxDepth; i++)
                {
                    positions[i] = -1;
                }

                return offset;
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                SetupChunk(occurrences);
                Reset();

                return this;
            }

            public override void Reset()
            {
                base.Reset();
                pathTypeOffset = 0;
                positionInPathTypeBuffer = -1;
            }
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory, new NPforDecodingWrapper());
        }
    }
}
