﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;
    using Documents.DocumentModel.Scope;

    using PostingFile;

    using Readers;
    using Utils.Collections;
    using Utils.Compression;
    using Utils.ByteArrays;

    public sealed class PtsCNPforScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private const int MaxPathTypes = 256;
            private const int MaxChunks = 256 + 128;
            private const int MaxDepth = 16;
            
            private readonly NPforDecodingWrapper wrapper;
            private byte[] dataBuffer = new byte[1024];

            private readonly uint[] pathTypeBuffer = new uint[128];
            private readonly int[] deweyColumnCount = new int[MaxDepth];
            private readonly int[] deweyColumnKnownOffsets = new int[MaxDepth];
            private readonly int[][] deweyColumnOffsets = new int[MaxDepth][];

            private readonly uint[][] loadedChunksTags = new uint[MaxDepth][];
            private readonly uint[][] loadedChunks = new uint[MaxDepth][];

            private readonly int[][] partitionPositions = new int[MaxPathTypes][];
            private readonly ushort[] pathTypeIdMap = new ushort[MaxPathTypes];
            private readonly int[] pathTypeFrequencies = new int[MaxPathTypes];

            private bool prepared;
            private uint sequence;

            private int maxDepth;
            private int uniquePathTypes;

            private int pathTypeOffset;
            private int positionInPathTypeBuffer;

            private uint[] headerLoadArray = new uint[256];

            private readonly ushort[] deweyArray = new ushort[MaxDepth];
            private readonly Dewey deweyHolder;

            public ChunkEnumerator(ScopeDirectory scopeDirectory, NPforDecodingWrapper wrapper) 
                : base(scopeDirectory)
            {
                this.wrapper = wrapper;

                deweyHolder = new Dewey(deweyArray);
                holder.Dewey = deweyHolder;

                for (int i = 0; i < MaxDepth; i++)
                {
                    deweyColumnOffsets[i] = new int[MaxChunks];
                    loadedChunks[i] = new uint[MaxChunks << 7];
                    loadedChunksTags[i] = new uint[MaxChunks];
                    for (int j = 0; j < MaxChunks; j++)
                    {
                        loadedChunksTags[i][j] = uint.MaxValue;
                    }
                }

                for (int i = 0; i < MaxPathTypes; i++)
                {
                    partitionPositions[i] = new int[MaxDepth];
                }
            }

            public override bool MoveNext()
            {
                if (!prepared)
                {
                    Prepare();
                }

                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                }

                CurrentOccurrenceId++;

                uint rawType = pathTypeBuffer[++positionInPathTypeBuffer];
                PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                holder.PathType = pathType;

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                if (CurrentOccurrenceId < TotalOccurrences)
                {
                    if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                    {
                        LoadDewey((int) rawType, pathType.DeweyDepth);
                    }

                    return true;
                }

                return false;
            }

            public override void MoveTo(int targetPosition)
            {
                if (!prepared)
                {
                    Prepare();
                }

                if (CurrentOccurrenceId >= targetPosition)
                {
                    return;
                }

                int[] encounteredPathTypes = new int[uniquePathTypes];
                uint rawType = 0;
                while (CurrentOccurrenceId < targetPosition)
                {
                    if (positionInPathTypeBuffer < 0)
                    {
                        DecodeChunk();
                    }

                    rawType = pathTypeBuffer[++positionInPathTypeBuffer];
                    ++encounteredPathTypes[rawType];
                    CurrentOccurrenceId++;

                    if (positionInPathTypeBuffer > 126)
                    {
                        positionInPathTypeBuffer = -1;
                    }
                }
               
                PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                holder.PathType = pathType;
                encounteredPathTypes[rawType]--;

                for (int i = 0; i < uniquePathTypes; i++)
                {
                    SkipDeweys(i, encounteredPathTypes[i]);
                }

                if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                {
                    LoadDewey((int) rawType, pathType.DeweyDepth);
                }
            }

            public override int SkipToPathType(MutableBitArray candidatePathTypes)
            {
                if(!prepared)
                {
                    Prepare();
                }

                if (holder.PathType == null)
                {
                    if (!MoveNext())
                    {
                        return TotalOccurrences + 1;
                    }

                    return SkipToPathType(candidatePathTypes);
                }

                if (candidatePathTypes[holder.PathType.PathId])
                {
                    return CurrentOccurrenceId;
                }

                for (int i = 0; i < uniquePathTypes; i++)
                {
                    if (candidatePathTypes[pathTypeIdMap[i]])
                    {
                        if (positionInPathTypeBuffer < 0)
                        {
                            positionInPathTypeBuffer = 127;
                        }

                        uint rawType = pathTypeBuffer[positionInPathTypeBuffer];
                        int[] encounteredPathTypes = new int[uniquePathTypes];

                        if (positionInPathTypeBuffer >= 127)
                        {
                            positionInPathTypeBuffer = -1;
                        }

                        while (!candidatePathTypes[pathTypeIdMap[rawType]])
                        {
                            if (positionInPathTypeBuffer < 0)
                            {
                                DecodeChunk();
                            }

                            CurrentOccurrenceId++;

                            rawType = pathTypeBuffer[++positionInPathTypeBuffer];

                            if (positionInPathTypeBuffer >= 127)
                            {
                                positionInPathTypeBuffer = -1;
                            }

                            if (CurrentOccurrenceId >= TotalOccurrences)
                            {
                                return TotalOccurrences + 1;
                            }
                        }

                        PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                        holder.PathType = pathType;

                        for (int j = 0; j < uniquePathTypes; j++)
                        {
                            SkipDeweys(j, encounteredPathTypes[j]);
                        }

                        if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                        {
                            LoadDewey((int)rawType, pathType.DeweyDepth);
                        }

                        return CurrentOccurrenceId;     
                    }
                }

                return TotalOccurrences + 1;
            }

            private void DecodeChunk()
            {
                pathTypeOffset = wrapper.DecodeSingle(
                    dataBuffer, pathTypeOffset, TotalOccurrences - CurrentOccurrenceId - 1, pathTypeBuffer, 0);
              
                positionInPathTypeBuffer = -1;
            }

            private void LoadDewey(int rawType, int depth)
            {
                int[] partition = partitionPositions[rawType];
                for (int i = 0; i < depth; i++)
                {
                    int targetChunk = partition[i] >> 7;
                    if (loadedChunksTags[i][targetChunk] != sequence)
                    {
                        LoadDeweyChunk(i, targetChunk);
                    }

                    deweyArray[i] = (ushort)loadedChunks[i][partition[i]++];
                }

                deweyHolder.Count = depth;
            }

            private void SkipDeweys(int rawType, int count)
            {
                int[] partition = partitionPositions[rawType];
                int depth = scopeDirectory[pathTypeIdMap[rawType]].DeweyDepth;
                for (int i = 0; i < depth; i++)
                {
                    partition[i] += count;
                }
            }

            private void LoadDeweyChunk(int column, int chunk)
            {
                int[] offsets = deweyColumnOffsets[column];
                int lastKnownOffset = deweyColumnKnownOffsets[column];
                if (lastKnownOffset < chunk)
                {
                    wrapper.SkipMultiple(
                        dataBuffer, 
                        offsets[lastKnownOffset],
                        chunk - lastKnownOffset,
                        deweyColumnCount[column], 
                        offsets,
                        lastKnownOffset + 1);
                    deweyColumnKnownOffsets[column] = chunk;
                }

                uint[] decodeBuffer = loadedChunks[column];
                int offset = chunk << 7;
                loadedChunksTags[column][chunk] = sequence;
                offsets[chunk + 1] = wrapper.DecodeSingle(
                    dataBuffer, offsets[chunk], deweyColumnCount[column] - offset, decodeBuffer, offset);
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                offset = PrefixedVByte.Decode(buffer, offset, out deweyColumnOffsets[0][0]);
                offset = PrefixedVByte.Decode(buffer, offset, out uniquePathTypes);
                offset = PrefixedVByte.Decode(buffer, offset, out maxDepth);

                int requiredHeaderValues = (uniquePathTypes << 1) + (maxDepth - 1);
                if (headerLoadArray.Length < requiredHeaderValues)
                {
                    headerLoadArray = new uint[requiredHeaderValues << 1];
                }
                
                int decoded = 0;
                while (decoded < requiredHeaderValues)
                {
                    decoded += Simple9.Decode(buffer.GetUInt32At(offset), headerLoadArray, decoded);
                    offset += 4;
                }

                return offset;
            }

            private void Prepare()
            {
                // mapping and frequencies
                Array.Clear(deweyColumnCount, 0, maxDepth);
                Array.Clear(partitionPositions[0], 0, maxDepth);
                for (int i = 0; i < uniquePathTypes; i++)
                {
                    pathTypeIdMap[i] = (ushort)headerLoadArray[i];
                    pathTypeFrequencies[i] = (int)headerLoadArray[uniquePathTypes + i];
                }

                for (int i = uniquePathTypes - 2; i >= 0; i--)
                {
                    pathTypeFrequencies[i] += pathTypeFrequencies[i + 1];
                }

                for (int i = 0; i < uniquePathTypes; i++)
                {
                    int typeCount = pathTypeFrequencies[i];
                    deweyColumnCount[scopeDirectory[pathTypeIdMap[i]].DeweyDepth] += typeCount;
                    int depth = scopeDirectory[pathTypeIdMap[i]].DeweyDepth;

                    for (int j = 0; j < depth; j++)
                    {
                        partitionPositions[i + 1][j] = typeCount + partitionPositions[i][j];
                    }

                    for (int j = depth; j < maxDepth; j++)
                    {
                        partitionPositions[i + 1][j] = partitionPositions[i][j];
                    }
                }

                deweyColumnCount[0] = TotalOccurrences;
                for (int i = 1; i < maxDepth; i++)
                {
                    deweyColumnCount[i] = deweyColumnCount[i - 1] - deweyColumnCount[i];
                }

                // dewey column offsets
                for (int i = 0, pos = uniquePathTypes << 1; i < maxDepth; i++)
                {
                    deweyColumnKnownOffsets[i] = 0;
                    deweyColumnOffsets[i + 1][0] = deweyColumnOffsets[i][0] + (int)headerLoadArray[pos++];
                }

                prepared = true;
                sequence++;
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                SetupChunk(occurrences);
                Reset();

                return this;
            }

            public override void Reset()
            {
                base.Reset();
                prepared = false;
                pathTypeOffset = 0;
                positionInPathTypeBuffer = -1;
            }
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory, new NPforDecodingWrapper());
        }
    }
}
