﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor.ColumnWise
{
    using System;

    using Documents.DocumentModel.Scope;

    using Utils.Compression;

    using Writers;
    using Utils.ByteArrays;

    public sealed class PtsCNPforScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly NPforEncodingWrapper wrapper;
        private readonly ScopeDirectory scopeDirectory;

        private int[] pathTypeId = new int[256];
        private uint[] pathTypeIdMap = new uint[256];
        private int[] pathTypeCounts = new int[256];
        private int[][] deweyPartitions = new int[256][];
        
        private uint[] valuesWrapped = new uint[1024];

        private readonly int[] depthCounts = new int[16];
        private readonly uint[][] deweyColumnBuffers = new uint[16][];

        private int maxDepth;
        private int uniquePathTypes;
        private int deweyStartOffset;
        private readonly int[] deweyColumnStartOffset = new int[16];

        internal PtsCNPforScopeColumnWriter(
            ScopeDirectory scopeDirectory, NPforEncodingWrapper wrapper)
        {
            this.wrapper = wrapper;
            this.scopeDirectory = scopeDirectory;

            for (int i = 0; i < deweyPartitions.Length; i++)
            {
                deweyPartitions[i] = new int[16];
            }

            for (int i = 0; i < deweyColumnBuffers.Length; i++)
            {
                deweyColumnBuffers[i] = new uint[1024];
            }
        }

        public override int MaxHeaderBytes
        {
            get { return 8 + (uniquePathTypes << 2); }
        }

        private uint[] headerEncodeValues = new uint[256];

        public override int WriteHeader(byte[] buffer, int offset)
        {
            offset = PrefixedVByte.Encode(deweyStartOffset, buffer, offset);
            offset = PrefixedVByte.Encode(uniquePathTypes, buffer, offset);
            offset = PrefixedVByte.Encode(maxDepth, buffer, offset);

            int requiredHeaderValues = (uniquePathTypes << 1) + (maxDepth - 1);

            if (headerEncodeValues.Length < requiredHeaderValues)
            {
                headerEncodeValues = new uint[requiredHeaderValues << 1];
            }

            pathTypeCounts[uniquePathTypes] = 0;
            for (int i = 0; i < uniquePathTypes; i++)
            {
                headerEncodeValues[i] = (uint)pathTypeId[i];
                pathTypeCounts[i] -= pathTypeCounts[i + 1];
                headerEncodeValues[i + uniquePathTypes] = (uint)pathTypeCounts[i];
            }

            int pos = uniquePathTypes << 1;
            for (int i = 0; i < maxDepth - 1; i++)
            {
                headerEncodeValues[pos++] = (uint)deweyColumnStartOffset[i];
            }

            int encoded = 0;
            while (encoded < pos)
            {
                uint word;
                encoded += Simple9.Encode(headerEncodeValues, encoded, out word);
                buffer.SetValueAt(offset, word);
                offset += 4;
            }

            return offset;
        }

        public override int Write(byte[] buffer, int offset)
        {
            if (pathTypeCounts.Length < scopeDirectory.NumberOfPaths + 128)
            {
                pathTypeCounts = new int[(scopeDirectory.NumberOfPaths + 128) << 1];    
            }

            maxDepth = 0;
            uniquePathTypes = 0;
            
            int minPathId = (scopeDirectory.NumberOfPaths + 1), maxPathId = 0;
            Array.Clear(pathTypeCounts, 0, scopeDirectory.NumberOfPaths + 1);
            Array.Clear(depthCounts, 0, 16);

            for (int i = 0; i < count; i++)
            {
                ushort id = pathTypes[i];
                uniquePathTypes += pathTypeCounts[id]++ == 0 ? 1 : 0;
                minPathId = Utils.Min(minPathId, id);
                maxPathId = Utils.Max(maxPathId, id);

                int depth = scopeDirectory[id].DeweyDepth;
                depthCounts[depth]++;
                maxDepth = Utils.Max(depth, maxDepth);
            }

            if (pathTypeId.Length < uniquePathTypes)
            {
                pathTypeId = new int[uniquePathTypes << 1];
                deweyPartitions = new int[uniquePathTypes << 1][];

                for (int i = 0; i < deweyPartitions.Length; i++)
                {
                    deweyPartitions[i] = new int[16];
                }
            }

            int pos = 0;
            for (int i = minPathId; i < maxPathId + 1; i++)
            {
                if (pathTypeCounts[i] > 0)
                {
                    pathTypeId[pos] = i;
                    pathTypeCounts[pos++] = -pathTypeCounts[i];
                }
            }

            // HACK: pathTypeCount contain negative counts for decreasing sort
            Array.Sort(pathTypeCounts, pathTypeId, 0, uniquePathTypes);

            for (uint i = 0; i < uniquePathTypes; i++)
            {
                pathTypeIdMap[pathTypeId[i] - minPathId] = i;
                Array.Clear(deweyPartitions[i], 0, 16);
            }

            
            for (int i = 0; i < uniquePathTypes; i++)
            {
                pathTypeCounts[i] = -pathTypeCounts[i];
                int depth = scopeDirectory[pathTypeId[i]].DeweyDepth;
                int typeCount = pathTypeCounts[i];
                for (int j = 0; j < depth; j++)
                {
                    deweyPartitions[i + 1][j] = typeCount + deweyPartitions[i][j];
                }

                for (int j = depth; j < maxDepth; j++)
                {
                    deweyPartitions[i + 1][j] = deweyPartitions[i][j];
                }
            }

            if (pathTypeIdMap.Length < maxPathId - minPathId + 1)
            {
                pathTypeIdMap = new uint[maxPathId - minPathId + 1];
            }
            
            if (valuesWrapped.Length < count + 128)
            {
                valuesWrapped = new uint[(count + 128) << 1];
            }

            depthCounts[0] = count;
            for (int i = 0; i < maxDepth; i++)
            {
                depthCounts[i + 1] = depthCounts[i] - depthCounts[i + 1];
                if (deweyColumnBuffers[i].Length < depthCounts[i] + 128)
                {
                    deweyColumnBuffers[i] = new uint[depthCounts[i] + 128];
                }
            }

            int deweyOffset = 0;
            for (int i = 0; i < count; i++)
            {
                int pathId = pathTypes[i];
                int depth = scopeDirectory[pathId].DeweyDepth;
                uint mapped = valuesWrapped[i] = pathTypeIdMap[pathId - minPathId];

                int[] partition = deweyPartitions[mapped];
                for (int j = 0; j < depth; j++)
                {
                    deweyColumnBuffers[j][partition[j]++] = deweyElements[deweyOffset++];
                }
            }

            int beginning = offset;
            offset = wrapper.EncodeMultiple(buffer, offset, valuesWrapped, count);
            deweyStartOffset = offset - beginning;

            int columnStart = offset;
            for (int i = 0; i < maxDepth; i++)
            {
                offset = wrapper.EncodeMultiple(buffer, offset, deweyColumnBuffers[i], depthCounts[i]);
                deweyColumnStartOffset[i] = offset - columnStart;
                columnStart = offset;
            }

            return offset;
        }
    }
}
