﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using System;
    
    using Utils.Compression;

    internal sealed class NPforDecodingWrapper
    {
        private readonly int padLimit;

        private readonly uint[] pforExceptions = new uint[128 + 28];
        private readonly uint[] pforExceptionIndexes = new uint[128 + 28];

        public unsafe int DecodeSingle(byte[] data, int offset, int count, uint[] output, int outOffset)
        {
            if (count < padLimit)
            {
                return DecodeWithFallbackMechanism(data, offset, count, output, outOffset);
            }
            
            /*
            uint bitWidth, chunkLength, numberOfExceptions, frameOfReference;
            fixed (byte* bufferPtr = &data[offset])
            {
                var ptr = (uint*) bufferPtr;
                PforUtils.UnpackControlWord(
                    ptr[0], out bitWidth, out chunkLength, out numberOfExceptions, out frameOfReference);

                int word = ((int) bitWidth << 2) + 1;
                for (int i = 0;
                        i < numberOfExceptions;
                        i += Simple9.Decode(ptr[word], pforExceptions, i), word++)
                {
                }

                for (int i = 0;
                        i < numberOfExceptions;
                        i += Simple9.Decode(ptr[word], pforExceptionIndexes, i), word++)
                {
                }

                Array.Clear(output, 0, 128);
                fixed (uint* outPtr = &output[outOffset])
                {
                    NewPFor.Decode(
                        128,
                        (int)bitWidth,
                        frameOfReference,
                        (int)numberOfExceptions,
                        ptr + 1,
                        outPtr,
                        pforExceptions,
                        pforExceptionIndexes);
                }

                offset += word << 2;
            }*/
            
            
            fixed (byte* bufferPtr = &data[offset])
            {
                Array.Clear(output, outOffset, 128);
                fixed (uint* outPtr = &output[outOffset])
                {
                    fixed (uint* exceptions = &pforExceptions[0])
                    {
                        fixed (uint* indexes = &pforExceptionIndexes[0])
                        {
                            offset += NativeCode.NPfor.Decode(
                                (uint*) bufferPtr,
                                outPtr, 
                                exceptions, 
                                indexes) << 2;
                        }
                    }
                }
            }
            
            return offset;
        }

        public unsafe int DecodeMulti(byte[] data, int offset, uint[] output, int outOffset, int blocks)
        {
            fixed (byte* bufferPtr = &data[offset])
            {
                Array.Clear(output, outOffset, 128 * blocks);
                fixed (uint* outPtr = &output[outOffset])
                {
                    fixed (uint* exceptions = &pforExceptions[0])
                    {
                        fixed (uint* indexes = &pforExceptionIndexes[0])
                        {
                            offset += NativeCode.NPfor.Decode(
                                (uint*)bufferPtr,
                                outPtr,
                                exceptions,
                                indexes,
                                blocks) << 2;
                        }
                    }
                }
            }

            return offset;
        }

        public int PadLimit
        {
            get { return padLimit; }
        }

        public NPforDecodingWrapper(int padLimit = NPforSettings.PadLimit)
        {
            this.padLimit = padLimit;
        }

        public unsafe int SkipSingle(byte[] data, int offset, int count)
        {
            if (count < padLimit)
            {
                return SkipWithFallbackMechanism(data, offset, count);
            }

            /*
            fixed (byte* ptr = &data[offset])
            {
                var intPtr = (uint*) ptr;
                int word = 0;
                uint bitWidth, chunkLength, numberOfExceptions, frameOfReference;
                PforUtils.UnpackControlWord(
                    intPtr[word], out bitWidth, out chunkLength, out numberOfExceptions, out frameOfReference);
                word += 1 + ((int)bitWidth << 2);

                for (int i = 0; i < numberOfExceptions; i += numValues[intPtr[word] >> 28], word += 1)
                {
                }

                for (int i = 0; i < numberOfExceptions; i += numValues[intPtr[word] >> 28], word += 1)
                {
                }

                offset += (word << 2);
            }*/

            fixed (byte* ptr = &data[offset])
            {
                offset += (NativeCode.NPfor.SkipBlocks((uint*)ptr, 1) << 2);
            }
            

            return offset;
        }

        public unsafe void SkipMultiple(byte[] data, int offset, int blocks, int remaining, int[] outputOffsets, int outputOffset)
        {
            int realBlocks = remaining >> 7;
            int rest = remaining & 127;

            if (realBlocks >= blocks)
            {
                realBlocks = blocks;
                rest = 0;
            }
            else if (rest >= padLimit)
            {
                realBlocks++;
                rest = 0;
            }

            fixed (byte* ptr = &data[offset])
            {
                fixed (int* outPtr = &outputOffsets[outputOffset])
                {
                    NativeCode.NPfor.SkipBlocks((uint*)ptr, realBlocks, offset, outPtr);    
                }
            }

            if (rest > 0)
            {
                outputOffsets[outputOffset + realBlocks] = SkipWithFallbackMechanism(data, offset, rest);
            }
        }

        public unsafe int SkipMultiple(byte[] data, int offset, int blocks, int remaining)
        {
            int realBlocks = remaining >> 7;
            int rest = remaining & 127;

            if (realBlocks >= blocks)
            {
                realBlocks = blocks;
                rest = 0;
            }
            else if (rest >= padLimit)
            {
                realBlocks++;
                rest = 0;
            }

            fixed (byte* ptr = &data[offset])
            {
                offset += (NativeCode.NPfor.SkipBlocks((uint*)ptr, realBlocks) << 2);
            }

            if (rest > 0)
            {
                return SkipWithFallbackMechanism(data, offset, rest);
            }

            /*
            fixed (byte* ptr = &data[offset])
            {
                var intPtr = (uint*) ptr;
                int word = 0;
                uint bitWidth, chunkLength, numberOfExceptions, frameOfReference;
                PforUtils.UnpackControlWord(
                    intPtr[word], out bitWidth, out chunkLength, out numberOfExceptions, out frameOfReference);
                word += 1 + ((int)bitWidth << 2);

                for (int i = 0; i < numberOfExceptions; i += numValues[intPtr[word] >> 28], word += 1)
                {
                }

                for (int i = 0; i < numberOfExceptions; i += numValues[intPtr[word] >> 28], word += 1)
                {
                }

                offset += (word << 2);
            }*/

            return offset;
        }

        private static int SkipWithFallbackMechanism(byte[] data, int offset, int count)
        {
            return PrefixedVByte.SkipMany(data, offset, count);
        }

        private static int DecodeWithFallbackMechanism(
            byte[] data, int offset, int count, uint[] output, int outOffset)
        {
            for (int i = 0; i < count; i++)
            {
                offset = PrefixedVByte.Decode(data, offset, out output[outOffset + i]);
            }

            return offset;
        }
    }
}
