﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using Documents.DocumentModel.Scope;
    using PostingFile;
    using Readers;
    using Utils.Collections;
    using Utils.Compression;

    public sealed class NPforScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private readonly NPforDecodingWrapper wrapper;
            private byte[] dataBuffer = new byte[1024 << 2];

            private readonly uint[] pathTypeDecodeBuffer = new uint[128 + 28];
            private readonly uint[] deweyDecodeBuffer = new uint[128 + 28];

            private uint deweyCount;

            private uint deweyOffset;

            private int positionInPathTypeBuffer;
            private int positionInDeweyBuffer;

            private int currentDeweyOffset;
            private int currentPathTypeOffset;

            private readonly ushort[] deweyArray = new ushort[32];
            private readonly Dewey deweyHolder;

            private uint[] chunkCounts = new uint[128];

            public ChunkEnumerator(ScopeDirectory scopeDirectory, NPforDecodingWrapper wrapper) 
                : base(scopeDirectory)
            {
                this.wrapper = wrapper;
                deweyHolder = new Dewey(deweyArray);
                holder.Dewey = deweyHolder;
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                deweyCount = 0;
                int chunks = (int)(totalOccurrences >> 7) + (((totalOccurrences & 127) > 0) ? 1 : 0);
                if (chunkCounts.Length < chunks)
                {
                    chunkCounts = new uint[chunks << 1];
                }

                for (int i = 0; i < chunks; i++)
                {
                    offset = PrefixedVByte.Decode(buffer, offset, out chunkCounts[i]);
                    deweyCount += chunkCounts[i];
                }

                return PrefixedVByte.Decode(buffer, offset, out deweyOffset);
            }

            private int MoveToPathTypeAt(int targetPosition)
            {
                int skipDistance = targetPosition - CurrentOccurrenceId - 1;
                int deweyDistance = 0;
                int remainigInCurrent = 128 - positionInPathTypeBuffer - 1;

                if (positionInPathTypeBuffer > -1 && skipDistance > remainigInCurrent)
                {
                    if (remainigInCurrent < 64)
                    {   // traverse forward
                        PathType pt = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                        for (int i = 0; i < remainigInCurrent; i++)
                        {
                            deweyDistance += pt.DeweyDepth;
                            pt = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                        }
                    }
                    else
                    {   // rollback
                        int rolledBack = 0;
                        int currentChunk = Utils.Max((CurrentOccurrenceId + 1) >> 7, 0);
                        for (int i = 0; i <= positionInPathTypeBuffer; i++)
                        {
                            rolledBack += pathTypes[pathTypeDecodeBuffer[i]].DeweyDepth;
                        }

                        int skipped = (int)chunkCounts[currentChunk] - rolledBack;
                        deweyDistance += skipped;
                    }

                    CurrentOccurrenceId += remainigInCurrent;
                    skipDistance -= remainigInCurrent;
                    positionInPathTypeBuffer = -1;
                }

                if (positionInPathTypeBuffer < 0 && skipDistance > 127)
                {   // clean start
                    int skipBuffers = (skipDistance >> 7);
                    currentPathTypeOffset = SkipBuffers(currentPathTypeOffset, skipBuffers, TotalOccurrences - CurrentOccurrenceId);
                    positionInPathTypeBuffer = -1;
                    int currentChunk = Utils.Max((CurrentOccurrenceId + 1) >> 7, 0);
                    for (int i = 0; i < skipBuffers; i++)
                    {
                        deweyDistance += (int)chunkCounts[currentChunk + i];
                    }

                    CurrentOccurrenceId += skipBuffers << 7;
                }


                if (positionInPathTypeBuffer < 0)
                {
                    currentPathTypeOffset = ReadOneBuffer(
                        currentPathTypeOffset, pathTypeDecodeBuffer, TotalOccurrences - CurrentOccurrenceId - 1);
                    // positionInPathTypeBuffer = 0;
                }

                PathType pathType = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                while (++CurrentOccurrenceId < targetPosition)
                {
                    if (positionInPathTypeBuffer >= 127)
                    {
                        currentPathTypeOffset = ReadOneBuffer(
                            currentPathTypeOffset, pathTypeDecodeBuffer, TotalOccurrences - CurrentOccurrenceId - 1);
                        positionInPathTypeBuffer = -1;
                    }

                    deweyDistance += pathType.DeweyDepth;
                    pathType = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                }

                pathType = pathTypes[pathTypeDecodeBuffer[positionInPathTypeBuffer]];
                holder.PathType = pathType;

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                return deweyDistance;
            }

            private void SkipDeweyElements(int deweyDistance)
            {
                if (positionInDeweyBuffer >= 0)
                {
                    int remainingInCurrent = 128 - positionInDeweyBuffer;
                    if (remainingInCurrent >= deweyDistance)
                    {
                        positionInDeweyBuffer += deweyDistance;
                        return;
                    }

                    deweyDistance -= remainingInCurrent;
                }

                if (deweyDistance > 127)
                {
                    int chunks = deweyDistance >> 7;
                    int skipped = chunks << 7;
                    currentDeweyOffset = SkipBuffers(currentDeweyOffset, chunks, (int)deweyCount);
                    deweyDistance -= skipped;
                    deweyCount -= (uint)skipped;
                }
                
                currentDeweyOffset = ReadOneBuffer(currentDeweyOffset, deweyDecodeBuffer, (int)deweyCount);
                deweyCount -= 128;

                positionInDeweyBuffer = deweyDistance;
            }

            public override void MoveTo(int targetPosition)
            {
                if (CurrentOccurrenceId >= targetPosition || targetPosition >= TotalOccurrences)
                {
                    return;
                }

                SkipDeweyElements(MoveToPathTypeAt(targetPosition));
                LoadDewey(holder.PathType.DeweyDepth);
            }

            public override int SkipToPathType(MutableBitArray candidatePathTypes)
            {
                bool hasMore = (holder.PathType == null || !candidatePathTypes[holder.PathType.PathId]) && MoveNext();
                if (hasMore && !candidatePathTypes[holder.PathType.PathId])
                {
                    PathType pathType = holder.PathType ?? pathTypes[0];
                    int deweySkip = 0;
                    while (!candidatePathTypes[pathType.PathId] && CurrentOccurrenceId < TotalOccurrences)
                    {
                        if (positionInPathTypeBuffer < 0)
                        {
                            currentPathTypeOffset = ReadOneBuffer(
                                currentPathTypeOffset, pathTypeDecodeBuffer, TotalOccurrences - CurrentOccurrenceId - 1);
                        }

                        CurrentOccurrenceId++;

                        pathType = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                        deweySkip += pathType.DeweyDepth;

                        if (positionInPathTypeBuffer >= 127)
                        {
                            positionInPathTypeBuffer = -1;
                        }
                    }

                    int depth = pathType.DeweyDepth;
                    holder.PathType = pathType;
                    holder.Dewey.Count = pathType.DeweyDepth;

                    deweySkip -= depth;
                    SkipDeweyElements(deweySkip);
                    LoadDewey(depth);
                }

                return CurrentOccurrenceId;
            }

            private void LoadDewey(int deweyDepth)
            {
                deweyHolder.Count = deweyDepth;

                if (positionInDeweyBuffer < 0)
                {
                    currentDeweyOffset = ReadOneBuffer(currentDeweyOffset, deweyDecodeBuffer, (int)deweyCount);
                    deweyCount -= 128;
                    positionInDeweyBuffer = 0;
                }

                int i = 0;
                while (i < deweyDepth && positionInDeweyBuffer < 128)
                {
                    deweyArray[i++] = (ushort)deweyDecodeBuffer[positionInDeweyBuffer++];
                }

                if (positionInDeweyBuffer > 127)
                {
                    positionInDeweyBuffer = -1;

                    if (i < deweyDepth)
                    {
                        currentDeweyOffset = ReadOneBuffer(currentDeweyOffset, deweyDecodeBuffer, (int)deweyCount);
                        deweyCount -= 128;
                        positionInDeweyBuffer = 0;

                        while (i < deweyDepth)
                        {
                            deweyArray[i++] = (ushort)deweyDecodeBuffer[positionInDeweyBuffer++];
                        }
                    }
                }
            }

            public override bool MoveNext()
            {
                if (positionInPathTypeBuffer < 0)
                {
                    currentPathTypeOffset = ReadOneBuffer(currentPathTypeOffset, pathTypeDecodeBuffer, TotalOccurrences - CurrentOccurrenceId - 1);
                }

                CurrentOccurrenceId++;

                PathType pathType = pathTypes[pathTypeDecodeBuffer[++positionInPathTypeBuffer]];
                holder.PathType = pathType;
                
                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                LoadDewey(pathType.DeweyDepth);
                return CurrentOccurrenceId < TotalOccurrences;
            }

            public override void Reset()
            {
                currentPathTypeOffset = 0;
                positionInPathTypeBuffer = -1;
                positionInDeweyBuffer = -1;
                currentDeweyOffset = (int)deweyOffset;
                base.Reset();
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                SetupChunk(occurrences);
                Reset();

                return this;
            }

            private int ReadOneBuffer(int offset, uint[] decodedBuffer, int remaining)
            {
                return wrapper.DecodeSingle(
                    dataBuffer,
                    offset,
                    Utils.Min(128, remaining),
                    decodedBuffer,
                    0);
            }

            private int SkipBuffers(int offset, int buffers, int remaining)
            {
                return wrapper.SkipMultiple(dataBuffer, offset, buffers, remaining);
            }
        }

        private readonly int padLimit;

        public NPforScopeColumnReader(int padLimit = NPforSettings.PadLimit)
        {
            this.padLimit = padLimit;
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory, new NPforDecodingWrapper(padLimit));
        }
    }
}
