﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using Documents.DocumentModel.Scope;
 
    using PostingFile;
    
    using Readers;
    using Utils.ByteArrays;
    using Utils.Collections;
    using Utils.Compression;

    public sealed class PtsNPforScopeColumnReader : AbstractScopeColumnReader
    {
        private sealed class ChunkEnumerator : AbstractScopeChunkEnumerator
        {
            private const int MaxChunks = 4096;

            private readonly NPforDecodingWrapper wrapper;
            private byte[] dataBuffer = new byte[1024];

            private readonly ushort[] deweyArray = new ushort[16];
            private readonly Dewey deweyHolder;

            private bool prepared;
            private int pathTypeOffset;
            private int deweyOffset;
            private int positionInPathTypeBuffer;

            private readonly uint[] headerLoadArray = new uint[256];

            private int uniquePathTypes;
            private readonly ushort[] pathTypeIdMap = new ushort[256];
            private readonly int[] pathTypeFrequencies = new int[256];
            private readonly int[] partitions = new int[256];

            private int sequence;
            private readonly uint[] pathTypeBuffer = new uint[128];
            private readonly uint[] deweyBuffer = new uint[MaxChunks << 7];
            private readonly int[] chunkSequences = new int[MaxChunks];
            private readonly int[] chunkOffsets = new int[MaxChunks];
            private int lastKnownOffset;
            private int totalDeweys;

            public ChunkEnumerator(ScopeDirectory scopeDirectory, NPforDecodingWrapper wrapper) 
                : base(scopeDirectory)
            {
                this.wrapper = wrapper;
                deweyHolder = new Dewey(deweyArray);
                holder.Dewey = deweyHolder;
            }

            public override bool MoveNext()
            {
                if (!prepared)
                {
                    Prepare();
                }

                if (positionInPathTypeBuffer < 0)
                {
                    DecodeChunk();
                }

                CurrentOccurrenceId++;

                uint rawType = pathTypeBuffer[++positionInPathTypeBuffer];
                PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                holder.PathType = pathType;

                if (positionInPathTypeBuffer >= 127)
                {
                    positionInPathTypeBuffer = -1;
                }

                if (CurrentOccurrenceId < TotalOccurrences)
                {
                    if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                    {
                        LoadDewey((int)rawType, pathType.DeweyDepth);
                    }

                    return true;
                }

                return false;
            }

            public override void MoveTo(int targetPosition)
            {
                if (!prepared)
                {
                    Prepare();
                }

                if (CurrentOccurrenceId >= targetPosition)
                {
                    return;
                }

                int[] encounteredPathTypes = new int[uniquePathTypes];
                uint rawType = 0;
                while (CurrentOccurrenceId < targetPosition)
                {
                    if (positionInPathTypeBuffer < 0)
                    {
                        DecodeChunk();
                    }

                    rawType = pathTypeBuffer[++positionInPathTypeBuffer];
                    ++encounteredPathTypes[rawType];
                    CurrentOccurrenceId++;

                    if (positionInPathTypeBuffer > 126)
                    {
                        positionInPathTypeBuffer = -1;
                    }
                }

                PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                holder.PathType = pathType;
                encounteredPathTypes[rawType]--;

                for (int i = 0; i < uniquePathTypes; i++)
                {
                    SkipDeweys(i, encounteredPathTypes[i]);
                }

                if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                {
                    LoadDewey((int)rawType, pathType.DeweyDepth);
                }
            }

            public override int SkipToPathType(MutableBitArray candidatePathTypes)
            {
                if (!prepared)
                {
                    Prepare();
                }

                if (holder.PathType == null)
                {
                    if (!MoveNext())
                    {
                        return TotalOccurrences + 1;
                    }

                    return SkipToPathType(candidatePathTypes);
                }

                if (candidatePathTypes[holder.PathType.PathId])
                {
                    return CurrentOccurrenceId;
                }

                for (int i = 0; i < uniquePathTypes; i++)
                {
                    if (candidatePathTypes[pathTypeIdMap[i]])
                    {
                        if (positionInPathTypeBuffer < 0)
                        {
                            positionInPathTypeBuffer = 127;
                        }

                        uint rawType = pathTypeBuffer[positionInPathTypeBuffer];
                        int[] encounteredPathTypes = new int[uniquePathTypes];

                        if (positionInPathTypeBuffer >= 127)
                        {
                            positionInPathTypeBuffer = -1;
                        }

                        while (!candidatePathTypes[pathTypeIdMap[rawType]])
                        {
                            if (positionInPathTypeBuffer < 0)
                            {
                                DecodeChunk();
                            }

                            CurrentOccurrenceId++;

                            rawType = pathTypeBuffer[++positionInPathTypeBuffer];

                            if (positionInPathTypeBuffer >= 127)
                            {
                                positionInPathTypeBuffer = -1;
                            }

                            if (CurrentOccurrenceId >= TotalOccurrences)
                            {
                                return TotalOccurrences + 1;
                            }
                        }

                        PathType pathType = pathTypes[pathTypeIdMap[rawType]];
                        holder.PathType = pathType;

                        for (int j = 0; j < uniquePathTypes; j++)
                        {
                            SkipDeweys(j, encounteredPathTypes[j]);
                        }

                        if (PathTypeMask == null || PathTypeMask[pathType.PathId])
                        {
                            LoadDewey((int)rawType, pathType.DeweyDepth);
                        }

                        return CurrentOccurrenceId;
                    }
                }

                return TotalOccurrences + 1;
            }

            private void LoadDewey(int rawType, int depth)
            {
                int partition = partitions[rawType];
                int targetChunk = partition >> 7;

                if (chunkSequences[targetChunk] != sequence)
                {
                    LoadDeweyChunk(targetChunk);
                }

                int remaining = 128 - (partition & 127);
                if (remaining < depth)
                {
                    LoadDeweyChunk(targetChunk + 1);
                }

                for (int i = 0; i < depth; i++)
                {
                    deweyArray[i] = (ushort)deweyBuffer[partition++];
                }

                partitions[rawType] = partition;
                deweyHolder.Count = depth;
            }

            private void SkipDeweys(int rawType, int count)
            {
                int depth = scopeDirectory[pathTypeIdMap[rawType]].DeweyDepth;
                partitions[rawType] += depth * count;
            }

            private void LoadDeweyChunk(int chunk)
            {
                if (lastKnownOffset < chunk)
                {
                    wrapper.SkipMultiple(
                        dataBuffer, chunkOffsets[lastKnownOffset], chunk - lastKnownOffset,
                        totalDeweys - (lastKnownOffset << 7), chunkOffsets, lastKnownOffset + 1);

                    lastKnownOffset = chunk;
                }

                int offset = chunk << 7;
                chunkSequences[chunk] = sequence;
                chunkOffsets[chunk + 1] = wrapper.DecodeSingle(
                    dataBuffer, chunkOffsets[chunk], totalDeweys - offset, deweyBuffer, offset);
                lastKnownOffset = Utils.Max(chunk + 1, lastKnownOffset);
            }

            private void DecodeChunk()
            {
                pathTypeOffset = wrapper.DecodeSingle(
                    dataBuffer, pathTypeOffset, TotalOccurrences - CurrentOccurrenceId - 1, pathTypeBuffer, 0);

                positionInPathTypeBuffer = -1;
            }

            public override int ReadHeader(byte[] buffer, int offset, uint totalOccurrences)
            {
                offset = PrefixedVByte.Decode(buffer, offset, out deweyOffset);
                offset = PrefixedVByte.Decode(buffer, offset, out uniquePathTypes);

                int decoded = 0, decode = uniquePathTypes << 1;
                while (decoded < decode)
                {
                    decoded += Simple9.Decode(buffer.GetUInt32At(offset), headerLoadArray, decoded);
                    offset += 4;
                }

                return offset;
            }

            private void Prepare()
            {
                // mapping and frequencies
                for (int i = 0; i < uniquePathTypes; i++)
                {
                    pathTypeIdMap[i] = (ushort)headerLoadArray[i];
                    pathTypeFrequencies[i] = (int)headerLoadArray[uniquePathTypes + i];
                }

                for (int i = uniquePathTypes - 2; i >= 0; i--)
                {
                    pathTypeFrequencies[i] += pathTypeFrequencies[i + 1];
                }

                partitions[0] = 0;
                for (int i = 0; i < uniquePathTypes; i++)
                {
                    int typeCount = pathTypeFrequencies[i];
                    int depth = scopeDirectory[pathTypeIdMap[i]].DeweyDepth;
                    partitions[i + 1] = partitions[i] + typeCount * depth;
                }

                totalDeweys = partitions[uniquePathTypes];

                prepared = true;
                sequence++;

                chunkOffsets[0] = deweyOffset;
                lastKnownOffset = 0;
            }

            public override AbstractScopeChunkEnumerator LoadChunk(
                PostingFileDataReader dataReader, long filePosition, uint chunkLength, uint occurrences)
            {
                if (dataBuffer.Length < chunkLength)
                {
                    dataBuffer = new byte[chunkLength << 1];
                }

                dataReader.LoadData(dataBuffer, 0, (int)chunkLength, filePosition);

                SetupChunk(occurrences);
                Reset();

                return this;
            }

            public override void Reset()
            {
                base.Reset();
                prepared = false;
                pathTypeOffset = 0;
                positionInPathTypeBuffer = -1;
            }
        }

        public override AbstractScopeChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator(ScopeDirectory, new NPforDecodingWrapper());
        }
    }
}
