﻿namespace NETing.Index.ColumnAccess.Implementations.NPfor
{
    using System;
    using Documents.DocumentModel.Scope;
    using Utils.ByteArrays;
    using Utils.Compression;
    using Writers;

    public sealed class PtsNPforScopeColumnWriter : AbstractScopeColumnWriter
    {
        private readonly NPforEncodingWrapper wrapper;
        private readonly ScopeDirectory scopeDirectory;

        private uint[] valuesWrapped = new uint[1024];
        private uint[] deweyBuffer = new uint[1024 << 3];

        private int deweyStartOffset;
        
        private int uniquePathTypes;
        
        private readonly int[] pathTypeId = new int[256];
        private readonly ushort[] pathTypeIdMap = new ushort[256];
        private readonly int[] pathTypeCounts = new int[256];
        private readonly int[] pathTypePartitions = new int[256];

        internal PtsNPforScopeColumnWriter(
            ScopeDirectory scopeDirectory, NPforEncodingWrapper wrapper)
        {
            this.wrapper = wrapper;
            this.scopeDirectory = scopeDirectory;
        }

        private readonly uint[] headerEncodeValues = new uint[256];

        public override int MaxHeaderBytes
        {
            get { return 8 + (uniquePathTypes << 2); }
        }

        public override int WriteHeader(byte[] buffer, int offset)
        {
            offset = PrefixedVByte.Encode(deweyStartOffset, buffer, offset);
            offset = PrefixedVByte.Encode(uniquePathTypes, buffer, offset);

            pathTypeCounts[uniquePathTypes] = 0;
            for (int i = 0; i < uniquePathTypes; i++)
            {
                headerEncodeValues[i] = (uint)pathTypeId[i];
                pathTypeCounts[i] -= pathTypeCounts[i + 1];
                headerEncodeValues[i + uniquePathTypes] = (uint)pathTypeCounts[i];
            }

            int encoded = 0;
            int values = uniquePathTypes << 1;
            while (encoded < values)
            {
                uint word;
                encoded += Simple9.Encode(headerEncodeValues, encoded, out word);
                buffer.SetValueAt(offset, word);
                offset += 4;
            }

            return offset;
        }

        public override int Write(byte[] buffer, int offset)
        {
            uniquePathTypes = 0;

            int minPathId = (scopeDirectory.NumberOfPaths + 1), maxPathId = 0;
            Array.Clear(pathTypeCounts, 0, scopeDirectory.NumberOfPaths + 1);

            for (int i = 0; i < count; i++)
            {
                ushort id = pathTypes[i];
                uniquePathTypes += pathTypeCounts[id]++ == 0 ? 1 : 0;
                minPathId = Utils.Min(minPathId, id);
                maxPathId = Utils.Max(maxPathId, id);
            }

            int pos = 0;
            for (int i = minPathId; i < maxPathId + 1; i++)
            {
                if (pathTypeCounts[i] > 0)
                {
                    pathTypeId[pos] = i;
                    pathTypeCounts[pos++] = -pathTypeCounts[i];
                }
            }

            // HACK: pathTypeCount contain negative counts for decreasing sort
            Array.Sort(pathTypeCounts, pathTypeId, 0, uniquePathTypes);
            
            pathTypePartitions[0] = 0;
            for (ushort i = 0; i < uniquePathTypes; i++)
            {
                pathTypeIdMap[pathTypeId[i] - minPathId] = i;
                pathTypeCounts[i] = -pathTypeCounts[i];
                pathTypePartitions[i + 1] = pathTypePartitions[i] 
                    + (pathTypeCounts[i] * scopeDirectory[pathTypeId[i]].DeweyDepth);
            }

            if (valuesWrapped.Length < count + 128)
            {
                valuesWrapped = new uint[(count + 128) << 1];
            }

            if (deweyBuffer.Length < deweyCount + 128)
            {
                deweyBuffer = new uint[(deweyCount + 128) << 1];
            }

            int deweyPos = 0;
            for (int i = 0; i < count; i++)
            {
                int reduced = pathTypeIdMap[pathTypes[i] - minPathId];
                valuesWrapped[i] = (uint)reduced;
                int depth = scopeDirectory[pathTypes[i]].DeweyDepth;
                
                for (int j = 0; j < depth; j++)
                {
                    deweyBuffer[pathTypePartitions[reduced]++] = deweyElements[deweyPos++];
                }
            }

            for (int i = 0; i < uniquePathTypes; i++)
            {
                pathTypePartitions[i] -= pathTypeCounts[i] * scopeDirectory[pathTypeId[i]].DeweyDepth;
            }

            int beginning = offset;
            offset = wrapper.EncodeMultiple(buffer, offset, valuesWrapped, count);
            deweyStartOffset = offset - beginning;

            return wrapper.EncodeMultiple(buffer, offset, deweyBuffer, deweyCount);
        }
    }
}
