﻿namespace NETing.Index.ColumnAccess.Implementations.Raw
{
    using System;
    using System.Runtime.InteropServices;
    
    using PostingFile;
 
    using Readers;

    public sealed class RawDocumentColumnReader : AbstractDocumentColumnReader
    {
        private sealed class ChunkEnumerator : AbstractDocumentChunkEnumerator
        {
            [StructLayout(LayoutKind.Explicit)]
            private struct BufferUnion
            {
                [FieldOffset(0)] public byte[] ByteBuffer;
                [FieldOffset(0)] readonly public uint[] UintBuffer;
                [FieldOffset(0)] readonly public ushort[] UshortBuffer;
            }

            private readonly BufferUnion buffers 
                = new BufferUnion { ByteBuffer = new byte[(128 << 2) + (128 << 1)] };

            private readonly uint[] documentIdBuffer;
            private readonly ushort[] frequencyBuffer;

            private int position;
            private int nextPosition;
            private int nextOccurrenceId;
            
            public ChunkEnumerator()
            {
                documentIdBuffer = buffers.UintBuffer;
                frequencyBuffer = buffers.UshortBuffer;
            }

            public override bool MoveNext()
            {
                position = nextPosition++;
                CurrentOccurrenceId = nextOccurrenceId;
                nextOccurrenceId += Frequency;
                return position < TotalDocuments;
            }

            public override int SkipTo(uint documentId)
            {
                uint lastDoc = documentIdBuffer[TotalDocuments - 1];
                if (lastDoc < documentId)
                {
                    return TotalOccurrences + 1;
                }

                if (CurrentOccurrenceId < 0)
                {
                    if (!MoveNext())
                    {
                        return TotalOccurrences + 1;
                    }
                }

                if (Current >= documentId)
                {
                    return CurrentOccurrenceId;
                }

                int index = Array.BinarySearch(
                    documentIdBuffer, position, TotalDocuments - position, documentId);

                index = index < 0 ? index ^ -1 : index;

                while (position < index)
                {
                    CurrentOccurrenceId = nextOccurrenceId;
                    position++;
                    nextOccurrenceId += Frequency;
                }

                nextPosition = position + 1;
                nextOccurrenceId = CurrentOccurrenceId + Frequency;
                return CurrentOccurrenceId;
            }

            public override ushort Frequency
            {
                get { return frequencyBuffer[FrequencyBufferOffset + position]; }
            }

            public override uint Current
            {
                get { return documentIdBuffer[position]; }
            }

            public override int RemainingDocuments
            {
                get { return TotalDocuments - position - 1; }
            }

            public override void LoadChunk(
                PostingFileDataReader dataReader,
                long documentFilePosition,
                uint documentChunkLength,
                long frequencyFilePosition,
                uint frequencyChunkLength,
                int chunkCount,
                uint occurrences)
            {
                int totalBytes = (chunkCount << 2) + (chunkCount << 1);
                dataReader.LoadData(buffers.ByteBuffer, 0, totalBytes, documentFilePosition);
                FrequencyBufferOffset = chunkCount << 1;
                SetupChunk(chunkCount, occurrences);
            }

            protected override void SetupChunk(int chunkCount, uint occurrences)
            {
                base.SetupChunk(chunkCount, occurrences);
                Reset();
            }

            private void Reset()
            {
                position = -1;
                nextPosition = 0;

                CurrentOccurrenceId = -1;
                nextOccurrenceId = 0;
            }
        }

        public override AbstractDocumentChunkEnumerator CreateEnumerator()
        {
            return new ChunkEnumerator();
        }
    }
}
