﻿namespace NETing.Index.Controller
{
    using Configurations;

    using Documents.DocumentModel.Scope;
    using Documents.Parser;
    
    using Services;
    
    using Storage.Buffer;

    public class IndexController
    {
        public ScopeDirectoryService ScopeDirectoryService { get; private set; }

        public StorageProfile StorageProfile { get; private set; }

        public IBufferManager BufferManager { get; private set; }

        internal TermStorage TermStorage { get; private set; }

        public DictionaryService DictionaryService { get; private set; }

        private readonly IPostingFileConfig postingFileConfig;
        private readonly IIndexSystemConfig systemConfig;

        private QueryService queryService;
        public QueryService QueryService
        {
            get
            {
                if (queryService == null)
                {
                    queryService = PrepareQuery();
                }

                return queryService;
            }
        }

        public IndexController(
            StorageProfile storageProfile,
            IIndexSystemConfig systemConfig,
            IPostingFileConfig postingFileConfig)
            : this(new ScopeDirectory(), storageProfile, systemConfig, postingFileConfig)
        {
        }

        public IndexController(
            ScopeDirectory scopeDirectory,
            StorageProfile storageProfile,
            IIndexSystemConfig systemConfig,
            IPostingFileConfig postingFileConfig)
        {
            ScopeDirectoryService = new ScopeDirectoryService(
                scopeDirectory, storageProfile);
            TermStorage = new TermStorage();
            StorageProfile = storageProfile;

            EnsureStorageExists();

            DictionaryService = new DictionaryService(StorageProfile);
            this.postingFileConfig = postingFileConfig;
            this.systemConfig = systemConfig;

            if (systemConfig.BufferSlices > 1)
            {
                BufferManager = new SlicedBufferManager(
                    systemConfig.BlockSize,
                    systemConfig.BufferCount,
                    systemConfig.BufferSlices,
                    systemConfig.BuffersReadOnly,
                    StorageProfile.PostingFile);
            }
            else
            {
                BufferManager = new BufferManager(
                    systemConfig.BlockSize,
                    systemConfig.BufferCount,
                    systemConfig.BuffersReadOnly,
                    StorageProfile.PostingFile);
            }
        }

        public BuilderService PrepareBuild(IBuilderServiceConfig builderConfig)
        {
            EnsureStorageExists();
            return new BuilderService(
                builderConfig,
                postingFileConfig,
                DictionaryService, 
                StorageProfile,
                ScopeDirectoryService,
                BufferManager,
                TermStorage);
        }

        public QueryService PrepareQuery()
        {
            return new QueryService(
                BufferManager,
                ScopeDirectoryService,
                DictionaryService,
                postingFileConfig);
        }

        public void Shutdown()
        {
            BufferManager.Dispose();
            BufferManager = null;
            ScopeDirectoryService = null;
            DictionaryService = null;
            
            if (queryService != null)
            {
                queryService.Dispose();
            }

            System.GC.Collect();
            System.GC.WaitForFullGCComplete();
        }

        public void Restart()
        {
            BufferManager.Dispose();
            BufferManager = new BufferManager(
                systemConfig.BlockSize,
                systemConfig.BufferCount,
                systemConfig.BuffersReadOnly,
                StorageProfile.PostingFile);
        }

        private void EnsureStorageExists()
        {
            if (!StorageProfile.DictionaryFile.Directory.Exists)
            {
                StorageProfile.DictionaryFile.Directory.Create();
            }

            if (!StorageProfile.ScopeDirectoryFile.Directory.Exists)
            {
                StorageProfile.ScopeDirectoryFile.Directory.Create();
            }

            if (!StorageProfile.PostingFile.Directory.Exists)
            {
                StorageProfile.PostingFile.Directory.Create();
            }
        }
    }
}
