﻿namespace NETing.Index.Controller.Services
{
    using System;
    using System.Collections.Concurrent;
    using System.Collections.Generic;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using System.Xml;

    using Configurations;
    
    using Documents.DocumentModel;
    using Documents.Parser;
    
    using FileProcessor;
    
    using Inversion;
    
    using PostingFile;
    
    using Storage.Buffer;
    
    using Utils.Collections;
    
    public sealed class BuilderService : IDisposable
    {
        private sealed class BuilderServiceMutex {}

        private readonly BuilderServiceMutex mutex = new BuilderServiceMutex();

        private DocumentInverter documentInverter;
        private DocumentParser parser;

        private readonly IBufferManager bufferManager;
        private readonly TermStorage termStorage;

        private readonly DictionaryService dictionaryService;
        private readonly ScopeDirectoryService scopeDirectoryService;

        private readonly IPostingFileConfig postingFileConfig;

        internal ScalingInMemoryInverter Inverter { get; set; }
        private readonly IBuilderServiceConfig config;

        private const int MaxDocumentQueueCapacity = 10000;
        private readonly BlockingCollection<Task<Document>> documents 
            = new BlockingCollection<Task<Document>>(
                new ConcurrentQueue<Task<Document>>(),
                MaxDocumentQueueCapacity);

        internal BuilderService(
            IBuilderServiceConfig config, 
            IPostingFileConfig postingFileConfig,
            DictionaryService dictionaryService,
            StorageProfile storageProfile,
            ScopeDirectoryService scopeDirectoryService,
            IBufferManager bufferManager,
            TermStorage termStorage)
        {
            this.dictionaryService = dictionaryService;
            this.bufferManager = bufferManager;
            this.scopeDirectoryService = scopeDirectoryService;

            this.config = config;
            this.termStorage = termStorage;

            this.postingFileConfig = postingFileConfig;
            this.postingFileConfig.ScopeDirectory = scopeDirectoryService.ScopeDirectory;
        }

        private void ProcessDocuments(object feed)
        {
            foreach (Document document in documents.GetConsumingEnumerable().Select(tsk => tsk.Result))            {
                try
                {
                    using (var occs = documentInverter.Invert(parser.ParseDocument(document.Content)))
                    {
                        while (occs.MoveNext())
                        {
                            Add(occs.Current.First, document.DocumentInfo.DocumentId, occs.Current.Second);
                        }
                    }
                }
                catch (XmlException e)
                {
                    Console.WriteLine("ERROR: Malformed document: " + document.DocumentInfo.DocumentId);
                    Console.WriteLine(">>>>>: " + e.Message);
                }
            }

            lock (mutex)
            {
                Monitor.PulseAll(mutex);
            }
        }

        private void Add(Term term, uint documentId, InvertedDocumentOccurrence docOccs)
        {
            Inverter.AddDocumentOccurences(term, documentId, docOccs);
        }

        public void Add(Document document)
        {
            documents.Add(DocumentLoader.LoadDocument(document));
        }

        public void Start()
        {
            parser = new DocumentParser(scopeDirectoryService.ScopeDirectory);
            documentInverter = new DocumentInverter(termStorage);

            Inverter = new ScalingInMemoryInverter(
                config.StoreSize, config.InitialListSize, scopeDirectoryService.ScopeDirectory);

            ThreadPool.QueueUserWorkItem(ProcessDocuments);
        }

        public void PrepareFinish()
        {
            lock (mutex)
            {
                documents.CompleteAdding();
                while (!documents.IsCompleted)
                {
                    Monitor.Wait(mutex);
                }    
            }
        }

        public void Finish()
        {
            var postingFileWriter = new PostingFileWriter(
                bufferManager,
                postingFileConfig.DocumentIdWriter,
                postingFileConfig.FrequencyWriter,
                postingFileConfig.PositionWriter,
                postingFileConfig.ScopeWriter);
            
            foreach (var term in Inverter.ReadInvertedIndex())
            {
                long position = postingFileWriter.NextTerm(term.First);
                int documentFrequency = 0;
                using (term.Second)
                {
                    while (term.Second.MoveNext())
                    {
                        documentFrequency++;
                        postingFileWriter.Add(term.Second.Current);
                    }
                }


                dictionaryService.Register(term.First, position, documentFrequency);
            }

            postingFileWriter.Finish();
            bufferManager.Force();
        }
        

        public void Write()
        {
            Parallel.Invoke(
                dictionaryService.Write,
                scopeDirectoryService.Write);
        }

        internal IEnumerator<Tuple<Term, IEnumerator<Posting>>> ReadInMemoryInversion()
        {
            foreach (Pair<Term, IEnumerator<InvertedDocumentOccurrence>> pair in Inverter.ReadInvertedIndex())
            {
                yield return new Tuple<Term, IEnumerator<Posting>>(
                    pair.First, UnwindDocumentOccurrences(pair.Second));
            }
        }

        private IEnumerator<Posting> UnwindDocumentOccurrences(IEnumerator<InvertedDocumentOccurrence> docOccs)
        {
            var holder = new Posting();
            while (docOccs.MoveNext())
            {
                InvertedDocumentOccurrence docOcc = docOccs.Current;
                holder.DocumentId = docOcc.DocumentId;
                holder.Frequency = (ushort)docOcc.Frequency;
                for (int i = 0; i < docOcc.Frequency; i++)
                {
                    holder.Position = docOcc.Positions[i];
                    holder.Scope = docOcc.ScopeAt(i, scopeDirectoryService.ScopeDirectory);
                    yield return holder;
                }
            }
        }

        public void Dispose()
        {
            Inverter = null;
        }
    }
}


