﻿namespace NETing.Index.Controller.Services
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;
    
    using Configurations;
    
    using Documents.DocumentModel;
    using Documents.DocumentModel.Scope;
    using Documents.Parser;
    
    using PathMatching;
    
    using PostingFile;
    
    using QueryLanguage;
    using Search;
    using Search.ScopeIntersection;
    using Storage.Buffer;

    public class QueryService : IDisposable
    {
        private readonly DictionaryService dictionaryService;
        private readonly ScopeDirectoryService scopeDirectoryService;
        private readonly PostingFileReader postingFileReader;

        private readonly QueryParser queryParser;

        public QueryService(
            IBufferManager bufferManager,
            ScopeDirectoryService scopeDirectoryService,
            DictionaryService dictionaryService,
            IPostingFileConfig config)
        {
            this.dictionaryService = dictionaryService;
            this.scopeDirectoryService = scopeDirectoryService;
            
            dictionaryService.Read();
            scopeDirectoryService.Read();

            queryParser = new QueryParser(new PathMatcher(scopeDirectoryService.ScopeDirectory));

            postingFileReader = new PostingFileReader(
                bufferManager,
                scopeDirectoryService.ScopeDirectory,
                config.DocumentReader,
                config.PositionReader,
                config.ScopeReader);
        }

        internal IEnumerator<object> Query(Query query)
        {
            if (query.Terms.Count() > 1)
            {
                var enumerators
                    = new PostingFileEnumerator[query.Terms.Count()];
                int pos = 0;
                foreach (Term term in query.Terms)
                {
                    enumerators[pos++] = Lookup(term);
                    if (enumerators[pos - 1] == null)
                    {
                        for (int i = 0; i < pos - 1; i++)
                        {
                            enumerators[i].Dispose();
                        }

                        return Enumerable.Empty<object>().GetEnumerator();
                    }
                }

                if (query.PathTypes != null)
                {
                    if (query.Dense)
                    {
                        return new DenseScopeIntersectionQueryItem(
                            scopeDirectoryService.ScopeDirectory, enumerators, query.PathTypes);
                    }

                    return new SparseScopeIntersectionQueryItem(
                        scopeDirectoryService.ScopeDirectory, enumerators, query.PathTypes);
                }

                return new IntersectionQueryItem(enumerators);         
            }

            if (query.PathTypes != null)
            {
                var results = Lookup(query.Terms.First(), query.PathTypes);
                if (results == null)
                {
                    return Enumerable.Empty<object>().GetEnumerator();
                }

                return results;
            }

            var res = Lookup(query.Terms.First());
            if (res == null)
            {
                return Enumerable.Empty<object>().GetEnumerator();
            }

            return res;
        }

        public IEnumerator<object> Query(string queryString)
        {
            return Query(queryParser.Parse(queryString));
        }

        public IEnumerator<object> Query(Term term)
        {
            return Lookup(term);
        }

        internal PostingFileEnumerator Lookup(Term term)
        {
            var entry = dictionaryService.Locate(term);
            if (entry == null)
            {
                return null;
            }

            return postingFileReader.Lookup(entry.Item1, entry.Item2);
        }

        private sealed class PathTypeFilterEnumerator : IEnumerator<Posting>
        {
            private readonly IEnumerator<Posting> baseEnumerator;
            private readonly BitArray filter;

            public PathTypeFilterEnumerator(
                IEnumerator<Posting> baseEnumerator, 
                IEnumerable<PathType> pathTypes,
                ScopeDirectory scopeDirectory)
            {
                this.baseEnumerator = baseEnumerator;
                filter = new BitArray(scopeDirectory.NumberOfPaths);
                foreach (PathType entry in pathTypes)
                {
                    filter.Set(entry.PathId, true);
                }
            }

            public void Dispose()
            {
                baseEnumerator.Dispose();
            }

            public bool MoveNext()
            {
                while (baseEnumerator.MoveNext())
                {
                    if (filter.Get(baseEnumerator.Current.Scope.PathType.PathId))
                    {
                        return true;
                    }
                }

                return false;
            }

            public void Reset()
            {
                baseEnumerator.Reset();
            }

            public Posting Current
            {
                get { return baseEnumerator.Current; }
            }

            object IEnumerator.Current
            {
                get { return Current; }
            }
        }

        internal IEnumerator<Posting> Lookup(Term term, IEnumerable<PathType> pathTypes)
        {
            var entry = dictionaryService.Locate(term);
            if (entry == null)
            {
                return Enumerable.Empty<Posting>().GetEnumerator();
            }

            return new PathTypeFilterEnumerator(
                postingFileReader.Lookup(entry.Item1, entry.Item2), 
                pathTypes,
                scopeDirectoryService.ScopeDirectory);
        }

        public void Dispose()
        {
            postingFileReader.Dispose();
        }
    }
}


