﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Text;
    using Utils.Compression;

    public class Dewey : IEnumerable<uint>
    {
        private struct DeweyEnumerator : IEnumerator<uint>
        {
            private int pos;
            private readonly int limit;
            private readonly ushort[] array;

            internal DeweyEnumerator(int count, int offset, ushort[] array)
            {
                pos = offset - 1;
                limit = count + offset;
                this.array = array;
            }

            public void Dispose()
            {
            }

            public bool MoveNext()
            {
                return ++pos < limit;
            }

            public void Reset()
            {
                pos = 0;
            }

            public uint Current
            {
                get { return array[pos]; }
            }

            object IEnumerator.Current
            {
                get { return Current; }
            }
        }

        public ushort[] DeweyArray { get; set; }

        public int Count;

        public int Offset;

        public Dewey()
        {
        }

        public Dewey(ushort[] dewey)
        {
            Offset = 0;
            DeweyArray = dewey;
            Count = dewey.Length;
        }

        public Dewey(ushort[] dewey, int count)
        {
            Offset = 0;
            DeweyArray = dewey;
            Count = count;
        }

        public Dewey(ushort[] dewey, int offset, int count)
        {
            DeweyArray = dewey;
            Count = count;
            Offset = offset;
        }

        public ushort this[int pos]
        {
            get { return DeweyArray[Offset + pos]; }
            set { DeweyArray[Offset + pos] = value; }
        }

        #region IEnumerable<int> Members

        public IEnumerator<uint> GetEnumerator()
        {
            return new DeweyEnumerator(Count, Offset, DeweyArray);
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        public override string ToString()
        {
            var sb = new StringBuilder();
            foreach (uint i in this)
            {
                sb.Append(i).Append(".");
            }

            return sb.ToString();
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj))
            {
                return false;
            }

            if (ReferenceEquals(this, obj))
            {
                return true;
            }

            if (obj.GetType() != typeof(Dewey))
            {
                return false;
            }

            return Equals((Dewey)obj);
        }

        public bool Equals(Dewey other)
        {
            if (ReferenceEquals(null, other))
            {
                return false;
            }

            if (ReferenceEquals(this, other))
            {
                return true;
            }

            if (other.Count == Count)
            {   // Start comparing in the middle of the dewey, this is where the most selectivity is
                int center = Count >> 1;
                for (int i = center; i < Count; i++)
                {
                    if (this[i] != other[i])
                    {
                        return false;
                    }
                }

                for (int i = 0; i < center; i++)
                {
                    if (this[i] != other[i])
                    {
                        return false;
                    }
                }

                return true;
            }

            return false;
        }

        public int PrefixCompare(Dewey other, int prefix)
        {
            int limit = Utils.Min(prefix, Utils.Min(other.Count, Count));
            int i;
            int cmp = 0;
            for (i = 0; i < limit && cmp == 0; i++)
            {
                cmp = this[i] - other[i];
            }

            return cmp;
        }

        public bool PrefixEqual(Dewey other, int prefix)
        {
            for (int i = 0; i < prefix; i++)
            {
                if (this[i] != other[i])
                {
                    return false;
                }
            }

            return true;
        }

        public override int GetHashCode()
        {
            uint hash = 0;
            unchecked
            {
                foreach (var i in this)
                {
                    hash = 31 * hash + i;
                }   
            }

            return (int)hash;
        }

        public Dewey Duplicate()
        {
            var newArray = new ushort[Count];
            Buffer.BlockCopy(DeweyArray, Offset * 2, newArray, 0, Count * 2);
            return new Dewey(newArray);
        }

        public Dewey Prefix(int prefixLength)
        {
            var newArray = new ushort[prefixLength];
            Buffer.BlockCopy(DeweyArray, Offset << 1, newArray, 0, prefixLength << 1);
            return new Dewey(newArray);
        }
    }
}