﻿namespace NETing.Index.Documents.DocumentModel.Scope
{
    using System;
    using System.Collections.Generic;
    using System.Linq;

    [Serializable]
    public sealed class ScopeDirectory
    {
        /// <summary>
        /// Mapping from path type id number to path type objects. These paths are only identified
        /// by their ordinal position in this list.
        /// </summary>
        private PathType[] pathTypes = new PathType[32];
        private int pathTypeCount;
        
        /// <summary>
        /// Mapping from tag name, to tag object, and ordinal.
        /// </summary>
        private readonly IDictionary<Tag, Tag> tags = new Dictionary<Tag, Tag>();
        private Tag[] tagArray = new Tag[64];
        private readonly Tag tagHolder = new Tag();

        public ScopeDirectory()
        {
            tagHolder.Name = "";
            tagHolder.IsAttribute = false;
            pathTypes[0] = new PathType { PathId = 0, LastTag = InternTag(tagHolder) };
            pathTypeCount = 1;
        }

        /// <summary>
        /// Returns the number of distinct paths.
        /// </summary>
        public int NumberOfPaths
        {
            get { return pathTypeCount; }
        }

        public IDictionary<Tag, Tag> Tags
        {
            get { return tags; }
        }

        public Tag[] TagArray
        {
            get { return tagArray; }
        }

        public PathType this[int pathTypeId]
        {
            get { return pathTypes[pathTypeId]; }
        }

        public PathType[] PathTypes
        {
            get { return pathTypes; }
        }

        public PathType GetPath(params string[] path)
        {
            return GetPath(path.Select(GetTag));
        }

        public Tag GetTag(string tag)
        {
            if (tag[0] == '@')
            {
                tagHolder.IsAttribute = true;
                tagHolder.Name = tag.Substring(1);
            }
            else
            {
                tagHolder.IsAttribute = false;
                tagHolder.Name = tag;
            }

            return tags[tagHolder];
        }

        private PathType GetPath(IEnumerable<Tag> path)
        {
            IEnumerator<Tag> pathEnum = path.GetEnumerator();
            PathType currentPath = pathTypes[0];
            while (pathEnum.MoveNext())
            {
                currentPath = currentPath.Enter(pathEnum.Current);
            }

            return currentPath;
        }

        internal Cursor GetCursor()
        {
            return new Cursor(this);
        }

        internal sealed class Cursor
        {
            private readonly Tag tagHolder = new Tag();
            private readonly ScopeDirectory directory;

            public Cursor(ScopeDirectory scopeDirectory)
            {
                directory = scopeDirectory;
                DeweyBuilder = new DeweyBuilder();
                CurrentPathType = scopeDirectory[0];
                CurrentPathDewey = new Dewey(new ushort[10], 0);
                CurrentScope = new Scope(null, CurrentPathDewey);
            }

            private PathType CurrentPathType { get; set; }

            private Dewey CurrentPathDewey { get; set; }

            private DeweyBuilder DeweyBuilder { get; set; }

            public Scope CurrentScope { get; private set; }

            public void UpdateCurrentScope()
            {
                DeweyBuilder.Build(CurrentPathDewey);
                CurrentScope.PathType = CurrentPathType;
            }

            public void Enter(string tag)
            {
                tagHolder.Name = tag;
                tagHolder.IsAttribute = false;
                Enter(directory.InternTag(tagHolder));
            }

            private void Enter(Tag tag)
            {
                DeweyBuilder.Enter();
                CurrentPathType = directory.Enter(CurrentPathType, tag);
            }

            public void Exit()
            {
                DeweyBuilder.Exit();
                CurrentPathType = ScopeDirectory.Exit(CurrentPathType);
            }

            public void AttributeEnter(string name)
            {
                tagHolder.Name = name;
                tagHolder.IsAttribute = true;
                AttributeEnter(directory.InternTag(tagHolder));
            }

            private void AttributeEnter(Tag tag)
            {
                DeweyBuilder.Enter();
                CurrentPathType = directory.Enter(CurrentPathType, tag);
            }

            public void AttributeExit()
            {
                DeweyBuilder.RevertingExit();
                CurrentPathType = ScopeDirectory.Exit(CurrentPathType);
            }

            public void TextEnter()
            {
                DeweyBuilder.Enter();
            }

            public void TextExit()
            {
                DeweyBuilder.Exit();
            }

            public void Clear()
            {
                DeweyBuilder.Clear();
                CurrentPathType = directory[0];
                CurrentPathDewey.Count = 0;
            }
        }

        private Tag InternTag(Tag tag)
        {
            Tag tagObj;
            if (!tags.TryGetValue(tag, out tagObj))
            {
                tagObj = new Tag { Name = tag.Name, IsAttribute = tag.IsAttribute, TagId = tags.Count };
                AddTag(tagObj);
            }

            return tagObj;
        }

        private void AddTag(Tag tag)
        {
            tags.Add(tag, tag);
            if (tagArray.Length < tag.TagId)
            {
                Array.Resize(ref tagArray, tag.TagId * 2);
            }

            tagArray[tag.TagId] = tag;
        }

        #region PathType management

        private PathType Enter(Tag tag)
        {
            return Enter(pathTypes[0], tag);
        }

        private PathType Enter(PathType path, Tag tag)
        {
            PathType child = path.Enter(tag);
            if (child.PathId == -1)
            {
                child.PathId = pathTypeCount;
                if (pathTypes.Length < pathTypeCount + 1)
                {
                    Array.Resize(ref pathTypes, pathTypes.Length * 2);
                }

                pathTypes[pathTypeCount++] = child;
            }

            return child;
        }

        private static PathType Exit(PathType path)
        {
            return path.Parent;
        }

        #endregion
    }
}