﻿namespace NETing.Index.Documents.Parser
{
    using System.Collections.Generic;
    using System.IO;
    using System.Xml;
    using DocumentModel;
    using DocumentModel.Scope;

    internal sealed class DocumentParser
    {
        // 1 MB buffer is long enough for all texts in test collections
        private const int TextBufferSize = 1024 * 1024; 

        private readonly ScopeDirectory.Cursor cursor;

        private readonly Tokenizer tokenizer = new Tokenizer();
        private readonly char[] valueChunkBuffer = new char[TextBufferSize];
        private readonly XmlReaderSettings readerSettings;

        private readonly Token tokenHolder = new Token();

        public DocumentParser(ScopeDirectory scopeDirectory)
        {
            cursor = scopeDirectory.GetCursor();
            readerSettings = new XmlReaderSettings();
            readerSettings.NameTable = new NameTable();
            readerSettings.CloseInput = true;
            readerSettings.IgnoreWhitespace = true;
            readerSettings.CheckCharacters = true;
        }

        public IEnumerator<Token> ParseDocument(Stream dataReader)
        {
            using (var reader = XmlReader.Create(dataReader, readerSettings))
            {
                cursor.Clear();
                ushort pos = 0;
                XmlNodeType previousNodeType = XmlNodeType.Document;
                while (reader.Read())
                {
                    int length;
                    switch (reader.NodeType)
                    {
                        case XmlNodeType.Element:
                            if (previousNodeType == XmlNodeType.Text || previousNodeType == XmlNodeType.CDATA)
                            {
                                cursor.TextExit();
                            }

                            bool exitAfterAttributes = reader.IsEmptyElement;
                            cursor.Enter(reader.Name);
                            while (reader.MoveToNextAttribute())
                            {
                                cursor.AttributeEnter(reader.Name);
                                cursor.UpdateCurrentScope();
                                tokenHolder.Scope = cursor.CurrentScope;
                                length = reader.ReadValueChunk(
                                    valueChunkBuffer, 0, valueChunkBuffer.Length);
                                tokenizer.Initialize(valueChunkBuffer, length);
                                while (tokenizer.MoveNext())
                                {
                                    tokenHolder.Position = pos++;
                                    tokenHolder.Term = tokenizer.Current;
                                    yield return tokenHolder;
                                }

                                cursor.AttributeExit();
                            }

                            if (exitAfterAttributes)
                            {
                                cursor.Exit();
                            }

                            break;
                        case XmlNodeType.EndElement:
                            if (previousNodeType == XmlNodeType.Text
                                || previousNodeType == XmlNodeType.CDATA)
                            {
                                cursor.TextExit();
                            }

                            cursor.Exit();
                            break;
                        case XmlNodeType.Text:
                        case XmlNodeType.CDATA:
                            if (previousNodeType != XmlNodeType.Text
                                && previousNodeType != XmlNodeType.CDATA)
                            {
                                cursor.TextEnter();
                                cursor.UpdateCurrentScope();
                                tokenHolder.Scope = cursor.CurrentScope;
                            }

                            length = reader.ReadValueChunk(valueChunkBuffer, 0, valueChunkBuffer.Length);
                            tokenizer.Initialize(valueChunkBuffer, length);
                            while (tokenizer.MoveNext())
                            {
                                tokenHolder.Position = pos++;
                                tokenHolder.Term = tokenizer.Current;
                                yield return tokenHolder;
                            }

                            break;
                    }

                    previousNodeType = reader.NodeType;
                }
            }
        }
    }
}


