﻿namespace NETing.Index.Inversion
{
    using System;
    using System.Collections.Generic;
    using Documents.DocumentModel;
    using Documents.DocumentModel.Scope;
    using Documents.Parser;
    using Utils.ByteArrays;
    using Utils.Collections;

    /// <summary>
    /// Invertes a document.
    /// </summary>
    internal class DocumentInverter
    {
        private const int EstimateMaxDeweyLength = 8;
        private const int InitialTermFrequecy = 128;
        private const int InitialTermSlots = 512;

        private readonly Pair<Term, InvertedDocumentOccurrence> inversionHolder;

        private readonly InvertedDocumentOccurrence invertedDocOcc = new InvertedDocumentOccurrence();

        private readonly Dictionary<Term, int> lexicon = new Dictionary<Term, int>(InitialTermSlots << 3);
        private int index;
        private int[][] append = new int[InitialTermSlots][];
        private ushort[][] pathTypes = new ushort[InitialTermSlots][];
        private ushort[][] positions = new ushort[InitialTermSlots][];
        private ushort[][] deweys = new ushort[InitialTermSlots][];

        private readonly List<Term> traversalList = new List<Term>(1024);

        private readonly TermStorage termStorage;

        public DocumentInverter(TermStorage termStorage)
        {
            inversionHolder = new Pair<Term, InvertedDocumentOccurrence>(null, invertedDocOcc);
            this.termStorage = termStorage;
        }

        private void EnsureCapacity(int requiredCapacity)
        {
            if (requiredCapacity >= append.Length)
            {
                Array.Resize(ref append, requiredCapacity * 2);
                Array.Resize(ref positions, requiredCapacity * 2);
                Array.Resize(ref pathTypes, requiredCapacity * 2);
                Array.Resize(ref deweys, requiredCapacity * 2);
            }
        }

        public IEnumerator<Pair<Term, InvertedDocumentOccurrence>> Invert(IEnumerator<Token> document)
        {
            index = 0;
            traversalList.Clear();
            lexicon.Clear();
            while (document.MoveNext())
            {
                int list;
                if (!lexicon.TryGetValue(document.Current.Term, out list))
                {
                    Term interned = termStorage.Intern(document.Current.Term);
                    EnsureCapacity(index);
                    traversalList.Add(interned);
                    
                    lexicon.Add(interned, index);
                    interned.InvertedPos = index;

                    if (append[index] == null)
                    {
                        append[index] = new[] { 0, 0 };    
                    }
                    else
                    {
                        append[index][0] = 0;
                        append[index][1] = 0;
                    }

                    list = index++;
                }

                Dewey dewey = document.Current.Scope.Dewey;
                EnsureOccurenceCapacity(list, dewey);
                int[] appendAt = append[list];
                positions[list][appendAt[0]] = document.Current.Position;
                pathTypes[list][appendAt[0]++] = (ushort)document.Current.Scope.PathType.PathId;
                
                BlockCopy.Copy(
                    dewey.DeweyArray, dewey.Offset << 1, deweys[list], appendAt[1] << 1, dewey.Count << 1);

                appendAt[1] += dewey.Count;
            }

            foreach (Term term in traversalList)
            {
                inversionHolder.First = term;
                int list = term.InvertedPos;
                invertedDocOcc.Frequency = append[list][0];
                invertedDocOcc.Positions = positions[list];
                invertedDocOcc.Deweys = deweys[list];
                invertedDocOcc.PathTypes = pathTypes[list];
                invertedDocOcc.TotalDeweys = append[list][1];
                yield return inversionHolder;
            }
        }

        private void EnsureOccurenceCapacity(int list, Dewey actualDewey)
        {
            int[] appendAt = append[list];
            if (positions[list] == null)
            {
                positions[list] = new ushort[InitialTermFrequecy];
                pathTypes[list] = new ushort[InitialTermFrequecy];
                deweys[list] = new ushort[InitialTermFrequecy * EstimateMaxDeweyLength];
            }

            if (positions[list].Length <= appendAt[0])
            {
                Array.Resize(ref positions[list], (appendAt[0] + 1) * 2);
                Array.Resize(ref pathTypes[list], (appendAt[0] + 1) * 2);
            }

            if (deweys[list].Length <= (appendAt[1] + actualDewey.Count))
            {
                Array.Resize(ref deweys[list], (appendAt[1] + actualDewey.Count + 1) * 2);
            }
        }
    }
}