﻿namespace NETing.Index.Inversion
{
    using System.Collections;
    using System.Collections.Generic;

    using Utils.ByteArrays;
    using Utils.Compression;

    internal sealed class InversionReader : IEnumerator<InvertedDocumentOccurrence>
    {
        private InvertedListRef currentList;
        private int position;
        private int localLimit;
        private int upperLimit;

        private byte[] readBuffer = new byte[1024];
        
        private readonly InvertedDocumentOccurrence holder = new InvertedDocumentOccurrence();

        public void Dispose()
        {
        }

        public bool MoveNext()
        {
            if (position < upperLimit || currentList.Overflow != null)
            {
                int docOccLength;
                if (SegmentRemainingBytes >= 4)
                {
                    docOccLength = currentList.Store.Buffer.GetInt32At(position);
                    position += 4;
                }
                else
                {
                    ReadListData(0, 4);
                    docOccLength = readBuffer.GetInt32At(0);
                }

                if (SegmentRemainingBytes >= docOccLength)
                {
                    holder.DeserializeFrom(currentList.Store.Buffer, position);
                    position += docOccLength;
                }
                else
                {
                    ReadListData(0, docOccLength);
                    holder.DeserializeFrom(readBuffer, 0);
                }
                return true;
            }

            return false;
        }

        private int SegmentRemainingBytes
        {
            get { return localLimit - position;  }
        }

        private void ReadListData(int offset, int length)
        {
            if (readBuffer.Length < offset + length)
            {   // resize read buffer
                readBuffer = new byte[(offset + length) * 2];   
            }

            if (SegmentRemainingBytes >= length)
            {   // simple within segment copy
                ReadFromSegment(readBuffer, offset, length);
            }
            else
            {
                int loaded = SegmentRemainingBytes;
                ReadFromSegment(readBuffer, offset, SegmentRemainingBytes);
                while (loaded < length && position < upperLimit)
                {
                    MoveNextSegment();
                    int bytesToRead = Utils.Min(SegmentRemainingBytes, length - loaded);
                    ReadFromSegment(readBuffer, offset + loaded, bytesToRead);
                    loaded += bytesToRead;
                }

                if (loaded < length)
                {   // upper limit hit, continue in overflow
                    currentList = currentList.Overflow;
                    StartReadingList();
                    ReadListData(offset + loaded, length - loaded);
                }
            }
        }

        private void ReadFromSegment(byte[] buffer, int offset, int count)
        {
            BlockCopy.Copy(currentList.Store.Buffer,position, buffer, offset, count);
            position += count;
        }

        public void Reset()
        {
        }

        public InvertedDocumentOccurrence Current
        {
            get { return holder; }
        }

        object IEnumerator.Current
        {
            get { return Current; }
        }

        public void Read(InvertedListRef listHead)
        {
            currentList = listHead;
            StartReadingList();
        }

        private void StartReadingList()
        {
            InitializeSegment(currentList.FirstPosition);
            upperLimit = currentList.Append;
        }

        private void MoveNextSegment()
        {
            int nextSegment = currentList.Store.Buffer.GetInt32At(localLimit);
            InitializeSegment(nextSegment);
        }

        private void InitializeSegment(int segmentStart)
        {
            position = segmentStart + 4;
            localLimit = currentList.Store.Buffer.GetInt32At(segmentStart);
        }
    }
}
