﻿namespace NETing.Index.Inversion
{
    using System;
    using System.Collections.Generic;
    
    using Documents.DocumentModel.Scope;
    using Documents.Parser;

    using Utils.Benchmark;
    using Utils.Collections;
    using Utils.Compression;

    /// <summary>
    /// Scaling is a joke, it does not scale very well, but it scales better that 
    /// one which uses a single byte array as storage.
    /// </summary>
    internal sealed class ScalingInMemoryInverter
    {
        private const int InitialReferences = 2048;
        
        internal const int SegmentOverhead = 4 + 4; // segment length + pointer

        private readonly IList<InversionStore> stores = new List<InversionStore>();
        private readonly int storeSize;
        private InversionStore activeStore;

        private int referenceIndex;
        private readonly IDictionary<int, int> lexicon = new Dictionary<int, int>();
        private InvertedListRef[] referencesHeads = new InvertedListRef[InitialReferences];
        private InvertedListRef[] referencesTails = new InvertedListRef[InitialReferences];
        
        private readonly List<Term> traversalList = new List<Term>(1 << 16);
        private readonly int initialListSize = 128;

        private readonly ScopeDirectory scopeDirectory;
        
        public ScalingInMemoryInverter(
            int storeSize, int initialListSize, ScopeDirectory scopeDirectory)
        {
            this.storeSize = storeSize;
            this.initialListSize = initialListSize + SegmentOverhead;
            this.scopeDirectory = scopeDirectory;
            AllocateNewStore();
        }

        private InversionStore AllocateNewStore()
        {
            activeStore = new InversionStore(storeSize, stores.Count);
            stores.Add(activeStore);
            Console.WriteLine("Store number " + activeStore.StoreIndex + " created!");
            return activeStore;
        }

        private InvertedListRef AllocateNewList(
            int listIndex,
            InversionStore store,
            int requiredSize)
        {
            var list = new InvertedListRef(store, requiredSize);
            referencesTails[listIndex] = list;
            return list;
        }

        private byte[] writeBuffer = new byte[2048];

        public void AddDocumentOccurences(Term term, uint documentId, InvertedDocumentOccurrence docOccs)
        {
            int listIndex;
            InvertedListRef entry = null;
            if (!lexicon.TryGetValue(term.Start, out listIndex))
            {
                traversalList.Add(term);
                if (referencesHeads.Length <= referenceIndex)
                {
                    Array.Resize(ref referencesHeads, referenceIndex * 2);
                    Array.Resize(ref referencesTails, referenceIndex * 2);
                }

                lexicon[term.Start] = referenceIndex;
                listIndex = referenceIndex++;
            }
            else
            {
                entry = referencesTails[listIndex];
            }

            int length = docOccs.SerializeTo(documentId, ref writeBuffer);
            if (entry == null)
            {
                InversionStore store = null;
                if (activeStore.Remaining >= length + SegmentOverhead)
                {
                    store = activeStore;
                }

                if (store == null)
                {   // need new store.
                    store = AllocateNewStore();
                }

                // allocate new list
                int size = Utils.Min(
                    Utils.Max(initialListSize, length) + SegmentOverhead,
                    activeStore.Remaining);
                entry = AllocateNewList(listIndex, store, size);
                referencesHeads[listIndex] = entry;
                entry.Write(writeBuffer, 0, length);
            }
            else
            {
                if (entry.Remaining < length)
                {   // no enough space, fill the list then allocate more.
                    int rest = length - entry.Remaining;
                    entry.Write(writeBuffer, 0, entry.Remaining);

                    if (entry.Store.Remaining < rest + SegmentOverhead)
                    {   // need to change store
                        InversionStore overflowStore 
                            = activeStore.Remaining >= rest + SegmentOverhead 
                            ? activeStore : AllocateNewStore();

                        InvertedListRef overflow = AllocateNewList(
                                    listIndex, overflowStore, rest + SegmentOverhead);
                        entry.Overflow = overflow;
                        overflow.Write(writeBuffer, length - rest, rest);
                    }
                    else
                    {   // allocate new segment
                        entry.Extend(rest);
                        entry.Write(writeBuffer, length - rest, rest);
                    }
                }
                else
                {   // Simple, just write into current list.
                    entry.Write(writeBuffer, 0, length);
                }
            }
        }

        public IEnumerable<Pair<Term, IEnumerator<InvertedDocumentOccurrence>>> ReadInvertedIndex()
        {
            var listReader = new InversionReader();
            var holder = new Pair<Term, IEnumerator<InvertedDocumentOccurrence>>(null, listReader);
            long stortStart = Timing.PerformanceCounter;
            traversalList.Sort();
            long sortElapset = Timing.PerformanceCounter - stortStart;
            Console.WriteLine("Sorting dictionary in: " 
                + (Timing.PerformanceCountToNanoSeconds(sortElapset) / 1e9) + " seconds.");
            foreach (var term in traversalList)
            {
                listReader.Read(referencesHeads[lexicon[term.Start]]);
                holder.First = term;
                yield return holder;
            }
        }
    }   
}
