﻿namespace NETing.Index.PathMatching
{
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using Documents.DocumentModel.Scope;
    using Syntax;

    public class PathMatcher
    {
        private readonly ScopeDirectory scopeDirectory;
        private readonly SyntaxChecker syntaxChecker;

        public PathMatcher(ScopeDirectory scopeDirectory)
        {
            this.scopeDirectory = scopeDirectory;
            syntaxChecker = new SyntaxChecker(scopeDirectory);
        }

        public IEnumerable<PathType> MatchingPaths(string query)
        {
            ExpressionToken[] tokens = syntaxChecker.Validate(Parser.ParseExpression(query)).ToArray();
            foreach (var pathType in scopeDirectory.PathTypes.Take(scopeDirectory.NumberOfPaths))
            {
                if (PathMatches(pathType, tokens))
                {
                    yield return pathType;
                }
            }
        }

        private bool PathMatches(PathType pathType, ExpressionToken[] query)
        {
            if (query.Length == 0)
            {
                return true;
            }

            Tag[] path = pathType.ToArray();
            if (query[0].Type == TokenType.Child)
            {   // absolute
                return MatchesSuffix(path, query, 0, 1);    
            }

            // relative
            bool match = false;
            int queryOffset = query[0].Type == TokenType.Descendant ? 1 : 0;
            for (int i = 0; !match && i < path.Length; i++)
            {
                match |= MatchesSuffix(path, query, i, queryOffset);
            }

            return match; 
        }

        private bool MatchesSuffix(
            Tag[] path, ExpressionToken[] query, int pathOffset, int queryOffset)
        {
            if (pathOffset >= path.Length)
            {
                return queryOffset >= query.Length;
            }

            if (queryOffset >= query.Length)
            {
                return true;
            }

            switch (query[queryOffset].Type)
            {
                case TokenType.Child:
                    return MatchesSuffix(path, query, pathOffset, queryOffset + 1);
                case TokenType.Descendant:
                    bool match = false;
                    int i = 0;
                    while (!match && pathOffset + i < path.Length)
                    {
                        match |= MatchesSuffix(path, query, pathOffset + i, queryOffset + 1);
                        i++;
                    }

                    return match;
                case TokenType.ElementTag:
                    if (path.Length <= pathOffset )
                    {
                        return false;
                    }

                    if (query[queryOffset].Value == "*")
                    {
                        return MatchesSuffix(path, query, pathOffset + 1, queryOffset + 1);
                    }

                    return path[pathOffset].Name.Equals(query[queryOffset].Value)
                           && MatchesSuffix(path, query, pathOffset + 1, queryOffset + 1);
                default:
                    return false;
            }
        }

        public static void PrintSyntax(TextWriter output)
        {
            output.WriteLine("Expression starts with: \"/\" absolute path.");
            output.WriteLine("\"/\": Child traversal.");
            output.WriteLine("\"//\": Decendant traversal.");
            output.WriteLine("\"*\": Matches any tag.");
            output.WriteLine("Example: /documents//author");
            output.WriteLine("Matches any author tag beneath the documents root node.");
        }
    }
}


