﻿namespace NETing.Index.PostingFile
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Diagnostics;
    
    using ColumnAccess.Readers;
    
    using Documents.DocumentModel;
    using Documents.DocumentModel.Scope;
    using Search;
    using Storage.Buffer;

    using Utils.Collections;
    using Utils.Compression;

    [DebuggerDisplay("Current = {Current}")]
    internal sealed class PostingFileEnumerator 
        : IEnumerator<Posting>, IComparable<PostingFileEnumerator>
    {
        private readonly PostingFileDataReader dataReader;
        private readonly PostingFileReader postingFileReader;
        private readonly IBufferManager bufferManager;
        
        private byte[] headerBuffer = new byte[1024];

        private readonly AbstractDocumentChunkEnumerator documentChunkEnumerator;
        private readonly AbstractPositionChunkEnumerator positionChunkEnumerator;
        private readonly AbstractScopeChunkEnumerator sourceScopeChunkEnumerator;
        private AbstractScopeChunkEnumerator scopeChunkEnumerator;

        private readonly Posting holder = new Posting();

        #region Posting file state

        private int remainingDocumentFrequency;
        
        private long currentChunkStart;
        private long nextChunkStart;

        private bool documentColumnReady;
        private bool positionColumnReady;
        private bool scopeColumnReady;

        private long columnHeadersStart;
        private long documentIdColumnStart;
        private long frequencyColumnStart;
        private long positionColumnStart;
        private long scopeColumnStart;

        private uint documentIdLength;
        private uint frequencyLength;
        private uint positionLength;
        private uint scopeLength;
        
        private int documentsInChunk;
        private uint chunkTotalOccurrences;

        private int documentSkipToOccId;

        #endregion
        
        public PostingFileEnumerator(
            PostingFileReader postingFileReader,
            PostingFileDataReader dataReader, 
            IBufferManager bufferManager,
            AbstractDocumentColumnReader documentColumnReader,
            AbstractPositionColumnReader positionColumnReader,
            AbstractScopeColumnReader scopeColumnReader)
        {
            this.postingFileReader = postingFileReader;
            this.dataReader = dataReader;
            this.bufferManager = bufferManager;

            documentChunkEnumerator = documentColumnReader.CreateEnumerator();
            positionChunkEnumerator = positionColumnReader.CreateEnumerator();
            sourceScopeChunkEnumerator = scopeColumnReader.CreateEnumerator();
        }

        internal void Initialize(int documentFrequency, long postingFileStart)
        {
            remainingDocumentFrequency = documentFrequency;
            currentChunkStart = postingFileStart;

            holder.DocumentId = 0;
            holder.Frequency = 0;
            holder.Position = 0;

            PrepareFirstPosting();
        }

        internal void FinalCleanup()
        {
        }

        #region Skip methods

        public MutableBitArray PathTypeMask 
        {
            set { sourceScopeChunkEnumerator.PathTypeMask = value; }
        }

        public bool SkipToDocumentId(uint documentId)
        {
            if (!documentColumnReady)
            {
                InitializeEnumerators(true, false, false);
            }

            int skipToOccId = documentChunkEnumerator.SkipTo(documentId);
            while (skipToOccId >= chunkTotalOccurrences)
            {
                if (!MoveNextChunk(true, false, false))
                {
                    return false;
                }

                skipToOccId = documentChunkEnumerator.SkipTo(documentId);
            }


            documentSkipToOccId = skipToOccId;
            holder.DocumentId = documentChunkEnumerator.Current;
            holder.Frequency = documentChunkEnumerator.Frequency;
            remainingDocumentOccurrences = holder.Frequency;

            chunkRemainingDocuments = documentChunkEnumerator.RemainingDocuments;
            return true;   
        }

        public void SynchronizeWithDocumentColumn()
        {
            if (!scopeColumnReady || !positionColumnReady)
            {
                InitializeEnumerators(true, true, true);
            }

            if (scopeChunkEnumerator.CurrentOccurrenceId < documentChunkEnumerator.CurrentOccurrenceId)
            {
                MoveNonDocumentIdColumns(documentSkipToOccId);
            }
        }

        public bool SkipToPathTypes(MutableBitArray candidateSet)
        {
            if (!scopeColumnReady)
            {
                InitializeEnumerators(false, false, true);
            }

            int skipToOccId = scopeChunkEnumerator.SkipToPathType(candidateSet);
            while (skipToOccId >= chunkTotalOccurrences)
            {
                if (!MoveNextChunk(false, false, true))
                {
                    return false;
                }

                skipToOccId = scopeChunkEnumerator.SkipToPathType(candidateSet);
            }

            if (!(documentColumnReady & positionColumnReady))
            {
                InitializeEnumerators(true, true, true);
            }

            MoveNonScopeColumns(skipToOccId);
            return true;
        }

        public bool SkipToPathIdAndDocumentId(MutableBitArray candidateSet, uint documentId)
        {
            if (SkipToPathTypes(candidateSet))
            {
                if (holder.DocumentId < documentId)
                {
                    if (SkipToDocumentId(documentId))
                    {
                        SynchronizeWithDocumentColumn();
                        return SkipToPathTypes(candidateSet);
                    }

                    return false;
                }

                return true;
            }
            
            return false;
        }

        public bool SkipToDocumentIdAndPathId(uint documentId, MutableBitArray candidateSet)
        {
            if (SkipToDocumentId(documentId))
            {
                SynchronizeWithDocumentColumn();
                return SkipToPathTypes(candidateSet);
            }
            
            return false;
        }

        public bool SkipToScope(
            Scope scope, int matchingPrefix, out bool foundMatch)
        {
            if (!scopeColumnReady)
            {
                InitializeEnumerators(false, false, true);
            }

            int skipToOccid = scopeChunkEnumerator.SkipToScope(
                scope, matchingPrefix, documentChunkEnumerator.LastOccurrenceInCurrent, out foundMatch);
            
            if (skipToOccid >= chunkTotalOccurrences)
            {
                return false;
            }

            if (!(documentColumnReady & positionColumnReady))
            {
                InitializeEnumerators(true, true, true);
            }

            MoveNonScopeColumns(skipToOccid);
            return true;
        }

        #endregion

        #region IEnumerator<Posting> memebers

        private uint remainingDocumentOccurrences;
        private int chunkRemainingDocuments;

        public bool MoveNext()
        {
            if (!(documentColumnReady & positionColumnReady & scopeColumnReady))
            {
                InitializeEnumerators(true, true, true);
            }

            if (remainingDocumentOccurrences <= 0)
            {
                if (documentChunkEnumerator.RemainingDocuments <= 0)
                {
                    if (MoveNextChunk(true, true, true))
                    {
                        return MoveNext();    
                    }

                    return false;
                }

                documentChunkEnumerator.MoveNext();
                holder.DocumentId = documentChunkEnumerator.Current;
                
                holder.Frequency = documentChunkEnumerator.Frequency;
                remainingDocumentOccurrences = holder.Frequency;

                positionChunkEnumerator.NewDocument(holder.Frequency);

                chunkRemainingDocuments--;
            }

            remainingDocumentOccurrences--;

            positionChunkEnumerator.MoveNext();
            scopeChunkEnumerator.MoveNext(); // holder updated
            holder.Position = positionChunkEnumerator.Current;
            return true;
        }

        public bool LoadResults(QueryResult result, int scopePrefix)
        {
            if (result.Positions == null || (remainingDocumentOccurrences + 1) > result.Positions.Length)
            {
                int startAt = result.Positions != null ? result.Positions.Length : 0;
                result.Positions = new ushort[(remainingDocumentOccurrences + 1) << 1];
                Array.Resize(ref result.Scopes, (int)(remainingDocumentOccurrences + 1) << 1);

                if (result.DeweyArray == null || result.DeweyArray.Length < (result.Scopes.Length * 9))
                {
                    result.DeweyArray = new ushort[result.Scopes.Length * 9];
                    for (int j = 0; j < startAt; j++)
                    {
                        result.Scopes[j].Dewey.DeweyArray = result.DeweyArray;
                    }
                }

                for (int i = startAt; i < result.Scopes.Length; i++)
                {
                    result.Scopes[i] = new Scope(null, new Dewey(result.DeweyArray));
                }
            }

            result.DocumentId = Current.DocumentId;
            result.Positions[0] = Current.Position;
            Scope firstScope = result.Scopes[0];
            firstScope.PathType = Current.Scope.PathType;
            firstScope.Dewey.Offset = 0;
            firstScope.Dewey.Count = firstScope.PathType.DeweyDepth;
            ushort[] deweyArray = result.DeweyArray;
            for (int i = 0; i < firstScope.Dewey.Count; i++)
            {
                deweyArray[i] = Current.Scope.Dewey[i];    
            }

            result.Count = 1;
            int offset = firstScope.PathType.DeweyDepth;

            
            bool hasMore = MoveNext();
            if (scopePrefix > 0)
            {
                
                while (hasMore && Current.DocumentId == result.DocumentId)
                {
                    if (!firstScope.Dewey.PrefixEqual(Current.Scope.Dewey, scopePrefix))
                    {
                        break;
                    }

                    result.Positions[result.Count] = Current.Position;
                    Scope scope = result.Scopes[result.Count++];
                    scope.PathType = Current.Scope.PathType;

                    for (int i = 0; i < scope.PathType.DeweyDepth; i++)
                    {
                        deweyArray[offset++] = Current.Scope.Dewey[i];
                    }

                    hasMore = MoveNext();
                }
            }
            else
            {


                while (hasMore && Current.DocumentId == result.DocumentId)
                {
                    result.Positions[result.Count] = Current.Position;
                    Scope scope = result.Scopes[result.Count++];
                    scope.PathType = Current.Scope.PathType;

                    for (int i = 0; i < scope.PathType.DeweyDepth; i++)
                    {
                        deweyArray[offset++] = Current.Scope.Dewey[i];
                    }


                    hasMore = MoveNext();
                }
            }

            result.Count = (ushort)result.Count;
            return hasMore;
        }

        public void Reset()
        {
            PrepareFirstPosting();
        }

        public Posting Current { get { return holder; } }
        object IEnumerator.Current { get { return Current; } }

        public int CompareTo(PostingFileEnumerator other)
        {
            return Current.CompareTo(other.Current);
        }

        public void Dispose()
        {
            dataReader.Dispose();
            PathTypeMask = null;
            postingFileReader.Finished(this);
        }

        #endregion

        #region Posting file decoding

        private void PrepareFirstPosting()
        {
            LoadListGlobalMetadata();
            PrepareChunk(false, false, false);
        }

        private bool MoveNextChunk(
            bool prepareDocumentColumn,
            bool preparePositionColumn,
            bool prepareScopeColumn)
        {
            remainingDocumentFrequency -= documentsInChunk;
            if (remainingDocumentFrequency > 0)
            {
                currentChunkStart = nextChunkStart;
                PrepareChunk(prepareDocumentColumn, preparePositionColumn, prepareScopeColumn);
                return true;
            }

            return false;
        }

        private void PrepareChunk(
            bool prepareDocumentColumn,
            bool preparePositionColumn,
            bool prepareScopeColumn)
        {
            // make sure header does not span buffer boundary
            long remaining = bufferManager.BlockSize - (currentChunkStart % bufferManager.BlockSize);
            if (remaining < Constants.MaxHeaderSize)
            {
                currentChunkStart += remaining;
            }

            uint headerBytes;
            BlockBuffer buffer = bufferManager.Pin(currentChunkStart, true);
            int offset = this.ReadChunkHeader(
                 buffer,
                 buffer.BufferOffset(currentChunkStart),
                 out headerBytes,
                 out chunkTotalOccurrences,
                 out documentIdLength,
                 out frequencyLength,
                 out positionLength,
                 out scopeLength);

            columnHeadersStart = buffer.FileOffset + offset;

            if (headerBytes > 0)
            {
                if (headerBuffer.Length < headerBytes)
                {
                    headerBuffer = new byte[headerBytes << 1];
                }

                dataReader.LoadData(headerBuffer, 0, (int)headerBytes, columnHeadersStart);
                int nextHeaderOffset = documentChunkEnumerator.ReadHeader(
                    headerBuffer, 0, chunkTotalOccurrences);
                nextHeaderOffset = positionChunkEnumerator.ReadHeader(
                    headerBuffer, nextHeaderOffset, chunkTotalOccurrences);
                sourceScopeChunkEnumerator.ReadHeader(
                    headerBuffer, nextHeaderOffset, chunkTotalOccurrences);
            }


            documentsInChunk = Utils.Min(128, remainingDocumentFrequency);
            
            bufferManager.UnPin(buffer);
            
            documentIdColumnStart = columnHeadersStart + headerBytes;
            frequencyColumnStart = documentIdColumnStart + documentIdLength;
            positionColumnStart = frequencyColumnStart + frequencyLength;
            scopeColumnStart = positionColumnStart + positionLength;
            
            chunkRemainingDocuments = documentsInChunk;
            remainingDocumentOccurrences = 0;

            nextChunkStart = scopeColumnStart + scopeLength;

            documentColumnReady = false;
            positionColumnReady = false;
            scopeColumnReady = false;

            InitializeEnumerators(prepareDocumentColumn, preparePositionColumn, prepareScopeColumn);
        }

        private void InitializeEnumerators(
            bool prepareDocumentColumn,
            bool preparePositionColumn,
            bool prepareScopeColumn)
        {
            if (prepareDocumentColumn & !documentColumnReady)
            {
                AbstractDocumentColumnReader.InitializeEnumerator(
                    documentChunkEnumerator,
                    dataReader,
                    documentIdColumnStart,
                    documentIdLength,
                    frequencyColumnStart,
                    frequencyLength,
                    documentsInChunk,
                    chunkTotalOccurrences);

                documentColumnReady = true;
            }

            if (preparePositionColumn & !positionColumnReady)
            {
                AbstractPositionColumnReader.InitializeEnumerator(
                    positionChunkEnumerator,
                    dataReader,
                    positionColumnStart,
                    positionLength,
                    chunkTotalOccurrences);

                positionColumnReady = true;
            }

            if (prepareScopeColumn & !scopeColumnReady)
            {
                scopeChunkEnumerator = AbstractScopeColumnReader.InitializeEnumerator(
                    sourceScopeChunkEnumerator,
                    dataReader,
                    scopeColumnStart,
                    scopeLength,
                    chunkTotalOccurrences);

                holder.Scope = scopeChunkEnumerator.Holder;
                scopeColumnReady = true;
            }
        }

        private void LoadListGlobalMetadata()
        {
        }

        #endregion
        
        #region OccurrenceId alignment

        private void MoveNonDocumentIdColumns(int occurrenceId)
        {            
            positionChunkEnumerator.MoveTo(occurrenceId, occurrenceId, holder.Frequency);
            holder.Position = positionChunkEnumerator.Current;
            scopeChunkEnumerator.MoveTo(occurrenceId);

            remainingDocumentOccurrences = holder.Frequency - 1;
        }

        private void MoveNonScopeColumns(int occurrenceId)
        {
            int documentOccId = documentChunkEnumerator.MoveToDocumentContaining(occurrenceId);
            holder.DocumentId = documentChunkEnumerator.Current;
            holder.Frequency = documentChunkEnumerator.Frequency;
            positionChunkEnumerator.MoveTo(documentOccId, occurrenceId, holder.Frequency);
            holder.Position = positionChunkEnumerator.Current;

            remainingDocumentOccurrences = (uint) (documentOccId + holder.Frequency - occurrenceId - 1);
        }

        #endregion
    }
}
