﻿namespace NETing.Index.PostingFile
{
    using System;
    using System.Collections.Concurrent;
    using System.Threading;
    
    using ColumnAccess.Readers;
    using Documents.DocumentModel.Scope;
    
    using Storage.Buffer;

    internal sealed class PostingFileReader : IDisposable
    {
        private const int MaxEnumerators = 100;

        private readonly AbstractDocumentColumnReader documentReader;
        private readonly AbstractPositionColumnReader positionReader;
        private readonly AbstractScopeColumnReader scopeReader;
        
        private readonly IBufferManager bufferManager;

        private int allocatedEnumerators;
        private readonly ConcurrentBag<PostingFileEnumerator> enumeratorPool
            = new ConcurrentBag<PostingFileEnumerator>();

        internal PostingFileReader(
            IBufferManager bufferManager,
            ScopeDirectory scopeDirectory,
            AbstractDocumentColumnReader documentReader,
            AbstractPositionColumnReader positionReader,
            AbstractScopeColumnReader scopeReader)
        {
            this.bufferManager = bufferManager;

            this.documentReader = documentReader;
            documentReader.ScopeDirectory = scopeDirectory;
            this.positionReader = positionReader;
            positionReader.ScopeDirectory = scopeDirectory;
            this.scopeReader = scopeReader;
            scopeReader.ScopeDirectory = scopeDirectory;
        }

        internal void Finished(PostingFileEnumerator enumerator)
        {
            if (allocatedEnumerators > MaxEnumerators)
            {
                enumerator.FinalCleanup();
            }

            enumeratorPool.Add(enumerator);
        }

        public void Dispose()
        {
            foreach (PostingFileEnumerator postingFileEnumerator in enumeratorPool)
            {
                postingFileEnumerator.FinalCleanup();
            }
        }

        public PostingFileEnumerator Lookup(long fileOffset, int documentFrequency)
        {
            PostingFileEnumerator enumerator;
            if (!enumeratorPool.TryTake(out enumerator))
            {
                if (allocatedEnumerators < MaxEnumerators)
                {
                    Interlocked.Increment(ref allocatedEnumerators);
                    enumerator = new PostingFileEnumerator(
                        this,
                        new PostingFileDataReader(bufferManager), 
                        bufferManager,
                        documentReader,
                        positionReader, 
                        scopeReader);
                }
                else
                {
                    while (!enumeratorPool.TryTake(out enumerator))
                    {
                        Thread.Sleep(50);
                    }
                }
            }

            enumerator.Initialize(documentFrequency, fileOffset);
            return enumerator;
        }
    }
}
