﻿namespace NETing.Index.PostingFile
{
    using System;
    
    using ColumnAccess.Writers;
    using Documents.Parser;
    
    using Inversion;
    using Storage.Buffer;
    using Utils.ByteArrays;
    using Utils.Compression;

    internal sealed class PostingFileWriter
    {
        private readonly AbstractDocumentIdColumnWriter documentIdWriter;
        private readonly AbstractFrequencyColumnWriter frequencyWriter;
        private readonly AbstractPositionColumnWriter positionColumnWriter;
        private readonly AbstractScopeColumnWriter scopeColumnWriter;

        private readonly IBufferManager bufferManager;

        private BlockBuffer currentBuffer;
        private byte[] chunkBuffer =  new byte[1024 * 1024 * 4];
        private byte[] headerBuffer =  new byte[1024 * 4];
        private long chunkPosition;

        private long totalSize;
        private long scopeColumnSize;

        private int chunkCount;

        private Term currentTerm;
        private uint totalTokens;

        public PostingFileWriter(
            IBufferManager bufferManager,
            AbstractDocumentIdColumnWriter documentIdWriter,
            AbstractFrequencyColumnWriter frequencyWriter,
            AbstractPositionColumnWriter positionColumnWriter,
            AbstractScopeColumnWriter scopeColumnWriter)
        {
            this.bufferManager = bufferManager;

            this.documentIdWriter = documentIdWriter;
            this.frequencyWriter = frequencyWriter;
            this.positionColumnWriter = positionColumnWriter;
            this.scopeColumnWriter = scopeColumnWriter;
        }

        internal long NextTerm(Term term)
        {
            if (currentTerm != null)
            {
                FinishTerm();
            }

            currentTerm = term;
            return chunkPosition;
        }

        internal void Add(InvertedDocumentOccurrence docOccs)
        {
            documentIdWriter.Add(docOccs.DocumentId);
            frequencyWriter.Add((ushort)(docOccs.Frequency));

            positionColumnWriter.Add(docOccs.Positions, docOccs.Frequency);
            scopeColumnWriter.Add(docOccs.PathTypes, docOccs.Deweys, docOccs.Frequency, docOccs.TotalDeweys);

            totalTokens += (uint)docOccs.Frequency;

            if (++chunkCount >= Constants.ChunkSize)
            {
                FinishChunk();    
            }
        }

        internal void Finish()
        {
            FinishTerm();
            if (currentBuffer != null)
            {
                bufferManager.UnPin(currentBuffer);
            }

            Console.WriteLine("$" + (scopeColumnSize / 1024) + "$\t$" + ((totalSize - scopeColumnSize) / 1024) + "$\t$" + (totalSize / 1024) + "$\\\\");
        }

        private void FinishTerm()
        {
            FinishChunk();
        }

        private void FinishChunk()
        {
            // Prepare data to be written.
            if (documentIdWriter.Count < 1)
            {
                return; // no data to be written.    
            }

            int maxBytes = documentIdWriter.MaxBytes + frequencyWriter.MaxBytes
                          + positionColumnWriter.MaxBytes + scopeColumnWriter.MaxBytes;

            if (maxBytes > chunkBuffer.Length)
            {
                Array.Resize(ref chunkBuffer, maxBytes << 1);    
            }

            int frequencyStart = documentIdWriter.Write(chunkBuffer, 0);
            int positionStart = frequencyWriter.Write(chunkBuffer, frequencyStart);
            int scopeStart = positionColumnWriter.Write(chunkBuffer, positionStart);
            int chunkEnd = scopeColumnWriter.Write(chunkBuffer, scopeStart);

            int maxHeader = documentIdWriter.MaxHeaderBytes
                            + frequencyWriter.MaxHeaderBytes
                            + positionColumnWriter.MaxHeaderBytes
                            + scopeColumnWriter.MaxHeaderBytes;

            int headerEnd = 0;
            if (maxHeader > 0)
            {
                if (headerBuffer.Length < maxHeader)
                {
                    headerBuffer = new byte[maxHeader << 1];
                }

                int nextHeaderStart = documentIdWriter.WriteHeader(headerBuffer, 0);
                nextHeaderStart = frequencyWriter.WriteHeader(headerBuffer, nextHeaderStart);
                nextHeaderStart = positionColumnWriter.WriteHeader(headerBuffer, nextHeaderStart);
                int scopeHeaderStart = nextHeaderStart;
                headerEnd = scopeColumnWriter.WriteHeader(headerBuffer, nextHeaderStart);
                scopeColumnSize += (headerEnd - scopeHeaderStart);
            }

            documentIdWriter.Reset();
            frequencyWriter.Reset();
            positionColumnWriter.Reset();
            scopeColumnWriter.Reset();

            // Write data to buffer(s).

            if (bufferManager.BlockSize - (chunkPosition % bufferManager.BlockSize) < Constants.MaxHeaderSize)
            {
                chunkPosition = ((chunkPosition / bufferManager.BlockSize) + 1) * bufferManager.BlockSize;
            }

            if (currentBuffer == null || 
                !(currentBuffer.FileOffset <= chunkPosition && currentBuffer.Remaining(chunkPosition) > 0))
            {
                if (currentBuffer != null)
                {
                    currentBuffer = bufferManager.ExchangeBuffers(currentBuffer, chunkPosition, false);
                }
                else
                {
                    currentBuffer = bufferManager.Pin(chunkPosition, false);
                }
            }

            int offset = this.WriteChunkHeader(
                currentBuffer,
                currentBuffer.BufferOffset(chunkPosition),
                headerEnd,
                totalTokens,
                frequencyStart,
                positionStart - frequencyStart,
                scopeStart - positionStart,
                chunkEnd - scopeStart);

            if (headerEnd > 0)
            {
                currentBuffer = WriteOutColumnHeaderData(currentBuffer, offset, headerEnd, out offset);
            }

            currentBuffer = WriteOutColumnData(currentBuffer, offset, chunkEnd, out offset);
            chunkPosition = currentBuffer.FileOffset + offset;
            
            chunkCount = 0;
            totalTokens = 0;

            scopeColumnSize += chunkEnd - scopeStart;
            totalSize = chunkPosition;
        }
        
        private BlockBuffer WriteOutColumnHeaderData(
            BlockBuffer buffer, int offset, int dataLength, out int nextByteOffset)
        {
            return WriteBytesToBuffers(buffer, offset, headerBuffer, dataLength, out nextByteOffset);
        }

        private BlockBuffer WriteOutColumnData(
            BlockBuffer buffer, int offset, int dataLength, out int nextByteOffset)
        {
            return WriteBytesToBuffers(buffer, offset, chunkBuffer, dataLength, out nextByteOffset);
        }

        private BlockBuffer WriteBytesToBuffers(
            BlockBuffer buffer,
            int offset, 
            byte[] data,
            int count,
            out int nextByteOffset)
        {
            long currentBufferFilePos = buffer.FileOffset;
            int written = 0;
            while (count - written > 0)
            {
                int bytes = Utils.Min(count - written, buffer.Remaining(offset));
                BlockCopy.Copy(data, written, buffer.Buffer, offset, bytes);
                buffer.IsDirty = true;
                written += bytes;
                offset += bytes;

                if (buffer.Remaining(offset) <= 0)
                {   // Swap buffers
                    currentBufferFilePos += offset;
                    offset = 0;
                    buffer = bufferManager.ExchangeBuffers(buffer, currentBufferFilePos, false);
                }
            }

            nextByteOffset = offset;
            return buffer;
        }
    }
}
