﻿namespace NETing.Index.Search.ScopeIntersection
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;

    using Documents.DocumentModel.Scope;
    
    using PostingFile;
    
    using Utils.Collections;
    using Utils.Compression;

    public abstract class AbstractScopeIntersectionQueryItem : IEnumerator<ScopeIntersectionResult>
    {
        private readonly ScopeDirectory scopeDirectory;

        protected readonly MutableBitArray candidatePathTypes;
        private readonly int[] queryPrefixOfCandidate;

        protected bool exhausted;
        protected readonly ScopeIntersectionResult holder
            = new ScopeIntersectionResult();

        internal readonly PostingFileEnumerator[] lists;
        internal readonly ListHolder[] listHolders;
        internal readonly BinaryMaxHeap<ListHolder> heap;

        internal AbstractScopeIntersectionQueryItem(
            ScopeDirectory scopeDirectory,
            PostingFileEnumerator[] lists,
            IEnumerable<PathType> queryPathTypes)
        {
            this.scopeDirectory = scopeDirectory;
            this.lists = lists;

            candidatePathTypes = CreateCandidatePathTypes(queryPathTypes);
            queryPrefixOfCandidate = CreateQueryPrefix(queryPathTypes);

            listHolders = new ListHolder[lists.Length];
            for (int i = 0; i < lists.Length; i++)
            {
                lists[i].PathTypeMask = candidatePathTypes;
                if (!lists[i].SkipToPathTypes(candidatePathTypes))
                {
                    exhausted = true;
                    return;
                }

                listHolders[i] = new ListHolder
                {
                    Enumerator = lists[i],
                    Index = i
                };
            }

            holder.Occurrences = new QueryResult[lists.Length];
            for (int i = 0; i < holder.Occurrences.Length; i++)
            {
                holder.Occurrences[i] = new QueryResult();
            }

            heap = new BinaryMaxHeap<ListHolder>(listHolders);
        }

        private int[] CreateQueryPrefix(
            IEnumerable<PathType> queryPathTypes)
        {
            int[] prefixes = new int[scopeDirectory.NumberOfPaths];
            int[] parentDepth = new int[scopeDirectory.NumberOfPaths];

            var queue = new Queue<PathType>(queryPathTypes.OrderBy(pt => pt.Depth));
            while (queue.Count > 0)
            {
                PathType pt = queue.Dequeue();
                if (prefixes[pt.PathId] < 1)
                {
                    if (parentDepth[pt.PathId] == 0)
                    {
                        parentDepth[pt.PathId] = pt.Depth;
                    }

                    prefixes[pt.PathId] = parentDepth[pt.PathId];
                    foreach (var child in pt.Children.Values)
                    {
                        if (prefixes[child.PathId] < 1)
                        {
                            if (parentDepth[child.PathId] == 0)
                            {
                                parentDepth[child.PathId] = parentDepth[pt.PathId];
                            }
                            else
                            {
                                parentDepth[child.PathId] = Utils.Min(
                                    parentDepth[pt.PathId], parentDepth[child.PathId]);
                            }

                            queue.Enqueue(child);
                        }
                    }
                }
            }

            return prefixes;
        }

        private static MutableBitArray CreateCandidatePathTypes(
            IEnumerable<PathType> queryPathTypes)
        {
            var queue = new Queue<PathType>(queryPathTypes);
            var array = new MutableBitArray();
            while (queue.Count > 0)
            {
                var pt = queue.Dequeue();
                if (!array[pt.PathId])
                {
                    array[pt.PathId] = true;
                    foreach (var child in pt.Children.Values)
                    {
                        if (!array[child.PathId])
                        {
                            queue.Enqueue(child);
                        }
                    }
                }
            }

            return array;
        }

        public void Dispose()
        {
            foreach (PostingFileEnumerator list in lists)
            {
                list.Dispose();
            }
        }

        public abstract bool MoveNext();

        protected bool FindTrueMatch()
        {
            while (!exhausted)
            {
                int i;
                var first = heap.Max.Enumerator.Current;
                int queryPerfix = queryPrefixOfCandidate[first.Scope.PathType.PathId];
                if (!candidatePathTypes[first.Scope.PathType.PathId] && !heap.Max.Enumerator.SkipToPathTypes(candidatePathTypes))
                {
                    exhausted = true;
                    return false;
                }

                for (i = 1; i < listHolders.Length; i++)
                {
                    if (heap.Array[i].Enumerator.Current.DocumentId != first.DocumentId)
                    {
                        return false;
                    }

                    bool matches;
                    if (!heap.Array[i].Enumerator.SkipToScope(
                        first.Scope, queryPerfix, out matches))
                    {
                        exhausted = true;
                        return false;
                    }

                    if (!matches)
                    {
                        i = 0;
                        heap.Heapify();
                        break;
                    }
                }

                if (i == listHolders.Length)
                {
                    BuildResult(queryPerfix);
                    return true;
                }
            }

            return false;
        }

        private void BuildResult(int prefixLength)
        {
            holder.DocumentId = listHolders[0].Enumerator.Current.DocumentId;
            holder.Scope = listHolders[0].Enumerator.Current.Scope.Prefix(prefixLength);

            foreach (ListHolder list in listHolders)
            {
                exhausted |= !list.Enumerator.LoadResults(holder.Occurrences[list.Index], prefixLength);
            }
        }

        public void Reset()
        {
            throw new NotImplementedException();
        }

        public ScopeIntersectionResult Current
        {
            get { return holder; }
        }

        object IEnumerator.Current
        {
            get { return Current; }
        }
    }
}