﻿namespace NETing.Index.Storage.Buffer
{
    using System.Collections.Concurrent;
    using System.Collections.Generic;
    using System.IO;
    using System.Threading;
    using System.Threading.Tasks;

    public sealed class BufferManager : IBufferManager
    {
        private sealed class BufferManagerMutex { }
        private readonly BufferManagerMutex mutex = new BufferManagerMutex();

        public int BlockSize { get; private set; }

        private readonly int bufferCount;
        private readonly bool readOnly;

        private readonly IBlockReader reader;
        private readonly IBlockWriter writer;

        private readonly IDictionary<long, LinkedListNode<BlockBuffer>> buffers
            = new Dictionary<long, LinkedListNode<BlockBuffer>>();
        
        private readonly LinkedList<BlockBuffer> cleanQueue = new LinkedList<BlockBuffer>();

        private BlockingCollection<BlockBuffer> dirtyQueue;
        private BlockingCollection<BlockBuffer> cleanedQueue;
        private Task cleanTask;
        
        public BufferManager(
            int blockSize,
            int bufferCount,
            bool readOnly,
            FileInfo file)
        {
            BlockSize = blockSize;
            this.bufferCount = bufferCount;
            this.readOnly = readOnly;

            var blockAccess = new BlockAccessFile(file.Open(FileMode.OpenOrCreate), blockSize);
            reader = blockAccess;
            writer = blockAccess;

            StartFlushing();
        }

        public BufferManager(
            int blockSize,
            int bufferCount,
            bool readOnly,
            BlockAccessFile file)
        {
            BlockSize = blockSize;
            this.readOnly = readOnly;
            this.bufferCount = bufferCount;
            reader = file;
            writer = file;
            StartFlushing();
        }

        
        public BlockBuffer Pin(long fileOffset, bool read)
        {
            return Pin(new BufferKey(fileOffset / BlockSize), read);
        }

        internal BlockBuffer Pin(BufferKey key, bool read)
        {
            lock (mutex)
            {
                return UnsafePin(key, read);
            }
        }
        
        private BlockBuffer UnsafePin(BufferKey key, bool read)
        {
            LinkedListNode<BlockBuffer> existing;
            BlockBuffer buffer;
            if (buffers.TryGetValue(key.BlockNumber, out existing))
            {
                buffer = existing.Value;
                if (existing.List != null)
                {
                    existing.List.Remove(existing);
                }
            }
            else
            {
                if (buffers.Count < bufferCount)
                {
                    buffer = new BlockBuffer(BlockSize);
                    PrepareBuffer(key, buffer, read);
                }
                else if (cleanQueue.Count > 0)
                {
                    buffer = cleanQueue.Last.Value;
                    buffers.Remove(buffer.Key.BlockNumber);
                    cleanQueue.RemoveLast();
                    PrepareBuffer(key, buffer, read);
                }
                else
                {
                    while (cleanQueue.Count < 1)
                    {
                        Monitor.Wait(mutex);
                    }

                    buffer = cleanQueue.Last.Value;
                    buffers.Remove(buffer.Key.BlockNumber);
                    cleanQueue.RemoveLast();
                    PrepareBuffer(key, buffer, read);
                }
            }
            
            buffer.PinCount++;
            return buffer;
        }

        public void UnPin(BlockBuffer buffer)
        {
            lock (mutex)
            {
                UnsafeUnPin(buffer);
                Monitor.PulseAll(mutex);
            }
        }

        public BlockBuffer ExchangeBuffers(BlockBuffer buffer, long fileOffset, bool read)
        {
            lock (mutex)
            {
                UnsafeUnPin(buffer);
                return UnsafePin(new BufferKey(fileOffset / BlockSize), read);
            }
        }

        private void UnsafeUnPin(BlockBuffer buffer)
        {
            if (--buffer.PinCount == 0)
            {
                if (buffer.IsDirty)
                {
                    buffers.Remove(buffer.Key.BlockNumber);
                    dirtyQueue.Add(buffer);
                }
                else
                {
                    cleanQueue.AddLast(buffer);
                    buffers[buffer.Key.BlockNumber] = cleanQueue.Last;
                    Monitor.Pulse(mutex);
                }
            }
        }

        private void PrepareBuffer(BufferKey key, BlockBuffer buffer, bool read)
        {
            if (read)
            {
                lock (reader)
                {
                    reader.ReadBlock(key.BlockNumber, buffer.Buffer, 0);
                }
            }

            buffer.Key = key;
            buffer.IsDirty = false;
            buffers[key.BlockNumber] = new LinkedListNode<BlockBuffer>(buffer);
        }

        private void StartFlushing()
        {
            if (!readOnly)
            {
                dirtyQueue = new BlockingCollection<BlockBuffer>(bufferCount);
                cleanedQueue = new BlockingCollection<BlockBuffer>(bufferCount);
                cleanTask = Task.Factory.StartNew(CleanBuffers);
                ThreadPool.QueueUserWorkItem(FlushWorker);
            }
        }

        private void FlushWorker(object state)
        {
            FlushDirtyBuffers();
        }

        private void FlushDirtyBuffers()
        {
            foreach (var blockBuffer in dirtyQueue.GetConsumingEnumerable())
            {
                lock (writer)
                {
                    writer.WriteBlock(blockBuffer.Key.BlockNumber, blockBuffer.Buffer, 0);
                }

                cleanedQueue.Add(blockBuffer);
            }

            cleanedQueue.CompleteAdding();
        }

        private void CleanBuffers()
        {
            foreach (BlockBuffer buffer in cleanedQueue.GetConsumingEnumerable())
            {
                lock (mutex)
                {
                    cleanQueue.AddLast(buffer);
                    buffers[buffer.Key.BlockNumber] = cleanQueue.Last;
                }
            }
        }

        public void Dispose()
        {
            if (!readOnly)
            {
                dirtyQueue.CompleteAdding();
                cleanTask.Wait();
                dirtyQueue.Dispose();
                cleanedQueue.Dispose();
            }

            buffers.Clear();
            cleanQueue.Clear();
            
            reader.Dispose();
            writer.Dispose();
        }

        public void Force()
        {
            if (!readOnly)
            {
                dirtyQueue.CompleteAdding();
                cleanTask.Wait();

                StartFlushing();
            }
        }
    }
}