﻿namespace NETing.Index.Storage.Buffer
{
    using System;
    using System.IO;

    public class SlicedBufferManager : IBufferManager
    {
        private readonly BufferManager[] slices;

        public SlicedBufferManager(
            int blockSize,
            int bufferCount,
            int sliceCount,
            bool readOnly,
            FileInfo file)
        {
            BlockSize = blockSize;
            slices = new BufferManager[sliceCount];
            int sliceBufferCount = bufferCount / sliceCount;
            var blockAccess = new BlockAccessFile(file.Open(FileMode.OpenOrCreate), blockSize);
            for (int i = 0; i < sliceCount; i++)
            {
                slices[i] = new BufferManager(blockSize, sliceBufferCount, readOnly, blockAccess);
            }
        }

        public int BlockSize { get; private set; }
        
        public BlockBuffer Pin(long fileOffset, bool read)
        {
            var key = new BufferKey(fileOffset / BlockSize);
            return slices[key.BlockNumber % slices.Length].Pin(key, read);
        }

        public void UnPin(BlockBuffer buffer)
        {
            slices[buffer.Key.BlockNumber % slices.Length].UnPin(buffer);
        }

        public BlockBuffer ExchangeBuffers(BlockBuffer buffer, long fileOffset, bool read)
        {
            var currentSlice = (int)(buffer.Key.BlockNumber % slices.Length);
            var key = new BufferKey(fileOffset / BlockSize);
            var newSlice = (int) (key.BlockNumber % slices.Length);
            if (currentSlice == newSlice)
            {
                return slices[currentSlice].ExchangeBuffers(buffer, fileOffset, read);
            }

            slices[currentSlice].UnPin(buffer);
            return slices[newSlice].Pin(key, read);
        }

        public void Force()
        {
            foreach (BufferManager bufferManager in slices)
            {
                bufferManager.Force();
            }
        }

        public void Dispose()
        {
            foreach (BufferManager bufferManager in slices)
            {
                bufferManager.Dispose();
            }

            Array.Clear(slices, 0, slices.Length);
        }
    }
}
