﻿namespace NETing.Index.Utils.Compression.Patched
{
    /// <summary>
    /// NewPFor compression algorithm as described in 
    /// "Inverted Index Compression and Query Processing with Optimized Document Ordering". (WWW2009)
    /// </summary>
    public static class NewPFor
    {
        private static uint[] maxCodes = new uint[33];

        static NewPFor()
        {
            for (int i = 0; i < 33; i++)
            {
                maxCodes[i] = (uint)((1 << i) - 1);
            }
        }

        public static int Encode(
            int width, 
            uint frameOfReference, 
            uint[] input, 
            uint[] intermediateBuffer,
            int inputOffset, 
            uint[] output, 
            uint[] exceptionIndexes, 
            uint[] exceptions)
        {
            uint maxCode = maxCodes[width];
            int j = 0;
            for (uint i = 0; i < 128; i++)
            {   // TODO: Double cursor?
                uint value = input[inputOffset + i] - frameOfReference;
                intermediateBuffer[i] = value & maxCode;
                exceptionIndexes[j] = i;
                exceptions[j] = value >> width;
                if (exceptions[j] > 0)
                {
                    j++;
                }

                /* this is slower than using IF
                bool exception = exceptions[j] > 0;
                j += *(byte*)&(exception) & 1;
                 */
            }

            for (int i = j - 1; i > 0; i--)
            {
                exceptionIndexes[i] -= exceptionIndexes[i - 1] + 1;
            }

            FastPack.Fast128Pack(width, intermediateBuffer, output);
            return j;
        }

        public static unsafe void Decode(int n, int width, uint frameOfReference, int numberOfExceptions, uint[] encoded, uint[] decoded, uint[] exceptions, uint[] exceptionIndexes, int encodedOffset, int decodedOffset)
        {
            fixed (uint* encPtr = &encoded[encodedOffset])
            {
                fixed (uint* decPtr = &decoded[decodedOffset])
                {
                    Decode(n, width, frameOfReference, numberOfExceptions, encPtr, decPtr, exceptions, exceptionIndexes);
                }
            }
        }

        public static unsafe void Decode(
            int n, 
            int width, 
            uint frameOfReference, 
            int numberOfExceptions, 
            uint* encoded, 
            uint* decoded, 
            uint[] exceptions, 
            uint[] exceptionIndexes)
        {
            FastPack.Fast128UnPack(width, encoded, decoded);

            int exceptionIndex = -1;
            for (int i = 0; i < numberOfExceptions; i++)
            {
                exceptionIndex += (int)exceptionIndexes[i] + 1;
                decoded[exceptionIndex] |= exceptions[i] << width;
            }

            if (frameOfReference > 0)
            {
                for (int i = 0; i < 128; i++)
                {
                    decoded[i] += frameOfReference;
                }
            }
        }
    }
}


