﻿namespace NETing.Index.Utils.Compression
{
    public static class PrefixedVByte 
    {
        private const ushort Length1Max = 128;
        private const uint Length1MaxI = Length1Max;
        private const ushort Length2Max = 16384 + Length1Max;
        private const uint Length2MaxI = Length2Max;

        private const uint Length3Max = 2097152 + Length2Max;
        private const uint Length4Max = 16777216 + Length3Max;

        private const ulong Length5Max = 34359738368 + Length4Max;
        private const ulong Length6Max = 4398046511104 + Length5Max;
        private const ulong Length7Max = 562949953421312 + Length6Max;
        private const ulong Length8Max = 72057594037927936 + Length7Max;

        private static readonly ushort[] offsetUint16
            = new ushort[] { 0, Length1Max, Length2Max };

        private static readonly uint[] offsetUint32 
            = new uint[] { 0, Length1Max, Length2Max, Length3Max, Length4Max };

        private static readonly ulong[] offsetUint64
            = new ulong[] { 0, Length1Max, Length2Max, Length3Max, Length4Max, Length5Max, Length6Max, Length7Max, Length8Max };

        private static readonly uint[] signalMasks = new uint[] { 1, 3, 7, 15, 31, 63, 127, 255 };
        private static readonly uint[] signalValues = new uint[] { 0, 1, 3, 7, 15, 31, 63, 127 };
        private static readonly uint[] codeLength = new uint[256];

        private static readonly ushort[] codeOffsetUint16 = new ushort[256];
        private static readonly uint[] codeOffsetUint32 = new uint[256];
        private static readonly ulong[] codeOffsetUint64 = new ulong[256];

        static PrefixedVByte()
        {
            for (uint i = 0; i < byte.MaxValue; i++)
            {
                for (uint j = 0; j < signalMasks.Length; j++)
                {
                    if ((i & signalMasks[j]) == signalValues[j])
                    {
                        Assign(i, j);
                        break;
                    }
                }
            }
        }

        private static void Assign(uint byteValue, uint length)
        {
            codeLength[byteValue] = length + 1;
            if (length < 3)
            {
                codeOffsetUint16[byteValue] = (ushort)((byteValue >> (int)(length + 1)) + offsetUint16[length]);
                codeOffsetUint32[byteValue] = codeOffsetUint16[byteValue];
                codeOffsetUint64[byteValue] = codeOffsetUint16[byteValue];
            }
            else if (length < 5)
            {
                codeOffsetUint32[byteValue] = (byteValue >> (int)(length + 1)) + offsetUint32[length];
                codeOffsetUint64[byteValue] = codeOffsetUint32[byteValue];
            }
            else
            {
                codeOffsetUint64[byteValue] = ((byteValue >> (int)length) + offsetUint64[length]) % ulong.MaxValue;
            }
        }

        public static int Encode(int value, byte[] buffer, int offset)
        {
            return Encode((uint) value, buffer, offset);
        }

        public static int Encode(uint value, byte[] buffer, int offset)
        {
            uint encode;
            if (value < Length1Max)
            {
                buffer[offset] = (byte)(value << 1);
                return offset + 1;
            }

            if (value < Length2Max)
            {
                encode = ((value - Length1Max) << 2) | 1;
                buffer[offset] = (byte) encode;
                buffer[offset + 1] = (byte)(encode >> 8);
                return offset + 2;
            }

            if (value < Length3Max)
            {
                encode = ((value - Length2Max) << 3) | 3;
                buffer[offset] = (byte)encode;
                buffer[offset + 1] = (byte)(encode >> 8);
                buffer[offset + 2] = (byte)(encode >> 16);
                return offset + 3;
            }

            if (value < Length4Max)
            {
                encode = ((value - Length3Max) << 4) | 7;
                buffer[offset] = (byte)encode;
                buffer[offset + 1] = (byte)(encode >> 8);
                buffer[offset + 2] = (byte)(encode >> 16);
                buffer[offset + 3] = (byte)(encode >> 24);
                return offset + 4;
            }

            ulong longEncode = (((ulong)value - Length4Max) << 5) | 15;
            buffer[offset] = (byte)longEncode;
            buffer[offset + 1] = (byte)(longEncode >> 8);
            buffer[offset + 2] = (byte)(longEncode >> 16);
            buffer[offset + 3] = (byte)(longEncode >> 24);
            buffer[offset + 4] = (byte)(longEncode >> 32);
            return offset + 5;
        }

        public static int Decode(byte[] input, int position, out int decoded)
        {
            uint result;
            int retval = Decode(input, position, out result);
            decoded = (int)result;
            return retval;
        }

        public static int Decode(byte[] input, int position, out uint decoded)
        {
            decoded = codeOffsetUint32[input[position]];
            if (decoded < Length1MaxI)
            {
                return position + 1;
            }

            if (decoded < Length2MaxI)
            {
                decoded += (uint)(input[position + 1] << 6);
                return position + 2;
            }

            if (decoded < Length3Max)
            {
                decoded += (uint)((input[position + 2] << 13) | (input[position + 1] << 5));
                return position + 3;
            }

            if (decoded < Length4Max)
            {
                decoded += (uint)((input[position + 3] << 20) | (input[position + 2] << 12) | (input[position + 1] << 4));
                return position + 4;
            }

            decoded += (uint)((input[position + 4] << 27) | (input[position + 3] << 19) | (input[position + 2] << 11) | (input[position + 1] << 3));
            return position + 5;
        }

        public static int Decode(byte[] input, int position, out ushort decoded)
        {
            decoded = codeOffsetUint16[input[position]];
            if (decoded < Length1Max)
            {
                return position + 1;
            }

            if (decoded < Length2Max)
            {
                decoded += (ushort)(input[position + 1] << 6);
                return position + 2;
            }
            
            decoded += (ushort)((input[position + 2] << 13) | (input[position + 1] << 5));
            return position + 3;
        }

        private static int SkipOne(byte[] input, int position)
        {
            return position + (int)codeLength[input[position]];
        }

        public static int SkipMany(byte[] input, int position, int n)
        {
            while (n > 31)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 32;
            }

            while (n > 15)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 16;
            }

            while (n > 7)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);

                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 8;
            }

            while (n > 3)
            {
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                position = SkipOne(input, position);
                n -= 4;
            }

            while (n > 0)
            {
                position = SkipOne(input, position);
                n--;
            }

            return position;
        }
    }
}


