﻿namespace NETing.Integration
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Threading.Tasks;
    
    using Index.ColumnAccess.Readers;
    using Index.ColumnAccess.Writers;
    using Index.Controller;
    using Index.Controller.Configurations;
    using Index.Controller.Configurations.ConcreteConfigurations;
    using Index.Controller.Services;
    using Index.Documents.DocumentModel;
    using Index.Documents.DocumentModel.Scope;
    using Index.Documents.FileSource;
    using Index.Documents.Parser;
    using Index.FileProcessor;
    using Index.PostingFile;
    using Index.Utils.Collections;
    using NUnit.Framework;
    
    [TestFixture]
    public sealed class SkippableIndexIntegrationTest
    {
        #region Generalized Test Code

        private static readonly DirectoryInfo testCollection 
            = new DirectoryInfo("C:\\Users\\Ola Natvig\\XmlData\\SmallSample");

        private static readonly FileProcessor fileProcessor = new FileProcessor();

        private sealed class TestIndexConfig : IIndexSystemConfig
        {
            public int BlockSize { get; internal set; }
            public int BufferCount { get; internal set; }
            public int BufferSlices { get; internal set; }
            public bool BuffersReadOnly { get; internal set; }
            public int BufferFlushInterval { get; internal set; }
        }

        private sealed class TestPostingFileConfig : IPostingFileConfig
        {
            public string Name
            {
                get { return "Test"; }
            }

            public ScopeDirectory ScopeDirectory { get; set; }

            public AbstractDocumentIdColumnWriter DocumentIdWriter { get; internal set; }
            public AbstractFrequencyColumnWriter FrequencyWriter { get; internal set; }
            public AbstractPositionColumnWriter PositionWriter { get; internal set; }
            public AbstractScopeColumnWriter ScopeWriter { get; internal set; }

            public AbstractDocumentColumnReader DocumentReader { get; internal set; }
            public AbstractPositionColumnReader PositionReader { get; internal set; }
            public AbstractScopeColumnReader ScopeReader { get; internal set; }
        }

        private sealed class TestBuilderServiceConfig : IBuilderServiceConfig
        {
            public int StoreSize { get; internal set; }
            public int InitialListSize { get; internal set; }
        }

        private static readonly IBuilderServiceConfig builderConfig = new TestBuilderServiceConfig
         {
             InitialListSize = 20,
             StoreSize = 1024 * 1024 * 50
         };

        private static readonly StorageProfile storageProfile = new StorageProfile(
            new FileInfo("dictionary.integration.neting"),
            new FileInfo("scopeDirectory.integration.neting"),
            new FileInfo("postingFile.integration.neting"));

        private static IndexController CreateController(
            int blockSize, 
            int bufferCount,
            int bufferSlices,
            int bufferFlushInterval,
            AbstractDocumentIdColumnWriter documentIdWriter,
            AbstractFrequencyColumnWriter freqColumnWriter,
            AbstractPositionColumnWriter posColumnColumnWriter,
            AbstractScopeColumnWriter scopeColumnWriter,
            AbstractDocumentColumnReader documentReader,
            AbstractPositionColumnReader posColumnReader,
            AbstractScopeColumnReader scopeColumnReader)
        {
            IPostingFileConfig postingConfig = new TestPostingFileConfig
                                                   {
                                                       DocumentIdWriter = documentIdWriter,
                                                       FrequencyWriter = freqColumnWriter,
                                                       PositionWriter = posColumnColumnWriter,
                                                       ScopeWriter = scopeColumnWriter,

                                                       DocumentReader =  documentReader,
                                                       PositionReader = posColumnReader,
                                                       ScopeReader = scopeColumnReader
                                                   };

            return CreateController(blockSize, bufferCount, bufferSlices, bufferFlushInterval, postingConfig);
        }

        private static IndexController CreateController(
            int blockSize,
            int bufferCount,
            int bufferSlices,
            int bufferFlushInterval,
            IPostingFileConfig postingFileConfig)
        {
            IIndexSystemConfig indexConfig = new TestIndexConfig
                                                 {
                                                     BlockSize = blockSize,
                                                     BufferCount = bufferCount,
                                                     BufferSlices = bufferSlices,
                                                     BufferFlushInterval = bufferFlushInterval
                                                 };

            return new IndexController(
                storageProfile,
                indexConfig,
                postingFileConfig); 
        }

        private static void TestConfiguration(IPostingFileConfig config)
        {
            var controller = CreateController(
                     1024 * 64,
                     20,
                     2,
                     100,
                     config);
            TestCorrectness(controller);
            controller.Shutdown();
        }

        private static void TestCorrectness(IndexController controller)
        {
            var crawler = new DirectoryCrawler(testCollection, true);
            BuilderService builder = controller.PrepareBuild(builderConfig);
            builder.Start();
            foreach (var file in crawler)
            {
                builder.Add(fileProcessor.PrepareFile(file));
            }

            builder.PrepareFinish();
            Console.WriteLine("Finished feeding, finishing index!");
            builder.Finish();
            Console.WriteLine("Finished index! Will start Query!");
            Console.WriteLine("Writing out data structures.");
            builder.Write();
            Console.WriteLine("Data structures written!");

            controller.Restart();

            QueryService query = controller.PrepareQuery();

            Console.WriteLine("Ready to run queries.");

            var terms = TestThatIndexEqualsInversion(
                builder.ReadInMemoryInversion(), query);

            Console.WriteLine("Reading lists in parallel!");

            TestReadIndexInParallel(terms, query);

            Console.WriteLine("Testing document id skipping.");
            TestDocumentIdSkipping(query, terms);

            Console.WriteLine("Testing path type skipping.");
            TestPathTypeSkipping(query, terms, controller.ScopeDirectoryService.ScopeDirectory);

            Console.WriteLine("Testing scope skipping");
            TestScopeSkipping(query, terms);
        }

        #endregion

        #region Test inverted list reading

        private static IEnumerable<Term> TestThatIndexEqualsInversion(
            IEnumerator<Tuple<Term, IEnumerator<Posting>>> inversion,
            QueryService queryService)
        {
            var terms = new List<Term>();
            while (inversion.MoveNext())
            {
                terms.Add(inversion.Current.Item1);
                AssertCorrectInvertedList(
                 queryService.Lookup(inversion.Current.Item1),
                 inversion.Current.Item2);
            }

            return terms;
        }

        private static void TestReadIndexInParallel(
            IEnumerable<Term> terms,
            QueryService queryService)
        {
            Assert.That(Parallel.ForEach(terms, term => AssertCorrectInvertedList(
                queryService.Lookup(term), queryService.Lookup(term))).IsCompleted);
        }

        #endregion  
        
        #region Test document Id skipping

        private static int[] skipDocs = new[] { 1, 5, 10, 20, 50, 200 };

        private static void TestDocumentIdSkipping(QueryService service, IEnumerable<Term> terms)
        {
            foreach (int interval in skipDocs)
            {
                Console.WriteLine("Skipping " + interval + " docuements in each list.");

                foreach (Term term in terms)
                {
                    using (var iter = service.Lookup(term))
                    {
                        using (var skip = service.Lookup(term))
                        {
                            TestDocumentIdSkip(iter, skip, interval);
                        }
                    }
                }
            }
        }

        private static void TestDocumentIdSkip(PostingFileEnumerator iterate, PostingFileEnumerator skip, int interval)
        {
            Assert.That(iterate.MoveNext()); // no lists are empty
            while (SkipNDocuments(iterate, interval))
            {
                Assert.That(skip.SkipToDocumentId(iterate.Current.DocumentId));
                skip.SynchronizeWithDocumentColumn();
                AssertSameDocument(iterate, skip);
            }
        }

        private static bool SkipNDocuments(PostingFileEnumerator enumerator, int skipDocs)
        {
            int skipped = 0;

            while (skipped < skipDocs)
            {
                uint docId = enumerator.Current.DocumentId;
                bool hasMore = enumerator.MoveNext();
                while (hasMore && enumerator.Current.DocumentId == docId)
                {
                    hasMore = enumerator.MoveNext();
                }

                if (!hasMore)
                {
                    return false;
                }

                skipped++;
            }

            return true;
        }

        private static void AssertSameDocument(
            PostingFileEnumerator enumerator1, PostingFileEnumerator enumerator2)
        {
            bool enum1HasMore, enum2HasMore;
            uint docId = enumerator1.Current.DocumentId;
            do
            {
                Assert.AreEqual(enumerator1.Current, enumerator2.Current);

                if (enumerator1.Current.DocumentId != docId)
                {
                    return;
                }

                enum1HasMore = enumerator1.MoveNext();
                enum2HasMore = enumerator2.MoveNext();
            } while (enum1HasMore & enum2HasMore);
        }

        private static void AssertCorrectInvertedList(
            IEnumerator<Posting> actual, 
            IEnumerator<Posting> expected)
        {
            using (actual)
            {
                using (expected)
                {
                    bool aMore = actual.MoveNext();
                    bool eMore = expected.MoveNext();
                    while (aMore & eMore)
                    {
                        Assert.AreEqual(expected.Current, actual.Current);
                        aMore = actual.MoveNext();
                        eMore = expected.MoveNext();
                    }

                    Assert.IsFalse(aMore);
                    Assert.IsFalse(eMore);
                }
            }
        }

        #endregion

        #region Test path type skipping

        private static int[] skipToPathTypeCount = new[] { 5, 10, 15, 20, 200 };

        private static void TestPathTypeSkipping(
            QueryService queryService, 
            IEnumerable<Term> terms,
            ScopeDirectory scopeDirectory)
        {
            foreach (int count in skipToPathTypeCount)
            {
                Console.WriteLine("Skipping to a subset of the paths of size: " + count);
                MutableBitArray queryPathTypes = GeneratePathTypeSubset(scopeDirectory, count);
                foreach (Term term in terms)
                {
                    using (var iterate = queryService.Lookup(term))
                    {
                        using (var skip = queryService.Lookup(term))
                        {
                            TestPathTypeSkip(iterate, skip, queryPathTypes);
                        }
                    }
                }
            }
        }

        private static void TestPathTypeSkip(
            PostingFileEnumerator iterate, PostingFileEnumerator skip, MutableBitArray queryPathTypes)
        {
            while (SkipToPathType(iterate, queryPathTypes))
            {
                Assert.That(skip.SkipToPathTypes(queryPathTypes));
                Assert.AreEqual(iterate.Current, skip.Current);

                if (!skip.MoveNext())
                {
                    return;
                }
            }
        }

        private static bool SkipToPathType(PostingFileEnumerator enumerator, MutableBitArray queryPathTypes)
        {
            bool hasMore = enumerator.MoveNext();
            while (hasMore)
            {
                if (queryPathTypes[enumerator.Current.Scope.PathType.PathId])
                {
                    break;
                }

                hasMore = enumerator.MoveNext();
            }

            return hasMore;
        }

        private static MutableBitArray GeneratePathTypeSubset(ScopeDirectory scopeDirectory, int count)
        {
            var random = new Random(100);
            var query = new MutableBitArray();
            for (int i = 0; i < count; i++)
            {
                int selected = random.Next(scopeDirectory.NumberOfPaths);
                int changed = 0;
                while (query[selected] && changed < scopeDirectory.NumberOfPaths)
                {
                    changed++;
                    selected = (selected + 1) % scopeDirectory.NumberOfPaths;
                } 

                if (changed == scopeDirectory.NumberOfPaths)
                {
                    break;
                }

                query[selected] = true;
            }

            return query;
        }

        #endregion

        #region Test scope skipping

        private static int[] skipIntoDocumentScopePositions = new[] { 1, 2, 3, 4, 5, 10, 20, 50, 100 };

        private static void TestScopeSkipping(
            QueryService queryService,
            IEnumerable<Term> terms)
        {
            foreach (int position in skipIntoDocumentScopePositions)
            {
                Console.WriteLine("Skipping scopes at " + position + "-th position.");
                foreach (Term term in terms)
                {
                    using (PostingFileEnumerator iterate = queryService.Lookup(term))
                    {
                        using (PostingFileEnumerator skip = queryService.Lookup(term))
                        {
                            TestScopeSkipping(iterate, skip, position);
                        }
                    }
                }
            }
        }

        private static void TestScopeSkipping(
            PostingFileEnumerator iterate,
            PostingFileEnumerator skip,
            int position)
        {
            int uniquePrefix;
            Assert.That(iterate.MoveNext());
            while (MoveToNthUniquePrefixInNextDocument(iterate, position, out uniquePrefix))
            {
                bool foundMatch;
                Assert.That(skip.SkipToDocumentId(iterate.Current.DocumentId));
                skip.SynchronizeWithDocumentColumn();
                Assert.That(skip.SkipToScope(iterate.Current.Scope, uniquePrefix, out foundMatch));
                Assert.That(foundMatch);
                Assert.AreEqual(iterate.Current, skip.Current);
            }
        }

        private static bool MoveToNthUniquePrefixInNextDocument(
            PostingFileEnumerator enumerator,
            int n,
            out int uniquePrefix)
        {
            uniquePrefix = 0;
            uint currentDocId = enumerator.Current.DocumentId;
            bool hasMore = enumerator.MoveNext();
            while (enumerator.Current.DocumentId == currentDocId && hasMore)
            {
                hasMore = enumerator.MoveNext();
            }

            if (!hasMore)
            {
                return false;
            }
            
            while (enumerator.Current.Frequency < n && hasMore)
            {
                hasMore = enumerator.MoveNext();  
            }

            Scope previous = enumerator.Current.Scope.Duplicate();
            for (int i = 0; i < n; i++)
            {
                previous = enumerator.Current.Scope.Duplicate();
                hasMore = enumerator.MoveNext(); // guarateed
            }

            currentDocId = enumerator.Current.DocumentId;
            while (enumerator.Current.DocumentId == currentDocId 
                && hasMore 
                && previous.Lcp(enumerator.Current.Scope) == enumerator.Current.Scope.PathType.Depth)
            {
                previous = enumerator.Current.Scope.Duplicate();
                hasMore = enumerator.MoveNext();
            }

            if (enumerator.Current.DocumentId != currentDocId)
            {
                return MoveToNthUniquePrefixInNextDocument(enumerator, n, out uniquePrefix);
            }

            if (hasMore)
            {
                uniquePrefix = previous.Lcp(enumerator.Current.Scope) + 1;
            }
            else
            {
                uniquePrefix = 0;
            }

            return hasMore;
        }

        #endregion

        [Test]
        public void TestRaw()
        {
            TestConfiguration(new RawPostingFileConfig());
        }

        [Test]
        public void TestNVByte()
        {
            TestConfiguration(new NVBytePostingFileConfig());
        }

        [Test]
        public void TestVByte()
        {
            TestConfiguration(new VBytePostingFileConfig());
        }

        [Test]
        public void TestNPfor()
        {
            TestConfiguration(new NPforPostingFileConfig());
        }

        [Test]
        public void TestCNPfor()
        {
            TestConfiguration(new CNPforPostingFileConfig());
        }

        [Test]
        public void TestDCNPfor()
        {
            TestConfiguration(new DCNPforPostingFileConfig());
        }

        [Test]
        public void TestDCNPfor32()
        {
            TestConfiguration(new DCNPforPostingFileConfig(32));
        }

        [Test]
        public void TestPfxCNPfor()
        {
            TestConfiguration(new PfxCNPforPostingFileConfig());
        }

        [Test]
        public void TestPtsNPfor()
        {
            TestConfiguration(new PtsNPforPostingFileConfig());
        }

        [Test]
        public void TestPtsCNPfor()
        {
            TestConfiguration(new PtsCNPforPostingFileConfig());
        }
    }
}
